<?php
/**
 * Refae\Controller\AppController
 */

namespace Refae\Controller;

use AsalaeCore\Controller\Component\ConditionComponent;
use AsalaeCore\Controller\Component\FilterComponent;
use AsalaeCore\Controller\Component\JstableComponent;
use AsalaeCore\Controller\Component\ModalComponent;
use AsalaeCore\Controller\Component\RestComponent;
use Cake\Cache\Cache;
use Cake\Controller\Controller;
use Cake\Core\Configure;
use Cake\Database\Expression\IdentifierExpression;
use Cake\Event\Event;
use Refae\Model\Table\CategoriesTable;
use Refae\Model\Table\DocumentsTable;

/**
 * Application Controller
 *
 * @author    Ludovic Ganée <ludovic.ganee@libriciel.coop>
 * @copyright (c) 2017-2018, Libriciel SCOP
 * @license   https://opensource.org/licenses/AGPL-3.0
 * @property RestComponent $Rest
 * @property ModalComponent $Modal
 * @property JstableComponent $Jstable
 * @property FilterComponent $Filter
 * @property ConditionComponent $Condition
 * @property DocumentsTable $Documents
 * @property CategoriesTable $Categories
 */
class AppController extends Controller
{

    /**
     * An array containing the names of helpers this controller uses. The array elements should
     * not contain the "Helper" part of the class name.
     *
     * Example:
     * ```
     * public $helpers = ['Form', 'Html', 'Time'];
     * ```
     *
     * @var  array
     * @link http://book.cakephp.org/3.0/en/controllers.html#configuring-helpers-to-load
     */
    public $helpers = [
        "Breadcrumbs" => ['className' => 'Bootstrap.Breadcrumbs'],
        'Form' => ['className' => 'AsalaeCore.Form'],
        'Html' => ['className' => 'Bootstrap.Html'],
        'Modal' => [
            'className' => 'Bootstrap.Modal',
            'templates' => [
                'modalTitle' => '<h3 class="h4 modal-title{{attrs.class}}"{{attrs}}>{{content}}</h3>',
                'modalContentEnd' => "<div class=\"ui-draggable-handle bottom-handle\"></div>\n</div>\n",
                'modalEnd' => "{{contentEnd}}{{dialogEnd}}</div>\n",
                'modalDialogEnd' => "</div>\n",
                'headerEnd' => "</div>\n",
                'bodyEnd' => "</div>\n",
                'footerEnd' => "</div>\n",
            ]
        ],
        'ModalForm' => ['className' => 'AsalaeCore.ModalForm'],
        'Navbar' => ['className' => 'Bootstrap.Navbar'],
        'Paginator' => ['className' => 'AsalaeCore.Paginator'],
        'Table' => ['className' => 'AsalaeCore.Table'],
        'Panel' => ['className' => 'Bootstrap.Panel'],
        'Upload' => ['className' => 'AsalaeCore.Upload'],
        'Filter' => ['className' => 'AsalaeCore.Filter'],
        'Select',
        'ModalView' => ['className' => 'AsalaeCore.ModalView'],
        'Fa' => ['className' => 'AsalaeCore.Fa'],
    ];

    /**
     * Liste les modals utilisés
     *
     * @var array
     */
    public $modals = [];

    /**
     * Initialization hook method.
     *
     * Use this method to add common initialization code like loading components.
     *
     * e.g. `$this->loadComponent('Security');`
     */
    public function initialize()
    {
        if (Configure::read('debug')) {
            Cache::disable();
        }

        ob_start();
        parent::initialize();

        $this->loadComponent('RequestHandler', ['enableBeforeRedirect' => false]);
        $this->loadComponent('Flash');
        $this->loadComponent('Auth', [
            'unauthorizedRedirect' => false,
            'loginAction' => '/',
            'logoutRedirect' => '/',
            'authError' => __("Vous n'avez pas l'autorisation d'accéder à cette page."),
            'flash' => ['element' => 'error']
        ]);
        $this->loadComponent('AsalaeCore.Rest');

        if ($this->getRequest()->is('ajax')) {
            $this->Auth->setConfig('unauthorizedRedirect', false);
            $connected = (bool)$this->getRequest()->getSession()->read('User.id');
            $this->setResponse($this->getResponse()->withHeader(
                'X-Connected',
                $connected ? 'true' : 'false'
            ));
            if ($this->getRequest()->getHeaderLine('No-Layout')) {
                $this->viewBuilder()->setLayout('no-layout');
            }
        }
    }

    /**
     * Ajoute à la vue le nécéssaire pour mettre à jour les données de recherche
     * pour la search-bar
     * @return string json_encoded searchData
     */
    protected function addSearchData(): string
    {
        $this->loadModel('Documents');
        $this->loadModel('Categories');
        $lastVersionSq = $this->Documents->withAlias('subDocuments')->find()
            ->select(['subDocuments.version'])
            ->where([
                'subDocuments.identifiant' => new IdentifierExpression('Documents.identifiant'),
                'subDocuments.status' => 'actif'
            ])
            ->order(['subDocuments.version' => 'desc'])
            ->limit(1)
        ;

        // searchbar
        $documents = $this->Documents->find()
            ->select(['Documents.id', 'Documents.name'])
            ->where([
                'Documents.status' => 'actif',
                'Documents.version' => $lastVersionSq,
            ])
            ->all()
            ->toArray()
        ;
        $categories = $this->Categories->find()
            ->select(['Categories.id', 'Categories.name'])
            ->all()
            ->toArray()
        ;
        $this->set('searchData', json_encode(compact('documents', 'categories')));
        return json_encode(compact('documents', 'categories'));
    }

    /**
     * Before render callback.
     *
     * @param  Event $event The beforeRender event.
     * @return \Cake\Network\Response|null
     */
    public function beforeRender(Event $event)
    {
        $this->set('user_id', $this->getRequest()->getSession()->read('User.id'));

        // modals
        $this->set('modals', $this->modals);

        $preHtml = ob_get_contents();
        $this->set('preHtml', $preHtml);

        if (ob_get_contents()) {
            ob_end_clean();
        }

        return null;
    }
}
