<?php
/**
 * Refae\Controller\DocumentsController
 */
namespace Refae\Controller;

use AsalaeCore\Controller\ApiTrait;
use AsalaeCore\Utility\XmlUtility;
use Cake\Database\Connection;
use Cake\Database\Expression\IdentifierExpression;
use Cake\Datasource\ConnectionManager;
use Cake\Datasource\EntityInterface;
use Cake\Event\Event;
use Cake\Http\Exception\BadRequestException;
use Cake\ORM\Entity;
use Cake\ORM\Query;
use Cake\ORM\TableRegistry;
use Cake\Utility\Hash;
use DateTime;
use DateTimeInterface;
use Exception;
use Libriciel\Filesystem\Utility\Filesystem;
use Refae\Model\Table\CategoriesTable;
use Refae\Model\Table\DocumentsTable;
use Refae\Model\Table\FilesTable;

/**
 * Controller de la table users
 *
 * @author    Ludovic Ganée <ludovic.ganee@libriciel.coop>
 * @copyright (c) 2017-2018, Libriciel SCOP
 * @license   https://opensource.org/licenses/AGPL-3.0
 * @property FilesTable $Files
 * @property bool|object CategoriesDocuments
 * @property DocumentsTable $Documents
 */
class DocumentsController extends AppController
{
    /**
     * Traits
     */
    use ApiTrait;

    /**
     * Options de pagination par défaut
     *
     * @var array
     */
    public $paginate = [
        'limit' => 10,
        'order' => [
            'Documents.id' => 'desc'
        ],
        'contain' => ['Categories'],
        'sortWhitelist' => [
            'identifiant',
            'name',
            'dua',
            'duc',
            'communicabilite',
            'sort_final',
            'support',
            'file_format',
            'created',
            'modified',
            'version',
            'status',
        ]
    ];

    /**
     * Called before the controller action. You can use this method to configure and customize components
     * or perform logic that needs to happen before each controller action.
     *
     * @param \Cake\Event\Event $event An Event instance
     * @return \Cake\Http\Response|null
     * @link https://book.cakephp.org/3.0/en/controllers.html#request-life-cycle-callbacks
     */
    public function beforeFilter(Event $event)
    {
        $this->Auth->allow(['index', 'view', 'find']);
        return parent::beforeFilter($event);
    }

    /**
     * Liste des documents
     * @param string $status
     * @return \Cake\Http\Response|null|void
     * @throws Exception
     */
    public function index(string $status = 'actif')
    {
        $this->loadComponent('Paginator');
        $this->loadComponent('AsalaeCore.Condition');
        $this->loadComponent('AsalaeCore.Jstable');
        $this->loadModel('Documents');
        $this->loadModel('Categories');
        $tableId = "documents-table";
        $this->Jstable->configurablePagination($tableId);
        $this->set('tableId', $tableId);
        $title = __("Liste des règles de gestion d’un document");
        /** @var Query $query */
        $query = $this->Documents->find()
            ->contain([
                'Categories' => ['sort' => ['name' => 'asc']],
                'Files',
            ])
        ;
        $lastVersionSq = $this->Documents->withAlias('subDocuments')->find()
            ->select(['subDocuments.version'])
            ->where([
                'subDocuments.identifiant' => new IdentifierExpression('Documents.identifiant')
            ])
            ->order(['version' => 'desc'])
            ->limit(1)
        ;
        if ($this->Auth->user('id') && ($stats = $this->getRequest()->getParam('?.status'))) {
            foreach ($stats as $stat) {
                $query->where($this->Condition->in('status', $stat));
            }
        } elseif ($status === 'actif' || !$this->Auth->user('id')) {
            $query->where([
                'Documents.status' => 'actif',
                'Documents.version' => $lastVersionSq->andWhere(['subDocuments.status' => 'actif']),
            ]);
        } elseif ($status === 'brouillon') {
            $query->where(['Documents.status' => 'brouillon']);
            $title .= ' ('.__("en attente de publication").')';
        } elseif ($status === 'obsolete') {
            $query->where(['Documents.status' => 'obsolete']);
            $title .= ' ('.__("obsolètes").')';
        }
        if ($this->getRequest()->getParam('?.last-version')) {
            $query->where(['version' => $lastVersionSq]);
        }
        $subcats = [];
        $subcat_id = (int)$this->getRequest()->getParam('?.subcat');
        $table_id = (int)$this->getRequest()->getParam('?.table_id');
        $searchcat_id = (int)$this->getRequest()->getParam('?.searchcat');
        $parents = [];
        $categoryDeletable = true;

        if (empty($table_id) && $subcat_id) {
            /** @var CategoriesTable $subCategories */
            $subCategories = clone $this->Categories;
            $subCategories->setAlias('subCategories');
            $subQuery = $subCategories->find()
                ->select(['subCategories.id'])
                ->where([
                    'subCategories.lft <' =>  new IdentifierExpression('Categories.lft'),
                    'subCategories.rght >' =>  new IdentifierExpression('Categories.rght'),
                ])
                ->order(['subCategories.lft' => 'asc'])
                ->limit(1)
            ;
            $table_id = $this->Categories->find()
                ->select(['sub_id' => $subQuery])
                ->where(['Categories.id' => $subcat_id])
                ->firstOrFail()
                ->get('sub_id')
            ;
        }
        if ($searchcat_id) {
            $cat = $this->Categories->find()->where(['id' => $searchcat_id])->first();
            $table_id = $this->Categories->find()
                ->select(['Categories.id'])
                ->where([
                    'Categories.lft <=' => $cat->get('lft'),
                    'Categories.rght >=' => $cat->get('rght')
                ])
                ->order(['Categories.lft' => 'asc'])
                ->first()
                ->get('id')
            ;
            if ($cat->get('id') !== $table_id) {
                $subcat_id = $cat->get('id');
            }
        }
        if ($table_id) {
            $cat = $this->Categories->find()->where(['id' => $table_id])->first();
            $subcats = $this->Categories->find('threaded')
                ->where([
                    'Categories.lft >' => $cat->get('lft'),
                    'Categories.rght <' => $cat->get('rght')
                ])
                ->order(['Categories.lft' => 'asc'])
                ->toArray()
            ;
            if (empty($subcat_id)) {
                $ids = $this->Categories->find()
                    ->select(['Categories.id'])
                    ->where([
                        'Categories.lft >=' => $cat->get('lft'),
                        'Categories.rght <=' => $cat->get('rght')
                    ])
                    ->all()
                    ->extract('id')
                    ->toArray()
                ;
                $Condition = $this->Condition;
                $query->contain([
                    'Categories' => function (Query $q) use ($ids, $Condition) {
                        return $q->where($Condition->in('Categories.id', $ids));
                    },
                    'Files',
                ]);
                $subquery = TableRegistry::getTableLocator()->get('CategoriesDocuments')->find()
                    ->select(['CategoriesDocuments.document_id'])
                    ->where($this->Condition->in('CategoriesDocuments.category_id', $ids))
                ;
                $query->where($this->Condition->in('Documents.id', $subquery));
            }
            $title = $cat->get('name');
            $catDesc = $cat->get('description');
            $parents[$table_id] = $title;
            $categoryDeletable = $cat->get('rght') === $cat->get('lft') +1 && $query->count() === 0;
        }
        if ($subcat_id) {
            $subcat = $this->Categories->find()->where(['id' => $subcat_id])->first();
            $query->contain([
                'Categories' => function (Query $q) use ($subcat) {
                    return $q->where([
                        'Categories.lft >=' => $subcat->get('lft'),
                        'Categories.rght <=' => $subcat->get('rght'),
                    ]);
                },
                'Files',
            ]);
            $CategoriesDocuments = TableRegistry::getTableLocator()->get('CategoriesDocuments');
            $CategoriesDocuments->belongsTo('Categories');
            $subquery = $CategoriesDocuments->find()
                ->select(['CategoriesDocuments.document_id'])
                ->innerJoinWith('Categories', function (Query $q) use ($subcat) {
                    return $q->where([
                        'Categories.lft >=' => $subcat->get('lft'),
                        'Categories.rght <=' => $subcat->get('rght'),
                    ]);
                })
            ;
            $query->where($this->Condition->in('Documents.id', $subquery));

            $title = $subcat->get('name');

            $parents += $this->Categories->find('list')
                ->where([
                    'Categories.lft <' => $subcat->get('lft'),
                    'Categories.rght >' => $subcat->get('lft'),
                    'Categories.id !=' => $table_id
                ])
                ->order(['Categories.lft' => 'asc'])
                ->toArray()
            ;
            $catDesc = $subcat->get('description');

            $categoryDeletable = $subcat->get('rght') === $subcat->get('lft') +1 && $query->count() === 0;
        }
        $this->set(compact('subcats', 'title', 'table_id', 'catDesc', 'subcat_id', 'parents', 'categoryDeletable'));

        // Order by favorite
        if ($this->getRequest()->getParam('?.sort') === 'favorite') {
            $this->loadComponent('AsalaeCore.Condition');
            $dir = $this->getRequest()->getParam('?.direction') ?: 'asc';
            $query->order($this->Condition->orderByFavorites($dir));
        } elseif (!$this->getRequest()->getParam('?.sort')) {
            $query->order(['Documents.id' => 'desc']);
        }

        try {
            $this->appendFiltersConditions($query);
        } catch (BadRequestException $e) {
            $this->Flash->error($e->getMessage());
        }

        if (in_array($export = $this->getRequest()->getParam('?.export', ''), ['csv', 'xml', 'json', 'database'])) {
            $output = array_map(function ($v) {
                if ($v instanceof Entity) {
                    $v = $v->toArray();
                }
                return $v;
            }, $query->all()->toArray());
            $response = $this->getResponse();
            switch ($export) {
                case 'csv':
                    $this->setResponse($response->withDownload('export.csv'));
                    $this->exportCsv($output);
                    break;
                case 'json':
                    $this->setResponse($response->withDownload('export.json'));
                    return $this->renderJson(json_encode($output));
                case 'xml':
                    $this->setResponse($response->withDownload('export.xml'));
                    return $this->renderXml(XmlUtility::arrayToXmlString(['root' => ['Document' => [$output]]]));
                case 'database':
                    return $this->exportDatabase($query);
            }
            return;
        }

        // Order by favorite
        if ($this->getRequest()->getParam('?.sort') === 'favorite') {
            $dir = $this->getRequest()->getParam('?.direction') ?: 'asc';
            $query->order($this->Condition->orderByFavorites($dir));
        }

        $data = $this->paginate($query)->toArray();
        $this->set('data', $data);

        $tableauxDeGestion = $this->Categories->find('list')
            ->where(['Categories.parent_id IS' => null])
            ->order(['Categories.name' => 'asc'])
            ->toArray()
        ;
        $this->set(compact("tableauxDeGestion"));

        $this->addSearchData();

        $categories = $this->Categories->find('list')->order(['name' => 'asc']);
        $this->set('categories', $categories->toArray());
        $this->set('status', $this->Documents->options('status'));
    }

    /**
     * Ajoute les filtres de recherche à $query
     * @param Query $query
     * @return $this
     * @throws Exception
     */
    private function appendFiltersConditions(Query $query)
    {
        $this->loadComponent('AsalaeCore.Filter');
        $this->Filter->loadAndSave();
        $params = $this->getRequest()->getParam('?') ?: [];

        $this->loadComponent('AsalaeCore.Condition');

        // Filtre identifiant
        foreach ((array)Hash::get($params, 'identifiant') as $value) {
            $query->where($this->Condition->ilike('Documents.identifiant', $value));
        }
        // Filtre name
        foreach ((array)Hash::get($params, 'name') as $value) {
            $query->where($this->Condition->ilike('Documents.name', $value));
        }
        // Filtre categories
        $this->loadModel('CategoriesDocuments');
        if ($categories = (array)Hash::get($params, 'categories')) {
            foreach ($categories as $values) {
                $whereCategories = $this->CategoriesDocuments->find()
                    ->select(['CategoriesDocuments.document_id'])
                    ->where($this->Condition->in('CategoriesDocuments.category_id', $values))
                ;
                $query->where($this->Condition->in('Documents.id', $whereCategories));
            }
        }
        // Filtre created
        foreach ((array)Hash::get($params, 'created') as $key => $value) {
            $operator = $this->getRequest()->getParam("?.dateoperator_created.$key");
            $query->where($this->Condition->dateOperator('Documents.created', $operator, $value));
        }
        // Filtre favoris
        if (Hash::get($params, 'favoris')) {
            $query->where($this->Condition->inFavorites('Documents.id'));
        }
        return $this;
    }

    /**
     * Ajout d'un document
     * @param string $category_id
     * @throws Exception
     */
    public function add(string $category_id = '')
    {
        $uploadId = 'add-upload-profil';
        $this->set('uploadId', $uploadId);
        $this->set('uploadBoxId', 'add-upload-document-box');
        $this->loadModel('Documents');
        $this->loadModel('Categories');
        $document = $this->Documents->newEntity();
        if ($category_id) {
            $this->Documents->patchEntity($document, ['categories' => ['_ids' => [$category_id]]]);
        }

        if ($this->getRequest()->is('post')) {
            $data = [
                'identifiant' => $this->getRequest()->getData('identifiant'),
                'name' => $this->getRequest()->getData('name'),
                'categories' => ['_ids' => $this->getRequest()->getData('categories._ids')],
                'support' => $this->getRequest()->getData('support'),
                'file_format' => $this->getRequest()->getData('file_format'),
                'localisation' => $this->getRequest()->getData('localisation'),
                'dua' => $this->getRequest()->getData('dua'),
                'duc' => $this->getRequest()->getData('duc'),
                'communicabilite' => $this->getRequest()->getData('communicabilite'),
                'depart_calcul_com' => $this->getRequest()->getData('depart_calcul_com'),
                'sort_final' => $this->getRequest()->getData('sort_final'),
                'description' => $this->getRequest()->getData('description'),
                'depart_calcul_dua' => $this->getRequest()->getData('depart_calcul_dua'),
                'depart_calcul_duc' => $this->getRequest()->getData('depart_calcul_duc'),
                'profil_archivage' => $this->getRequest()->getData('profil_archivage'),
                'status' => 'brouillon',
                'note_version' => $this->getRequest()->getData('note_version'),
                'version' => 1,
                'app_meta' => $this->getRequest()->getData('app_meta'),
            ];
            $conn = $this->Documents->getConnection();
            $conn->begin();
            $this->appendFileToData($data, $document);
            $this->Documents->patchEntity($document, $data);
            $this->loadComponent('AsalaeCore.Modal');
            $this->loadModel('Files');
            if ($this->Documents->save($document, ['associated' => 'Categories'])) {
                $this->Files->filesystemCommit();
                $conn->commit();
                $this->Modal->success();
            } else {
                $this->Files->filesystemRollback();
                $conn->rollback();
                $this->Modal->fail();
            }
        }

        $categories = $this->Categories->find('treeList')->toArray();
        $this->set('categories', $categories);
        $this->set('document', $document);
        $this->set('status', $this->Documents->options('status'));
        $this->set('sort_final', $this->Documents->options('sort_final'));
        $this->set('support', $this->Documents->options('support'));
        $this->set('communicabilite', $this->Documents->options('communicabilite'));
        $this->set('comInfo', json_encode(DocumentsTable::getCommunicabiliteDocs()));
    }

    /**
     * Déplace le fichier temporaire dans la base de donnée
     * @param array $data de sauvegarde
     * @param EntityInterface $document
     * @throws Exception
     */
    private function appendFileToData(array &$data, EntityInterface $document)
    {
        $file_upload_profil = $this->getRequest()->getData('file_upload_profil');
        $file = $document->get('file');
        if ($file_upload_profil) {
            $this->loadModel('Files');
            $filename = TMP . urldecode($file_upload_profil);
            if (is_file($filename)) {
                $file = $this->Files->newEntityFromUploadedFile($filename, [
                    'user_id' => $this->Auth->user('id')
                ]);
                $this->Files->save($file);
                $data['schema_file_id'] = $file->get('id');
            } elseif ($file && $file_upload_profil === $file->get('filename')) {
                $data['schema_file_id'] = $file->get('id');
            } else {
                throw new Exception(__("Le fichier {0} n'existe pas", $filename));
            }
        }
    }

    /**
     * Modification d'un document
     * @param string $id
     * @throws \Exception
     */
    public function edit(string $id)
    {
        $uploadId = 'edit-upload-profil';
        $this->set('uploadId', $uploadId);
        $this->set('uploadBoxId', 'edit-upload-document-box');
        $this->loadModel('Documents');
        $this->loadModel('Categories');
        $document = $this->Documents->find()
            ->where(['Documents.id' => $id, 'status <>' => 'actif'])
            ->contain(['Categories', 'Files'])
            ->firstOrFail()
        ;
        $newerExist = $this->Documents->exists([
            'identifiant' => $document->get('identifiant'),
            'version >' => $document->get('version')
        ]);
        if ($newerExist) {
            throw new BadRequestException(__("Impossible de modifier une version autre que la dernière"));
        }
        if ($file = $document->get('file')) {
            $filename = $file->get('filename');
            $this->set('uploadData', [[
                'id' => $filename,
                'name' => $filename,
                'size' => $file->get('size'),
                'select' => '<input type="radio" name="file_upload_profil" value="'.$filename.'" checked="checked">'
            ]]);
        }

        if ($this->getRequest()->is('put')) {
            $data = [
                'id' => $id,
                'identifiant' => ((int)$document->get('version') ?: 1) > 1
                    ? $document->get('identifiant')
                    : $this->getRequest()->getData('identifiant'),
                'name' => $this->getRequest()->getData('name'),
                'categories' => ['_ids' => $this->getRequest()->getData('categories._ids')],
                'support' => $this->getRequest()->getData('support'),
                'file_format' => $this->getRequest()->getData('file_format'),
                'localisation' => $this->getRequest()->getData('localisation'),
                'dua' => $this->getRequest()->getData('dua'),
                'duc' => $this->getRequest()->getData('duc'),
                'communicabilite' => $this->getRequest()->getData('communicabilite'),
                'depart_calcul_com' => $this->getRequest()->getData('depart_calcul_com'),
                'sort_final' => $this->getRequest()->getData('sort_final'),
                'description' => $this->getRequest()->getData('description'),
                'depart_calcul_dua' => $this->getRequest()->getData('depart_calcul_dua'),
                'depart_calcul_duc' => $this->getRequest()->getData('depart_calcul_duc'),
                'profil_archivage' => $this->getRequest()->getData('profil_archivage'),
                'status' => 'brouillon',
                'note_version' => $document->get('note_version'),
                'schema_file_id' => $document->get('schema_file_id'),
                'version' => $document->get('version'),
                'app_meta' => $this->getRequest()->getData('app_meta'),
            ];
            $conn = $this->Documents->getConnection();
            $conn->begin();
            $this->appendFileToData($data, $document);
            $this->Documents->patchEntity($document, $data);
            $this->loadComponent('AsalaeCore.Modal');
            $this->loadModel('Files');
            if ($this->Documents->save($document, ['associated' => 'Categories'])) {
                $this->Files->filesystemCommit();
                $this->Modal->success();
                $conn->commit();
            } else {
                $this->Files->filesystemRollback();
                $this->Modal->fail();
                $conn->rollback();
            }
        }

        $categories = $this->Categories->find('treeList')->toArray();
        $this->set('categories', $categories);
        $this->set('document', $document);
        $this->set('status', $this->Documents->options('status'));
        $this->set('sort_final', $this->Documents->options('sort_final'));
        $this->set('support', $this->Documents->options('support'));
        $this->set('communicabilite', $this->Documents->options('communicabilite'));
        $this->set('comInfo', json_encode(DocumentsTable::getCommunicabiliteDocs()));
    }

    /**
     * Converti $query en csv
     * @param array $results
     * @throws \Aura\Intl\Exception
     */
    private function exportCsv(array $results)
    {
        $data = [];
        $_header = [
            __('Identifiant'),
            __('Nom'),
            __('Format de fichier'),
            __("Durée d'utilité administrative"),
            __("Durée d'utilité courante"),
            __('Sort final'),
            __('Date de départ de calcul de la DUA'),
            __('Date de départ de calcul de la DUC'),
            __("Nom du profil d'archivage"),
            __('Description'),
            __("Statut"),
            __("Version"),
            __('Note sur la version'),
            __('Date de création'),
            __('Date de modification'),
        ];
        foreach ($results as $value) {
            $created = $value['created'];
            if ($created instanceof DateTimeInterface) {
                $created = $created->format('d/m/Y H:i:s');
            }
            $modified = $value['modified'];
            if ($modified instanceof DateTimeInterface) {
                $modified = $modified->format('d/m/Y H:i:s');
            }

            $data[] = [
                $value['identifiant'],
                $value['name'],
                $value['file_format'],
                is_numeric($value['dua']) ? __n("{0} an", "{0} ans", $value['dua'], $value['dua']) : '',
                is_numeric($value['duc']) ? __n("{0} an", "{0} ans", $value['duc'], $value['duc']) : '',
                $value['sort_finaltrad'],
                $value['depart_calcul_dua'],
                $value['depart_calcul_duc'],
                $value['profil_archivage'],
                $value['description'],
                $value['statustrad'],
                $value['version'],
                $value['note_version'],
                $created,
                $modified,
            ];
        }
        $_serialize = 'data';
        $this->set(compact('data', '_serialize', '_header'));
        $this->viewBuilder()->setClassName('CsvView.Csv');
    }

    /**
     * Permet d'exporter la documents et dossiers sous format json, ou
     * d'importer des données via le json obtenu
     */
    public function importExport()
    {
        $uploadId = 'uploader';
        $this->set('uploadId', $uploadId);
    }

    /**
     * @var array Compteurs d'import
     */
    private $importReport = [
        'documents' => [
            'added' => 0,
            'ignored' => 0,
            'skipped' => 0,
            'errors' => 0,
            'updated' => 0,
        ],
        'categories' => [
            'added' => 0,
        ],
    ];

    /**
     * Importe les données précédement exportés
     * @param string $filename
     * @return \Cake\Http\Response|null
     * @throws Exception
     */
    public function importDatabase(string $filename)
    {
        $this->loadComponent('AsalaeCore.Modal');
        /** @var Connection $conn */
        $conn = ConnectionManager::get('default');
        $conn->begin();
        try {
            $user_id = $this->Auth->user('id');
            $path = TMP.'upload'.DS.'user_'.$user_id;
            $finalPath = $path.DS.$filename;
            if (!is_readable($finalPath)) {
                throw new Exception(__("Impossible de lire le fichier ({0})", $finalPath));
            }
            $data = json_decode(file_get_contents($finalPath), true);
            if (empty($data)) {
                throw new Exception(__("Données illisibles. Impossible d'importer la base."));
            }

            $this->log(__("Début de l'import des catégories"), 'debug', ['import-database']);
            $this->importCategories($data['categories']);

            $this->log(__("Début de l'import des documents"), 'debug', ['import-database']);
            $this->importDocuments($data);

            $this->log(__("Fin de l'import"), 'debug', ['import-database']);
            $conn->commit();

            $this->Modal->success();
            return $this->renderJson(json_encode($this->importReport));
        } catch (Exception $e) {
            $this->Modal->fail();
            $conn->rollback();

            $this->log(__("Erreur lors de l'import: {0}\n{1}", $e->getMessage(), $e->getTraceAsString()), 'error', ['import-database']);
            return $this->renderJson(json_encode($e->getMessage()));
        }
    }

    /**
     * Table des correspondances entre les ids de l'import.json, et les entités
     * Rempli par importCategories(), utilisé par importDocuments()
     * @var array [import_id => actual_id]
     */
    private $catIds = [];

    /**
     * Permet l'import des categories d'un json d'export
     * fonction récursive
     * @param array $data
     * @param null|int $parent_id
     * @throws \Aura\Intl\Exception
     */
    private function importCategories(array $data, $parent_id = null)
    {
        $this->loadModel('Categories');
        foreach ($data as $value) {
            $category = $this->Categories->find()
                ->where(['Categories.name' => $value['name']])
                ->first()
            ;
            if (!$category) {
                $this->log('5015 '.__("Création de la catégorie {0}", $value['name']), 'debug', ['context' => 'import-database']);
                $category = $this->Categories->newEntity([
                    'name' => $value['name'],
                    'parent_id' => $parent_id,
                    'type' => $value['type'],
                    'description' => $value['description'],
                ]);
                $this->Categories->saveOrFail($category);
                $this->importReport['categories']['added']++;
            }
            $this->catIds[$value['id']] = $category;
            if (!empty($value['children'])) {
                $this->importCategories($value['children'], $category->get('id'));
            }
        }
    }

    /**
     * Permet l'import des documents d'un json d'export
     * @param array $data
     * @throws \Aura\Intl\Exception
     */
    private function importDocuments(array $data)
    {
        // On tri par date de création pour l'ajout de plusieurs versions
        usort($data['documents'], function ($a, $b) {
            return (new DateTime($a['created'])) > (new DateTime($b['created']));
        });
        $this->loadModel('Documents');
        $lasts = [];
        foreach ($data['documents'] as $value) {
            $exist = $this->Documents->exists(['uid' => $value['uid']]);
            if ($exist) {
                $this->log('5014 '.__("Le document {0} existe déjà", $value['uid']), 'debug', ['context' => 'import-database']);
                $this->importReport['documents']['ignored']++;
                continue;
            }
            if (!isset($lasts[$value['identifiant']])) {
                $last = $this->Documents->find()
                    ->where(['Documents.identifiant' => $value['identifiant']])
                    ->order(['Documents.version' => 'desc'])
                    ->contain(['Categories'])
                    ->first()
                ;
                if ($last) {
                    $lasts[$value['identifiant']] = $last;
                }
            } else {
                $last = $lasts[$value['identifiant']];
            }
            if ($last) {
                if ($last->get('created') > (new DateTime($value['created']))) {
                    $this->importReport['documents']['skipped']++;
                    $this->log('5001 '.__("Le document {0} n'a pas été importé car sa date de création précède celle de la dernière version.", $value['uid']), 'notice', ['context' => 'import-database']);
                    continue;
                }
                $nData = $value;
                unset($nData['id'], $nData['version'], $nData['schema_file_id']);
                $dirty = false;
                if ($last->get('status') === 'actif') {
                    $actualVersion = $last->get('version');
                    $nData['version'] = $actualVersion +1;
                    $newDocument = clone $last;
                    $newDocument->unsetProperty(['id', 'created', 'modified', 'uid']);
                    $newDocument->isNew(true);
                    $newDocument->set($nData);
                    if ($err = $newDocument->getErrors()) {
                        $errors = $this->flatErrors($err);
                        $this->importReport['documents']['errors']++;
                        $this->log('5002 '.__("Lors de la montée de version du document {0}, une erreur a été détecté : {1}.", $value['uid'], $errors), 'error', ['context' => 'import-database']);
                        continue;
                    }
                } else {
                    $newDocument = $last;
                    $newDocument->set($nData);
                    $dirty = $newDocument->isDirty();
                    $actualVersion = 0;
                }
                if (!empty($nData['file'])) {
                    unset($nData['file']['id']);
                    $nData['file']['user_id'] = $this->getRequest()->getSession()->read('User.id');
                    $newDocument->set('file', $nData['file']);
                }
                $newDocument->set('categories', $this->appendCategories($data['categories_documents'], $value['id']));
                $success = $this->Documents->save($newDocument, ['associated' => ['Categories', 'Files']]);
                if (!$success) {
                    $errors = $this->flatErrors($last->getErrors());
                    $this->importReport['documents']['errors']++;
                    $this->log('5003 '.__("Lors de la montée de version du document {0}, une erreur a été détecté : {1}.", $value['uid'], $errors), 'error', ['context' => 'import-database']);
                } elseif ($actualVersion < $newDocument->get('version')) {
                    $this->importReport['documents']['added']++;
                    $this->log('5004 '.__("Montée de version du document {0}", $value['uid']), 'debug', ['context' => 'import-database']);
                } elseif ($dirty) {
                    $this->importReport['documents']['updated']++;
                    $this->log('5005 '.__("Mise à jour du document {0} sans montée de version.", $value['uid']), 'debug', ['context' => 'import-database']);
                } else {
                    $this->importReport['documents']['skipped']++;
                    $this->log('5006 '.__("Le document {0} n'a pas été modifié.", $value['uid']), 'warning', ['context' => 'import-database']);
                }
            } else {
                $nData = $value;
                unset($nData['id'], $nData['categories'], $nData['schema_file_id']);
                if (!empty($nData['file'])) {
                    unset($nData['file']['id']);
                    $nData['file']['user_id'] = $this->getRequest()->getSession()->read('User.id');
                }
                $entity = $this->Documents->newEntity($nData);
                if ($err = $entity->getErrors()) {
                    $this->importReport['documents']['errors']++;
                    $errors = $this->flatErrors($err);
                    $this->log('5011 '.__("Lors de la création du document {0}, une erreur a été détecté : {1}.", $value['uid'], $errors), 'error', ['context' => 'import-database']);
                    continue;
                }
                $entity->set('categories', $this->appendCategories($data['categories_documents'], $value['id']));

                if ($this->Documents->save($entity, ['associated' => ['Categories', 'Files']])) {
                    $this->importReport['documents']['added']++;
                    $this->log('5012 '.__("Création du document {0}.", $value['uid']), 'debug', ['context' => 'import-database']);
                } else {
                    $this->importReport['documents']['errors']++;
                    $errors = $this->flatErrors($entity->getErrors());
                    $this->log('5013 '.__("Lors de la création du document {0}, une erreur a été détecté : {1}.", $value['uid'], $errors), 'error', ['context' => 'import-database']);
                }
            }
        }
    }

    /**
     * Permet d'ajouter les categories dans une entité
     * Fait la correspondance entre les ids présent dans le fichier d'import
     * avec ceux en base de donnée
     * @param array $categoriesDocuments
     * @param int $importId
     * @return array
     * @throws \Aura\Intl\Exception
     */
    private function appendCategories(array $categoriesDocuments, int $importId): array
    {
        $filtered = array_filter($categoriesDocuments, function ($v) use ($importId) {
            return $v['document_id'] === $importId;
        });
        $this->loadModel('CategoriesDocuments');
        $data = [];
        foreach ($filtered as $value) {
            if (isset($this->catIds[$value['category_id']])) {
                $data[] = $this->catIds[$value['category_id']];
            } else {
                $this->log(__("Une catégorie liée à un document n'existe pas : {0}.", var_export($value, true)), 'error', ['context' => 'import-database']);
            }
        }
        return $data;
    }

    /**
     * Converti un array d'erreur d'une entité en chaine
     * @param array $errors
     * @return string
     */
    private function flatErrors(array $errors): string
    {
        $flat = Hash::flatten($errors);
        $eStrFunc = function ($acc, $k) use ($flat) {
            return $acc.$k.": ".$flat[$k].", ";
        };
        return rtrim(array_reduce(array_keys($flat), $eStrFunc), ', ');
    }

    /**
     * Exporte les données en base
     * @param null|Query $query
     * @return \Cake\Http\Response|null
     * @throws Exception
     */
    public function exportDatabase($query = null)
    {
        $this->loadModel('Categories');
        $this->loadModel('Documents');
        $this->loadModel('CategoriesDocuments');
        $this->loadComponent('AsalaeCore.Condition');

        if ($query) {
            $search = $query->all()->toArray();
            $categoriesIds = array_unique(
                array_reduce(
                    array_map(function ($v) {
                        return array_map(function ($v) {
                            return $v['id'];
                        }, $v['categories'] ?? []);
                    }, $search),
                    'array_merge',
                    []
                )
            );
            $categories = $this->Categories->find('threaded')->all()->toArray();
            $this->removeUnwanted($categories, $categoriesIds);
            $categories_documents = $this->CategoriesDocuments->find()
                ->where($this->Condition->in('CategoriesDocuments.category_id', $this->getFlattenIds($categories)))
                ->all()
                ->toArray()
            ;
            $output = [
                'categories' => $categories,
                'categories_documents' => $categories_documents,
                'documents' => $search,
            ];
        } else {
            $output = [
                'categories' => $this->Categories->find('threaded')->all()->toArray(),
                'categories_documents' => $this->CategoriesDocuments->find()->all()->toArray(),
                'documents' => $this->Documents->find()->contain(['Files'])->all()->toArray(),
            ];
        }
        $datetime = date('Y-m-d H:i:s');
        $this->setResponse($this->getResponse()->withDownload('export-database-'.$datetime.'.json'));
        return $this->renderJson(json_encode($output));
    }

    /**
     * Supprime les catégories qui ne contiennent pas un id dans $keepIds
     * @param array $data
     * @param array $keepIds
     * @return bool true si dans la liste
     */
    private function removeUnwanted(array &$data, array $keepIds): bool
    {
        $keeped = false;
        foreach ($data as $key => $values) {
            $keepedChildren = false;
            if ($values['children']) {
                $keepedChildren = $this->removeUnwanted($data[$key]['children'], $keepIds);
            }
            if (in_array($values['id'], $keepIds) || $keepedChildren) {
                $keeped = true;
            } else {
                unset($data[$key]);
            }
        }
        return $keeped;
    }

    /**
     * Donne les ids d'un data threaded
     * @param array $data
     * @return array
     */
    private function getFlattenIds(array $data): array
    {
        $results = [];
        foreach ($data as $key => $values) {
            $results[] = $values['id'];
            if ($values['children']) {
                $results = array_merge($results, $this->getFlattenIds($values['children']));
            }
        }
        return $results;
    }

    /**
     * Action de publication
     * @param string $id
     * @return \Cake\Http\Response|null
     * @throws Exception
     */
    public function publish(string $id)
    {
        $this->getRequest()->is('update');
        $this->loadModel('Documents');
        $document = $this->Documents->get($id);
        $document->set('status', 'actif');
        $document->set('public_date', new DateTime);
        $this->loadComponent('AsalaeCore.Modal');
        if ($this->Documents->save($document)) {
            $this->Modal->success();
            return $this->renderJson($this->addSearchData());
        }
        $this->Modal->fail();
        return $this->renderJson('"failed"');
    }

    /**
     * Action de visualisation des données d'une règle de gestion
     * @param string $id
     */
    public function view(string $id)
    {
        $this->loadModel('Documents');
        $SubDocuments = clone $this->Documents;
        $SubDocuments->setAlias('doc');
        $subquery = $SubDocuments->find()
            ->select(['doc.identifiant'])
            ->where(['doc.id' => $id])
            ->limit(1)
        ;
        $query = $this->Documents->find()
            ->where(['Documents.identifiant' => $subquery])
            ->contain(['Categories' => ['sort' => ['name' => 'asc']], 'Files'])
            ->order(['Documents.version' => 'asc'])
        ;

        if (!$this->Auth->user('id')) {
            $query->where(['status' => 'actif']);
        }

        $documents = $query->all()->toArray();
        $this->set('documents', $documents);
    }

    /**
     * Action de création de la nouvelle version
     * @param string $id
     * @return \Cake\Http\Response|null
     * @throws Exception
     */
    public function newVersion(string $id)
    {
        $this->loadModel('Documents');
        $document = $this->Documents->find()
            ->where(['Documents.id' => $id])
            ->contain(['Categories'])
            ->first()
        ;
        $document->unsetProperty('note_version');
        $this->set('document', $document);

        $newVersionIsPossible = !$this->Documents->exists([
            'identifiant' => $document->get('identifiant'),
            'version >' => $document->get('version'),
        ]);
        $this->set('newVersionIsPossible', $newVersionIsPossible);

        if ($newVersionIsPossible && $this->getRequest()->is('put')) {
            $data = [
                'status' => 'brouillon',
                'note_version' => $this->getRequest()->getData('note_version'),
                'version' => $document->get('version') +1
            ];
            $conn = $this->Documents->getConnection();
            $conn->begin();

            if ($file_id = $document->get('schema_file_id')) {
                $this->loadModel('Files');
                $file = $this->Files->get($file_id);
                $file->unsetProperty(['id', 'created', 'modified']);
                $file->isNew(true);
                $this->Files->saveOrFail($file);
                $data['schema_file_id'] = $file->get('id');
            }

            $newDocument = clone $document;
            $newDocument->unsetProperty(['id', 'created', 'modified', 'uid']);
            $newDocument->isNew(true);
            $newDocument->set($data);

            $this->loadComponent('AsalaeCore.Modal');
            if ($this->Documents->save($newDocument, ['associated' => 'Categories'])) {
                $this->Modal->success();
                $conn->commit();
                return $this->renderJson(json_encode($newDocument->toArray()));
            } else {
                $this->Modal->fail();
                $conn->rollback();
            }
        }
    }

    /**
     * Action d'obsolescence
     * @param string $id
     * @return \Cake\Http\Response|null
     * @throws Exception
     */
    public function obsolete(string $id)
    {
        $this->getRequest()->is('update');
        $this->loadModel('Documents');
        $document = $this->Documents->get($id);
        $this->loadComponent('AsalaeCore.Modal');
        if ($this->Documents->updateAll(
            ['status' => 'obsolete', 'obsolete_date' => new DateTime],
            ['identifiant' => $document->get('identifiant')]
        )) {
            $this->Modal->success();
            return $this->renderJson($this->addSearchData());
        }
        $this->Modal->fail();
        return $this->renderJson('"failed"');
    }

    /**
     * Permet de récupérer une règle avec son identifiant
     * @param string $identifiant
     * @return \Cake\Http\Response
     */
    public function find(string $identifiant)
    {
        $this->loadModel('Documents');
        /** @var Query $query */
        $query = $this->Documents->find()
            ->where([
                'Documents.identifiant' => $identifiant,
                'Documents.status' => 'actif',
            ])
            ->contain(['Categories', 'Files'])
            ->order(['Documents.version' => 'desc'])
        ;
        $result = $query->firstOrFail();
        return $this->renderData($result);
    }

    /**
     * Action de suppression pour une régle brouillon
     * @param string $id
     * @return \Cake\Http\Response|null
     * @throws Exception
     */
    public function delete(string $id)
    {
        $this->getRequest()->allowMethod('delete');
        $this->loadModel('Documents');
        $document = $this->Documents->find()
            ->where([
                'Documents.id' => $id,
                'Documents.status' => 'brouillon'
            ])
            ->firstOrFail()
        ;
        $this->loadComponent('AsalaeCore.Modal');
        if ($this->Documents->delete($document)) {
            $this->Modal->success();
            return $this->renderJson('"done"');
        }
        $this->Modal->fail();
        return $this->renderJson('"failed"');
    }
}
