<?php
/**
 * Refae\Model\Table\DocumentsTable
 */

namespace Refae\Model\Table;

use ArrayObject;
use AsalaeCore\Model\Behavior\AliasBehavior;
use Cake\Datasource\EntityInterface;
use Cake\Event\Event;
use Cake\ORM\Table;
use Cake\Validation\Validator;
use AsalaeCore\Model\Behavior\OptionsBehavior;
use Exception;

/**
 * Table documents
 *
 * @category Table
 *
 * @author    Ludovic Ganée <ludovic.ganee@libriciel.coop>
 * @copyright (c) 2017-2018, Libriciel SCOP
 * @license   https://opensource.org/licenses/AGPL-3.0
 * @mixin OptionsBehavior
 * @mixin AliasBehavior
 */
class DocumentsTable extends Table
{
    /**
     * Configuration initale de la table
     *
     * @param array $config
     * @throws Exception
     */
    public function initialize(array $config)
    {
        $this->addBehavior('Timestamp');
        $this->addBehavior('AsalaeCore.Alias');
        $this->addBehavior('AsalaeCore.Options', [
            'status' => $this->enum('status'),
            'sort_final' => $this->enum('sort_final'),
            'support' => $this->enum('support'),
            'communicabilite' => $this->enum('communicabilite'),
        ]);
        $this->belongsToMany('Categories');
        $this->belongsTo('Files', [
            'foreignKey' => 'schema_file_id',
            'dependent' => true,
        ]);

        parent::initialize($config);
    }

    /**
     * Default validation rules.
     *
     * @param Validator $validator Validator instance.
     * @return Validator
     * @throws Exception
     */
    public function validationDefault(Validator $validator): Validator
    {
        $validator
            ->notEmpty('name')
            ->notEmpty('support')
            ->notEmpty('dua')
            ->notEmpty('communicabilite')
            ->notEmpty('sort_final')
            ->notEmpty('identifiant')
            ->notEmpty('version')
            ->notEmpty('status')
            ->notEmpty('uid')
            ->allowEmpty('file_format')
            ->allowEmpty('localisation')
            ->allowEmpty('description')
            ->allowEmpty('note_version')
            ->allowEmpty('duc')
            ->allowEmpty('depart_calcul_dua')
            ->allowEmpty('depart_calcul_duc')
            ->allowEmpty('depart_calcul_com')
            ->allowEmpty('profil_archivage')
            ->allowEmpty('schema_file_id')
        ;

        $validator->add('identifiant', ['unique' => [
            'rule' => ['validateUnique', ['scope' => 'version']],
            'provider' => 'table',
            'message' => __("Cet identifiant est déjà utilisé")]
        ]);
        $validator->add('status', ['inList' => [
            'rule' => ['inList', $this->enum('status')],
        ]]);

        return $validator;
    }

    /**
     * Fired before each entity is saved. Stopping this event will abort the save
     * operation. When the event is stopped the result of the event will be returned.
     * @param Event $event
     * @param EntityInterface $entity
     * @param ArrayObject $options
     * @return bool
     */
    public function beforeSave(/** @noinspection PhpUnusedParameterInspection */
        Event $event,
        EntityInterface $entity,
        ArrayObject $options
    ) {
        if ($entity->isNew() && empty($entity->get('uid'))) {
            $entity->set('uid', uniqid('', true));
        }
        return true;
    }

    /**
     * Renvoi la liste des choix possible pour un champ
     * @param string $field
     * @return array
     * @throws Exception
     */
    public function enum(string $field)
    {
        switch ($field) {
            case 'status':
                return ['actif', 'brouillon', 'obsolete'];
            case 'sort_final':
                return ['D', 'C', 'T'];
            case 'support':
                return ['E', 'P', 'P et E'];
            case 'communicabilite':
                return array_keys(self::getCommunicabiliteOptions());
            default:
                throw new Exception(__("{0} n'est pas un champ de type enum", $field));
        }
    }

    /**
     * @var array Liste des options de select ['AR038' => '0 an', ...]
     */
    private static $communicabiliteOptions = [];

    /**
     * @var array Liste la documentation liée ['AR038' => 'Documents...', ...]
     */
    private static $communicabiliteDocs = [];

    /**
     * Liste les options du champ communicabilite
     * Est construit à partir du fichier situé dans
     *                  src/Data/accessrestrictioncodes.json
     * @return array ['AR038' => '0 an', ...]
     * @throws Exception
     */
    public static function getCommunicabiliteOptions(): array
    {
        if (empty(self::$communicabiliteOptions)) {
            $filename = APP . 'Data' . DS . 'accessrestrictioncodes.json';
            if (!is_readable($filename)) {
                throw new Exception(__("Impossible de lire le fichier {0}", $filename));
            }
            $data = json_decode(file_get_contents($filename), true);
            foreach ($data as $values) {
                self::$communicabiliteOptions[$values['code']] = $values['code'].': '.$values['name'];
                self::$communicabiliteDocs[$values['code']] = $values['description'];
            }
        }
        return self::$communicabiliteOptions;
    }

    /**
     * Liste la documentation liée au code de communicabilite
     * @return array
     * @throws Exception
     */
    public static function getCommunicabiliteDocs(): array
    {
        self::getCommunicabiliteOptions();
        return self::$communicabiliteDocs;
    }
}
