<?php
/**
 * Refae\Model\Table\FilesTable
 */

namespace Refae\Model\Table;

use ArrayObject;
use Cake\Event\Event;
use Cake\ORM\Entity;
use Cake\ORM\Table;
use Cake\Validation\Validator;
use Libriciel\Filesystem\Utility\Filesystem;
use Refae\Model\Entity\File;

/**
 * Table files
 *
 * @category Table
 *
 * @author    Ludovic Ganée <ludovic.ganee@libriciel.coop>
 * @copyright (c) 2017-2018, Libriciel SCOP
 * @license   https://opensource.org/licenses/AGPL-3.0
 */
class FilesTable extends Table
{
    /**
     * Configuration initale de la table
     *
     * @param array $config
     */
    public function initialize(array $config)
    {
        $this->addBehavior('Timestamp');
        $this->belongsTo('Users');
        $this->hasOne('Documents', ['foreignKey' => 'schema_file_id']);

        parent::initialize($config);
    }

    /**
     * Default validation rules.
     *
     * @param Validator $validator Validator instance.
     * @return Validator
     */
    public function validationDefault(Validator $validator): Validator
    {
        return $validator;
    }

    /**
     * Confirme la suppression d'un fichier utilisé pour la création
     *
     * @param Event $event
     * @param Entity $entity
     * @param ArrayObject $options
     */
    public function afterSave(/** @noinspection PhpUnusedParameterInspection */
        Event $event,
        Entity $entity,
        ArrayObject $options
    ) {
        if ($entity->isNew() && $path = $entity->get('path')) {
            if ($this->getConnection()->inTransaction()) {
                $this->fileNamespaceToCommit[] = $path;
            } else {
                Filesystem::commit($path);
            }
        }
    }

    /**
     * Lors d'une sauvegarde, place les namespace ici en cas de transaction
     * Il faudra utiliser la méthode filesystemCommit()
     * @var array
     */
    private $fileNamespaceToCommit = [];

    /**
     * Commit du Filesystem après une transaction
     * @return $this
     */
    public function filesystemCommit()
    {
        foreach ($this->fileNamespaceToCommit as $namespace) {
            Filesystem::commit($namespace);
        }
        $this->fileNamespaceToCommit = [];
        return $this;
    }

    /**
     * Rollback du Filesystem après une transaction
     * @return $this
     */
    public function filesystemRollback()
    {
        foreach ($this->fileNamespaceToCommit as $namespace) {
            Filesystem::rollback($namespace);
        }
        $this->fileNamespaceToCommit = [];
        return $this;
    }

    /**
     * Crée une entité à partir d'un chemin de fichier
     * Provoque la suppression du fichier lors de l'afterSave. Il est possible
     * d'empécher la suppression avec un <b>Filesystem::rollback($filename);</b>
     * après l'appel de cette fonction
     * @param string $filename
     * @param null|array $data
     * @param array $options
     * @return Entity
     */
    public function newEntityFromUploadedFile(string $filename, $data = null, array $options = []): Entity
    {
        $data = (array)$data + [
            'filename' => basename($filename),
            'size' => filesize($filename),
            'content' => file_get_contents($filename),
            'path' => $filename
        ];
        $entity = $this->newEntity($data, $options);
        Filesystem::begin($filename);
        Filesystem::remove($filename);
        return $entity;
    }

    /**
     * Génère un fichier à partir des données d'une entité
     * @param File $file
     * @param string $targetDir
     * @param null $filename
     * @return bool
     */
    public function buildFile(File $file, string $targetDir, $filename = null)
    {
        if ($filename === null) {
            $filename = $file->get('filename');
        }
        Filesystem::dumpFile($targetDir . DS . $filename, $file->get('content'));
        return is_file($targetDir . DS . $filename)
            ? $targetDir . DS . $filename
            : false
        ;
    }
}
