<?php
/**
 * AsalaeCore\Controller\ApiTrait
 */

namespace AsalaeCore\Controller;

use AsalaeCore\Controller\Component\RestComponent;
use AsalaeCore\Http\AppServerRequest;
use Cake\Controller\Component\AuthComponent;
use Cake\Datasource\ResultSetInterface;
use Cake\Http\Response;
use Cake\Http\Exception\BadRequestException;
use Cake\Http\Exception\ForbiddenException;
use Cake\Http\Exception\MethodNotAllowedException;
use Cake\Http\Exception\NotAcceptableException;
use Cake\Http\Exception\NotFoundException;
use Cake\ORM\Entity;
use Cake\ORM\Table;

/**
 * Trait d'un controller api
 *
 * @category Controller
 *
 * @author    Ludovic Ganée <ludovic.ganee@libriciel.coop>
 * @copyright (c) 2017, Libriciel
 * @license   http://www.cecill.info/licences/Licence_CeCILL_V2-fr.html
 * @property AppServerRequest $request
 * @property Response $response
 * @property RestComponent $Rest
 * @property AuthComponent $Auth
 */
trait ApiTrait
{
    /**
     * Traits
     */
    use RenderDataTrait;

    /**
     * Permet d'accèder à l'api rest du controller
     * @param string $id
     * @return Response|null
     * @throws \Exception
     */
    public function api(string $id = '')
    {
        $this->loadComponent('AsalaeCore.Rest');
        $method = strtolower($this->request->getEnv('REQUEST_METHOD'));
        $action = $id ?: 'index';
        if (is_numeric($action)) {
            $action = 'index';
        }
        if ($method !== 'get' || !in_array($action, $this->Auth->allowedActions)) {
            $this->Rest->auth();
        }
        if ($route = $this->Rest->route()) {
            return call_user_func_array($route['callback'], $route['args']);
        }

        // Vérifications d'usage
        $accepts = $this->apiCheck($method, $id);

        // Récupération des données
        $controller = $this->request->getParam('controller');
        /** @var Table $Model */
        $Model = $this->loadModel($controller);
        $results = [];
        if ($this->request->is(['get', 'put', 'delete'])) {
            $query = $Model->find();
            if ($id) {
                $query->where(['id' => $id]);
                $results = $query->first();
            } else {
                $results = $query->all();
            }
        }
        $this->apiCheckData($method, $results, $id);

        // Actions et résultats
        $functionName = 'api'.ucfirst($method).'Action';
        $output = $this->$functionName($Model, $results, $id);

        // Rendu
        return $this->renderByAccepts($accepts, $output);
    }

    /**
     * Vérifications du formatage de la requête
     * @param string $method
     * @param string $id
     * @return array
     * @throws MethodNotAllowedException
     * @throws BadRequestException
     * @throws NotAcceptableException
     * @throws \Aura\Intl\Exception
     */
    protected function apiCheck(string $method, string $id): array
    {
        if (!in_array($method, ['get', 'put', 'post', 'delete'])) {
            throw new MethodNotAllowedException(__("405 Method Not Allowed"));
        }
        if ((in_array($method, ['put', 'delete']) && empty($id))
            || ($method === 'post' && !empty($id))
        ) {
            throw new BadRequestException(__("400 Bad Request"));
        }
        $acceptable = ['json', 'xml', 'html', 'xhtml'];
        $accepts = current($this->response->mapType($this->request->parseAccept()));
        $intersect = array_intersect($accepts, $acceptable);
        if (!$intersect) {
            throw new NotAcceptableException(__("406 Not Acceptable"));
        }
        return $intersect;
    }

    /**
     * Vérifications lié à la présence de data
     * @param string $method
     * @param array|ResultSetInterface|Entity $results
     * @param string $id
     * @throws NotFoundException
     * @throws BadRequestException
     * @throws \Aura\Intl\Exception
     */
    protected function apiCheckData(string $method, $results, $id)
    {
        if (empty($results)
            && (
                in_array($method, ['put', 'delete'])
                || ($method === 'get' && !empty($id))
            )
        ) {
            throw new NotFoundException(__("404 Not Found"));
        }
        if (empty($this->request->getData()) && in_array($method, ['put', 'post'])) {
            throw new BadRequestException(__("400 Bad Request"));
        }
    }

    /**
     * Action get
     * @param Table $Model
     * @param array|ResultSetInterface|Entity $results
     * @param string $id
     * @return array
     */
    protected function apiGetAction(
        /** @noinspection PhpUnusedParameterInspection */
        Table $Model,
        $results,
        $id = ''
    ): array {
        if (empty($id)) {
            $output = array_map(function (Entity $a) {
                return $a->toArray();
            }, $results->toArray());
        } else {
            $output = !empty($results) ? $results->toArray() : [];
        }
        return $output;
    }

    /**
     * Action put
     * @param Table $Model
     * @param array|ResultSetInterface|Entity $results
     * @param string $id
     * @return array
     */
    protected function apiPutAction(
        /** @noinspection PhpUnusedParameterInspection */
        Table $Model,
        $results,
        $id = ''
    ): array {
        return $this->apiSave($Model, $results);
    }

    /**
     * Logique de sauvegarde commune entre put et post
     * @param Table $Model
     * @param Entity $entity
     * @return array
     */
    protected function apiSave(Table $Model, Entity $entity): array
    {
        $Model->patchEntity($entity, $this->request->getData());
        if ($Model->save($entity)) {
            $output['success'] = true;
        } else {
            $output['success'] = false;
            $output['errors'] = $entity->getErrors();
        }
        $output['id'] = $entity->get('id');
        return $output;
    }

    /**
     * Action post
     * @param Table $Model
     * @param array|ResultSetInterface|Entity $results
     * @param string $id
     * @return array
     */
    protected function apiPostAction(
        /** @noinspection PhpUnusedParameterInspection */
        Table $Model,
        $results,
        $id = ''
    ): array {
        /* @var Entity $entity */
        $entity = $Model->newEntity();
        return $this->apiSave($Model, $entity);
    }

    /**
     * Action delete
     * @param Table $Model
     * @param array|ResultSetInterface|Entity $results
     * @param string $id
     * @return array
     */
    protected function apiDeleteAction(
        /** @noinspection PhpUnusedParameterInspection */
        Table $Model,
        $results,
        $id = ''
    ): array {
        $output['id'] = $results->get('id');
        if ($Model->delete($results)) {
            $output['success'] = true;
        } else {
            $output['success'] = false;
            $output['errors'] = $results->getErrors();
        }
        return $output;
    }

    /**
     * Définit les actions possibles via l'api
     *
     * exemple :
     * Le controller Users permet d'obtenir les infos d'un utilisateur et
     * possède également une méthode pour réinitialiser son mot de passe :
     *  $this->setApiActions([
     *      'default' => [          // pour https://mon.url/users[/:id]
     *          'get' => [
     *              'params' => ['id' => ['numeric', 'optional']],
     *              'function' => 'apiGet',
     *          ]
     *      ],
     *      'resetPassword' => [    // pour https://mon.url/users/reset-password/15
     *          'put' => [
     *              'function' => 'resetPassword'
     *          ]
     *      ]
     *  ])
     *
     * @param array $actions
     */
    protected function setApiActions(array $actions)
    {
        $this->Rest->setApiActions($actions);
    }

    /** @noinspection PhpUnusedPrivateMethodInspection */
    /**
     * Donne un callback pour interdire une action
     * Utile pour les fonctions non implémentées pour éviter d'accéder à des
     * informations confidentielles (ex: mot de passe), en effet, par défaut, un
     * get permet d'obtenir tous les champs d'une table si les droits le
     * permettent
     *
     * ex:
     *  $this->setApiActions([
     *      'default' => ['*' => ['function' => 'forbidden']]
     * ])
     */
    private function forbidden()
    {
        throw new ForbiddenException(__("403 Forbidden"));
    }
}
