<?php
/**
 * AsalaeCore\Controller\Component\ConditionComponent
 */

namespace AsalaeCore\Controller\Component;

use Cake\Controller\Component;
use Cake\Controller\ComponentRegistry;
use Cake\Database\Connection;
use Cake\Database\Driver;
use Cake\Datasource\ConnectionManager;
use Cake\I18n\Date;
use Cake\Http\Exception\BadRequestException;
use Cake\Utility\Inflector;
use DateTime;
use PDO;

/**
 * Component d'un controller utilisant les filtres de recherche
 *
 * @category Controller
 *
 * @author    Ludovic Ganée <ludovic.ganee@libriciel.coop>
 * @copyright (c) 2017, 2018 Libriciel
 * @license   http://www.cecill.info/licences/Licence_CeCILL_V2-fr.html
 */
class ConditionComponent extends Component
{
    /**
     * @var \Cake\Datasource\ConnectionInterface Classe de connexion
     */
    public $connection;

    /**
     * @var string driver name (ex: mysql, postgres etc...)
     */
    public $driverName = '';

    /**
     * ConditionComponent constructor.
     * @param ComponentRegistry $registry
     * @param array $config
     */
    public function __construct(ComponentRegistry $registry, array $config = [])
    {
        parent::__construct($registry, $config);
        $this->connection = ConnectionManager::get('default');
        if ($this->connection instanceof Connection) {
            /** @var Driver $driver */
            $driver = $this->connection->getDriver();
            if (!$driver->isConnected()) {
                $driver->connect();
            }
            /** @var PDO $pdo */
            $pdo = $driver->getConnection();
            $this->driverName = $pdo->getAttribute(PDO::ATTR_DRIVER_NAME);
        }
    }

    /**
     * Condition pour un $field ilike $value
     * @param string $field
     * @param string $value
     * @return array
     */
    public function ilike(string $field, string $value)
    {
        $value = preg_replace('/\?/', '_', $value);
        $value = preg_replace('/\*/', '%', $value);
        if ($this->driverName === 'mysql') {
            return [$field.' LIKE' => $value];
        }
        return [$field.' ILIKE' => $value];
    }

    /**
     * Conditions pour un $field in $values
     * @param string $field
     * @param mixed $values
     * @return array
     */
    public function in(string $field, $values)
    {
        if (empty($values)) {
            return ['FALSE'];
        }
        return [$field.' IN' => $values];
    }

    /**
     * Permet d'obtenir la condition pour $fieldBegin <= $date <= $fieldEnd
     * @param string|DateTime $date
     * @param string $fieldBegin
     * @param string $fieldEnd
     * @param string $format Format de $date (default= 'auto')
     * @return array
     */
    public function dateBetweenFields($date, string $fieldBegin, string $fieldEnd, string $format = 'auto')
    {
        if (is_string($date)) {
            $date = $format === 'auto'
                ? Date::parseDateTime($date)->format('Y-m-d')
                : DateTime::createFromFormat($format, $date)->format('Y-m-d')
            ;
        } else {
            $date = $date->format('Y-m-d');
        }
        return [
            "date($fieldBegin) <=" => $date,
            "date($fieldEnd) >=" => $date,
        ];
    }

    /**
     * Condition $field $operator $date
     * @param string $field
     * @param string $operator
     * @param $date
     * @param string $format
     * @return array
     * @throws \Aura\Intl\Exception
     */
    public function dateOperator(string $field, string $operator, $date, string $format = 'auto')
    {
        if (!in_array($operator, ['=', '<=', '>=', '<', '>'])) {
            throw new BadRequestException(__("L'opérateur indiqué n'est pas autorisé"));
        }
        if (is_string($date)) {
            $date = $format === 'auto'
                ? Date::parseDateTime($date)->format('Y-m-d')
                : DateTime::createFromFormat($format, $date)->format('Y-m-d')
            ;
        } elseif (method_exists($date, 'format')) {
            $date = $date->format('Y-m-d');
        }
        if ($date[0] === '-' || substr($date, 0, 4) === '0000') {
            throw new BadRequestException(__("La date saisie ne peut pas être aussi lointaine"));
        }
        return ["date($field) ".$operator => $date];
    }

    /**
     * Condition $field IN favorites
     * @param string $field
     * @param string $table Si défini sur 'auto', essayera de trouver le nom
     *                      par rapport à $field, sinon par le nom du Controller
     * @return array
     */
    public function inFavorites(string $field, string $table = 'auto')
    {
        if ($table === 'auto') {
            if (strpos($field, '.')) {
                $table = explode('.', $field, 2)[0];
            } else {
                $table = $this->getController()->getName();
            }
        }
        $table = Inflector::underscore($table);
        return $this->in($field, $this->getFavoritesCookies($table));
    }

    /**
     * Order by favorites $direction
     * @param string $direction
     * @param string $table
     * @return array
     */
    public function orderByFavorites(string $direction = 'asc', string $table = 'auto')
    {
        if ($table === 'auto') {
            $table = $this->getController()->getName();
        }
        $table = Inflector::underscore($table);
        $favoritesIds = $this->getFavoritesCookies($table);
        if (empty($favoritesIds)) {
            return ['FALSE' => $direction];
        }
        return [Inflector::camelize($table).'.id NOT IN ('.implode(', ', $favoritesIds).')' => $direction];
    }

    /**
     * Permet d'obtenir les ids des favories d'une table
     * @param string $table
     * @return array
     */
    private function getFavoritesCookies(string $table)
    {
        $cookies = $this->getController()->getRequest()->getCookieParams();
        $favoritesIds = [];
        foreach (array_keys($cookies) as $cookieName) {
            if (preg_match("/^table-favorite-$table-table-([\d]+)$/", $cookieName, $match)) {
                $favoritesIds[] = $match[1];
            }
        }
        return $favoritesIds;
    }
}
