<?php
/**
 * AsalaeCore\Controller\Component\FilterComponent
 */

namespace AsalaeCore\Controller\Component;

use AsalaeCore\Utility\Notify;
use Cake\Controller\Component;
use Cake\ORM\Entity;
use Cake\ORM\TableRegistry;

/**
 * Component d'un controller utilisant les filtres de recherche
 *
 * @category Controller
 *
 * @author    Ludovic Ganée <ludovic.ganee@libriciel.coop>
 * @copyright (c) 2017, Libriciel
 * @license   http://www.cecill.info/licences/Licence_CeCILL_V2-fr.html
 */
class FilterComponent extends Component
{

    /**
     * Permet de charger dans la vue les filtres sauvegardés dans
     * le contexte actuel sans la variable <b>$savedFilters<b>
     */
    public function loadAndSave()
    {
        $SavedFilters = TableRegistry::getTableLocator()->get('SavedFilters');

        $this->saveFilters();

        $result = $SavedFilters->find()
            ->where([
                'user_id' => $this->getController()->Auth->user('id'),
                'controller' => $this->getController()->getRequest()->getParam('controller'),
                'action' => $this->getController()->getRequest()->getParam('action'),
            ])
            ->contain(['Filters'])
            ->all()
            ->toArray()
        ;
        $this->getController()->set('savedFilters', $result);
    }


    /**
     * Permet d'obtenir la liste des filtres dans l'url
     *
     * @return void|array
     * @throws \ZMQSocketException
     * @throws \Aura\Intl\Exception
     */
    private function extractFilters()
    {
        $SavedFilters = TableRegistry::getTableLocator()->get('SavedFilters');
        $params = $this->getController()->getRequest()->getParam('?');
        $result = $SavedFilters->find('list')
            ->where([
                'user_id' => $this->getController()->Auth->user('id'),
                'controller' => $this->getController()->getRequest()->getParam('controller'),
                'action' => $this->getController()->getRequest()->getParam('action'),
            ])
            ->toArray();
        foreach ($result as $name) {
            if ($name === $params['SaveFilterName']) {
                Notify::send(
                    $this->getController()->Auth->user('id'),
                    __("Les filtres n'ont pas été sauvegardés car un nom de sauvegarde identique existe déjà")
                );
                return;
            }
        }
        unset($params['SaveFilterSelect'], $params['SaveFilterName'], $params['filter'], $params['OverrideFilter']);
        $filters = $this->dataToFilters($params);
        return array_values($filters);
    }

    /**
     * Converti les données brut en filtres près à être enregistré
     * ['monfiltre' => ['mavaleur']] => [['key' => 'monfiltre[0]', 'value' => 'mavaleur']]
     * @param array $data
     * @return array
     */
    public function dataToFilters(array $data)
    {
        $filters = [];
        foreach ($data as $field => $values) {
            foreach ($values as $k => $value) {
                if (is_array($value)) {
                    foreach ($value as $sK => $sValue) {
                        $key = $field . '['.$k.']['.$sK.']';
                        $filters[$key] = [
                            'key' => $key,
                            'value' => $sValue
                        ];
                    }
                } else {
                    $key = $field . '['.$k.']';
                    $filters[$key] = compact('key', 'value');
                }
            }
        }
        return $filters;
    }

    /**
     * Logique de sauvegarde pour écraser une sauvegarde existante
     * (ne sauvegarde que si modifications, supprime le reste)
     *
     * @param Entity $entity SavedFilter
     * @param array $filters
     * @return void|bool success
     */
    private function overrideSavedFilters(Entity $entity, array $filters)
    {
        $SavedFilters = TableRegistry::getTableLocator()->get('SavedFilters');
        $Filters = TableRegistry::getTableLocator()->get('Filters');
        if (empty($filters)) {
            return $SavedFilters->delete($entity);
        }
        foreach ($entity->filters as $k => $e_filter) {
            $exists = false;
            foreach ($filters as $key => $filter) {
                if ($filter['key'] === $e_filter->key
                    && $filter['value'] === $e_filter->value
                ) {
                    $exists = true;
                    unset($filters[$key]);
                    break;
                }
            }
            if (!$exists) {
                $Filters->delete($e_filter);
                unset($entity->filters[$k]);
            }
        }
        foreach ($filters as $filter) {
            $newFilter = $Filters->newEntity([
                'saved_filter_id' => $entity->id,
                'key' => $filter['key'],
                'value' => $filter['value'],
            ]);
            $Filters->save($newFilter);
            $entity->filters[] = $newFilter;
        }
    }

    /**
     * Sauvegarde des filtres
     */
    private function saveFilters()
    {
        $SavedFilters = TableRegistry::getTableLocator()->get('SavedFilters');
        $newSave = $this->getController()->getRequest()->getParam('?.SaveFilterSelect') === '-1'
            && $this->getController()->getRequest()->getParam('?.SaveFilterName');
        $overrideSave = $this->getController()->getRequest()->getParam('?.SaveFilterSelect') !== '-1'
            && $this->getController()->getRequest()->getParam('?.OverrideFilter');
        $entity = null;
        if ($overrideSave) {
            $entity = $SavedFilters->find()
                ->where([
                    'id' => $this->getController()->getRequest()->getParam('?.SaveFilterSelect'),
                    'user_id' => $this->getController()->Auth->user('id'),
                    'controller' => $this->getController()->getRequest()->getParam('controller'),
                    'action' => $this->getController()->getRequest()->getParam('action'),
                ])
                ->contain(['Filters'])
                ->first();
            if (!$entity) {
                $overrideSave = false;
            }
        }

        if ($newSave) {
            $filters = $this->extractFilters();
            if (empty($filters)) {
                return;
            }
            $entity = $SavedFilters->newEntity([
                'name' => $this->getController()->getRequest()->getParam('?.SaveFilterName'),
                'user_id' => $this->getController()->Auth->user('id'),
                'controller' => $this->getController()->getRequest()->getParam('controller'),
                'action' => $this->getController()->getRequest()->getParam('action'),
                'filters' => $filters
            ], ['associated' => 'Filters']);
            $SavedFilters->save($entity);
        } elseif ($overrideSave) {
            /* @var Entity $entity */
            $this->overrideSavedFilters($entity, $this->extractFilters());
        }
    }
}
