<?php
/**
 * AsalaeCore\Controller\Component\RestComponent
 */

namespace AsalaeCore\Controller\Component;

use Acl\Model\Entity\Aro;
use AsalaeCore\Model\Table\AccessTokensTable;
use AsalaeCore\Model\Entity\Webservice;
use Cake\Controller\Controller;
use Cake\Controller\Component;
use Cake\Http\Exception\BadRequestException;
use Cake\Http\Exception\ForbiddenException;
use Cake\Http\Exception\NotAcceptableException;
use Cake\Http\Exception\NotFoundException;
use Cake\Http\Exception\UnauthorizedException;
use Cake\ORM\TableRegistry;
use Cake\Utility\Hash;
use Cake\Utility\Security;
use DateTime;
use DateTimeZone;

/**
 * Permet de désigner des actions comme action / rest
 *
 * @category Component
 *
 * @author    Ludovic Ganée <ludovic.ganee@libriciel.coop>
 * @copyright (c) 2017, Libriciel
 * @license   http://www.cecill.info/licences/Licence_CeCILL_V2-fr.html
 */
class RestComponent extends Component
{
    /**
     * @var string Identité ex: webservice_1
     */
    public $identity = 'anonymous';

    /**
     * @var Webservice Entité du webservice
     */
    public $webservice;

    /**
     * @var array Règles par défaut pour self::route()
     */
    private $defaultRules = [
        'auth' => false,
        'accept' => '*',
    ];

    /**
     * @var array Liste des actions possibles
     */
    private $actions = [];

    /**
     * Constructor hook method.
     *
     * Implement this method to avoid having to overwrite
     * the constructor and call parent.
     *
     * @param array $config The configuration settings provided to this component.
     * @return void
     */
    public function initialize(array $config)
    {
        /* @var Controller $Controller */
        $Controller = $this->getController();
        $Controller->Auth->allow('api');
        if ($Controller->getRequest()->getParam('action') === 'api') {
            $this->getController()->autoRender = false;
        }
    }

    /**
     * Permet de renvoyer une route vers une fonction privée choisie (callable)
     * tout en controllant les droits si besoin
     * @return array|bool
     * @throws \Aura\Intl\Exception
     */
    public function route()
    {
        /* @var Controller $Controller */
        $Controller = $this->getController();
        $routes = array_map(
            function ($v) {
                return array_change_key_case((array)$v) + ['*' => []];
            },
            Hash::normalize($this->actions)
        );
        $action = $Controller->getRequest()->getParam('pass.0');
        if (is_numeric($action) || $action === false) {
            $route = 'default';
        } else {
            $route = $action;
        }
        if ($route && array_key_exists($route, $routes)) {
            $method = strtolower($Controller->getRequest()->getMethod());
            if (!isset($routes[$route][$method])) {
                $method = '*';
            }
            $rules = $routes[$route][$method];
            if (!empty($routes[$route]['*']) && $method !== '*') {
                $rules += $routes[$route]['*'];
            }
            $rules += ['function' => $route] + $this->defaultRules;
            if ($rules['auth']) {
                $this->auth();
            }
            if ($rules['accept'] !== '*') {
                if (!$Controller->getRequest()->accepts($rules['accept'])) {
                    throw new NotAcceptableException(__("406 Not Acceptable"));
                }
            }
            $args = $Controller->getRequest()->getParam('pass');
            if ($route !== 'default') {
                unset($args[0]);
            }
            return [
                'callback' => [$Controller, $rules['function']],
                'args' => $args
            ];
        }
        return false;
    }

    /**
     * Vérifi l'autentification d'un webservice
     *
     * Javascript Basic auth :
     *      strValue = username.val() + ":" + password.val(),
     *      strBase64 = btoa(strValue).replace(/.{76}(?=.)/g,'$&\n')
     */
    public function auth()
    {
        /* @var Controller $Controller */
        $Controller = $this->getController();
        $username = $this->getUsername();
        $password = $this->getPassword();
        $webservice = null;

        // OAuth2
        $token = $Controller->getRequest()->getParam('?.access_token');
        $auth = $Controller->getRequest()->getHeader('Authorization');
        if ($auth && preg_match('/^Bearer ([\w]+)$/', current($auth), $match)) {
            $token = $match[1];
        }
        if ($token) {
            $AccessTokens = TableRegistry::getTableLocator()->get('AccessTokens');
            $token = $AccessTokens->find()
                ->where(['code' => $token])
                ->contain(['Webservices'])
                ->first();
            if (!$token) {
                $this->accessDenied();
            }
            $webservice = $token->get('webservice');

        // Basic
        } elseif (empty($username) || empty($password)) {
            $this->accessDenied();
        } else {
            $Webservices = TableRegistry::getTableLocator()->get('Webservices');
            $comparable = $Webservices->newEntity(['password' => $password]);
            $webservice = $Webservices->find()
                ->where([
                    'username' => $username,
                    'password' => $comparable->get('password'),
                    'enabled' => true
                ])
                ->first();
        }

        if (!$webservice) {
            $this->accessDenied();
        }

        $tableMethods = [
            'post' => 'create',
            'get' => 'read',
            'put' => 'update',
            'delete' => 'delete'
        ];
        $method = strtolower($Controller->getRequest()->getEnv('REQUEST_METHOD'));
        $action = $Controller->getRequest()->getParam('pass.0');
        $path = 'api/'.$Controller->getRequest()->getParam('controller');
        if (!is_numeric($action) && is_string($action)) {
            $path .= '/'.$action;
        }

        if ($token && $token->get('user_id')) {
            $user = TableRegistry::getTableLocator()->get('Users')->get($token->get('user_id'))->toArray();
            $aroData = ['model' => 'Users', 'foreign_key' => $user['id']];
            $Controller->Auth->setUser($user);
            $Controller->getRequest()->getSession()->write('User', $user);
        } else {
            $Controller->getRequest()->getSession()->destroy();
            $Controller->Auth->setUser([]);
            $aroData = ['model' => 'Webservices', 'foreign_key' => $webservice->get('id')];
        }

        /* @var Aro $aro */
        $aro = TableRegistry::getTableLocator()->get('Aros')
            ->find()
            ->where($aroData)
            ->first()
        ;
        $access = $Controller->Acl->check(
            $aro->get('id'),
            $path,
            $tableMethods[$method]
        );
        if (!$access) {
            throw new ForbiddenException(__("403 Forbidden"));
        }
        $this->identity = 'webservice_'.$webservice->get('id');
        $this->webservice = $webservice;
    }


    /**
     * Getter du client_id
     * @return string
     * @throws \Aura\Intl\Exception
     */
    public function getClientId(): string
    {
        return $this->genericGetter('client_id');
    }

    /**
     * Getter généric
     * @param string $varname
     * @param bool $throwException
     * @return string
     * @throws \Aura\Intl\Exception
     */
    private function genericGetter(string $varname, bool $throwException = true): string
    {
        $Request = $this->getController()->getRequest();
        $var = $Request->getParam('?.'.$varname)
            ?: $Request->getData($varname, '')
        ;
        if (!$var && $throwException) {
            throw new BadRequestException(
                __("400 Bad Request").' '.__("{0} est obligatoire dans ce contexte", $varname)
            );
        }
        return $var;
    }

    /**
     * Getter du username
     * @return string
     */
    public function getUsername(): string
    {
        return env('PHP_AUTH_USER') ?: (string)$this->getController()->getRequest()->getData('username');
    }

    /**
     * Getter du password
     * @return string
     */
    public function getPassword(): string
    {
        return env('PHP_AUTH_PW') ?: (string)$this->getController()->getRequest()->getData('password');
    }

    /**
     * Demande l'authentification
     * @param string $type 'Basic' ou 'Bearer'
     */
    private function accessDenied(string $type = 'Basic')
    {
        header('WWW-Authenticate: '.$type);
        throw new UnauthorizedException("401 Unauthorized");
    }

    /**
     * Définit les actions possibles via l'api
     *
     * exemple :
     * Le controller Users permet d'obtenir les infos d'un utilisateur et
     * possède également une méthode pour réinitialiser son mot de passe :
     *  $this->setApiActions([
     *      'default' => [          // pour https://mon.url/users[/:id]
     *          'get' => [
     *              'params' => ['id' => ['numeric', 'optional']],
     *              'function' => 'apiGet',
     *          ]
     *      ],
     *      'resetPassword' => [    // pour https://mon.url/users/reset-password/15
     *          'put' => [
     *              'function' => 'resetPassword'
     *          ]
     *      ]
     *  ])
     *
     * @param array $actions
     */
    public function setApiActions(array $actions)
    {
        $this->actions = $actions;
    }

    /**
     * Permet d'authentification via OAuth2
     * @return array data de la réponse
     * @throws \Aura\Intl\Exception
     */
    public function authorizeAction(): array
    {
        $Controller = $this->getController();
        $clientId = $this->genericGetter('client_id');
        $responseType = $this->genericGetter('response_type');
        if ($responseType !== 'code') {
            throw new BadRequestException(
                __("400 Bad Request").' '.__("{0} n'est pas supporté", 'response_type: ', $responseType)
            );
        }
        /** @var Webservice $webservice */
        $webservice = TableRegistry::getTableLocator()->get('Webservices')->find()
            ->where(['client_id' => $clientId, 'enabled' => true])
            ->first()
        ;
        if (!$webservice) {
            throw new NotFoundException(__("404 Not Found"));
        }

        if ($Controller->getRequest()->is(['put', 'post'])) {
            $isRequester = $webservice->get('username') === $this->getUsername()
                && $webservice->comparePassword($this->getPassword())
            ;
            $isClient = null;
            if (!$isRequester) {
                $isClient = TableRegistry::getTableLocator()->get('Users')->find()
                    ->where([
                        'name' => $this->getUsername(),
                        'actif' => true,
                        'password' => Security::hash(
                            $this->getPassword(),
                            'sha256',
                            true
                        )
                    ])
                    ->first()
                ;
            }
            if ($isRequester || $isClient) {
                /** @var AccessTokensTable $AccessTokens */
                $AccessTokens = TableRegistry::getTableLocator()->get('AccessTokens');
                /** @var Webservice $webservice */
                $token = $AccessTokens->createCodeToken($webservice, $isClient);
                $url = $webservice->get('callback');
                $url .= strpos($url, '?') ? '&' : '?';
                $url .= 'code='.$token->get('code');
                $url .= '&client_id='.$clientId;
                if ($state = $this->genericGetter('state', false)) {
                    $url .= '&state='.$state;
                    $data['state'] = $state;
                }
                return [
                    'code' => $token->get('code'),
                    'client_id' => $clientId,
                    'callback' => $webservice->get('callback'),
                    'redirect_url' => $url
                ];
            } else {
                $Controller->Flash->error(__("Login ou mot de passe invalide"));
            }
        }

        $Controller->set('webservice', $webservice);
        return [];
    }

    /**
     * Permet d'obtenir un jeu de jetons
     * @param string $type 'code' ou 'refresh_token'
     * @return array
     * @throws \Aura\Intl\Exception
     */
    private function tokenAction(string $type): array
    {
        $code = $this->genericGetter('code');
        $client_id = $this->genericGetter('client_id');
        $secret = $this->genericGetter('client_secret');
        $grantType = $this->genericGetter('grant_type');
        if ($grantType === 'authorization_code') {
            $grantType = 'code';
        }

        if ($grantType !== $type) {
            throw new BadRequestException(
                __("400 Bad Request").' '.__("{0} n'est pas supporté", 'grant_type: '.$grantType.' type: '.$type)
            );
        }
        $Webservices = TableRegistry::getTableLocator()->get('Webservices');
        /** @var Webservice $webservice */
        $webservice = $Webservices->find()
            ->where(['client_id' => $client_id, 'enabled' => true])
            ->first()
        ;

        if ($webservice) {
            /** @var AccessTokensTable $AccessTokens */
            $AccessTokens = TableRegistry::getTableLocator()->get('AccessTokens');
            $codeToken = $AccessTokens->find()
                ->where([
                    'webservice_id' => $webservice->get('id'),
                    'code' => $code,
                    'type' => $type,
                ])
                ->first();
            if ($codeToken->get('expiration_date') < new DateTime('now', new DateTimeZone('Europe/Paris'))) {
                $AccessTokens->delete($codeToken);
                return $this->accessDenied('Bearer');
            }
        } else {
            throw new NotFoundException(__("404 Not Found"));
        }

        if (empty($codeToken)
            || $webservice->get('client_secret_unsecure') !== $secret
        ) {
            throw new NotFoundException(__("404 Not Found"));
        }
        /** @var Webservice $webservice */
        $webservice = $Webservices->get($codeToken->get('webservice_id'));
        $AccessTokens->delete($codeToken);

        return $AccessTokens->createAccessTokens($webservice, $codeToken->get('user_id'));
    }

    /**
     * Echange d'un access_token en échange d'un code d'autentification
     * @return array access_token
     * @throws \Aura\Intl\Exception
     */
    public function accessTokenAction(): array
    {
        return $this->tokenAction('code');
    }

    /**
     * Echange un refresh token contre un nouveau jeu de tokens
     * @return array
     * @throws \Aura\Intl\Exception
     */
    public function refreshTokenAction(): array
    {
        return $this->tokenAction('refresh_token');
    }
}
