<?php
/**
 * Asalae\Controller\FiltersControllerTrait
 */

namespace AsalaeCore\Controller;

use AsalaeCore\Controller\Component\FilterComponent;
use AsalaeCore\Controller\Component\ModalComponent;
use AsalaeCore\Model\Entity\Filter;
use Cake\Controller\Component\AuthComponent;
use Cake\Utility\Hash;
use Cake\View\ViewBuilder;
use Exception;
use Cake\View\Helper\UrlHelper;
use Cake\View\View;

/**
 * Filtres de rechercher
 *
 * @category Controller
 *
 * @author    Ludovic Ganée <ludovic.ganee@libriciel.coop>
 * @copyright (c) 2017, Libriciel
 * @license   http://www.cecill.info/licences/Licence_CeCILL_V2-fr.html
 * @property \AsalaeCore\Model\Table\SavedFiltersTable $SavedFilters
 * @property \AsalaeCore\Model\Table\FiltersTable $Filters
 * @property AuthComponent $Auth
 * @property ModalComponent $Modal
 * @property FilterComponent $Filter
 */
trait FiltersControllerTrait
{
    /**
     * Traits
     */
    use RenderDataTrait;

    /**
     * Permet de rediriger vers l'url construite à partir des filtres sauvegardés
     * @param string $id
     * @return \Cake\Http\Response|null
     * @throws Exception
     */
    public function ajaxRedirectUrl(string $id)
    {
        $this->request->allowMethod('ajax');
        return $this->redirect($this->getUrlById($id));
    }

    /**
     * Permet de récupérer une url à partir des filtres sauvegardé
     * @param string $id
     * @return string
     * @throws Exception
     */
    private function getUrlById(string $id): string
    {
        $this->loadModel('SavedFilters');
        $save = $this->SavedFilters->find()
            ->where([
                'SavedFilters.id' => $id,
                'user_id' => $this->Auth->user('id')])
            ->contain(['Filters'])
            ->first();

        if (!$save) {
            throw new Exception(__("Valeur non trouvée"));
        }

        $UrlHelper = new UrlHelper(new View());
        $params = [
            'controller' => $save->get('controller'),
            'action' => $save->get('action'),
            '?' => ["SaveFilterSelect" => $id]
        ];
        /** @var Filter $filter */
        foreach ($save->get('filters') as $filter) {
            $params['?'][$filter->get('key')] = $filter->get('value');
        }

        return $UrlHelper->build($params, ['escape' => false, 'fullBase' => true]);
    }

    /**
     * Permet de supprimer une sauvegarde de filtre de recherche
     * @param string $id
     * @throws Exception
     */
    public function ajaxDeleteSave(string $id)
    {
        $this->request->allowMethod('ajax');
        $this->loadModel('SavedFilters');

        /** @var ViewBuilder $viewBuild */
        $viewBuild = $this->viewBuilder();
        $viewBuild->setClassName('Json');

        $save = $this->SavedFilters->find()
            ->where([
                'SavedFilters.id' => $id,
                'user_id' => $this->Auth->user('id')])
            ->contain(['Filters'])
            ->first();
        if (empty($save)) {
            throw new Exception(__("La sauvegarde du filtre n'a pas été trouvée"));
        }
        $success = $this->SavedFilters->delete($save);
        $this->loadComponent('AsalaeCore.Modal');
        if ($success) {
            $this->Modal->success();
        } else {
            $this->Modal->fail();
        }

        $this->set('success', $success);
        $this->set('_serialize', ['success']);
    }

    /**
     * @param string $id
     * @throws Exception
     */
    public function ajaxGetFilters(string $id)
    {
        $this->request->allowMethod('ajax');
        $this->loadModel('SavedFilters');
        $save = $this->SavedFilters->find()
            ->where([
                'SavedFilters.id' => $id,
                'user_id' => $this->Auth->user('id')])
            ->contain(['Filters'])
            ->first();

        if (!$save) {
            throw new Exception("Valeur non trouvée");
        }

        $this->set('savedFilters', $save);
        $this->set('helperConfig', json_decode($this->request->getData('helper'), true));
    }

    /**
     * Enregistre un jeu de filtres
     */
    public function ajaxNewSave()
    {
        $this->request->allowMethod('ajax');
        $this->request->allowMethod('post');
        $this->loadModel('SavedFilters');

        $this->loadComponent('AsalaeCore.Modal');
        $this->loadComponent('AsalaeCore.Filter');
        $data = $this->request->getData();
        $name = $data['savename'];
        $controller = $data['controller'];
        $action = $data['action'];
        $filters = $this->Filter->dataToFilters(array_filter($data, 'is_array'));
        if (empty($filters)) {
            $this->Modal->fail();
            $report = "Empty filters";
        } else {
            $entity = $this->SavedFilters->newEntity(
                compact('name', 'controller', 'action') + [
                    'user_id' => $this->Auth->user('id'),
                    'filters' => $filters
                ],
                ['associated' => 'Filters']
            );
            if ($this->SavedFilters->save($entity)) {
                $this->Modal->success();
                $report = [
                    'id' => $entity->get('id'),
                    'name' => h($entity->get('name')),
                ];
            } else {
                $this->Modal->fail();
                $report = $entity->getErrors();
            }
        }
        return $this->renderData($report);
    }

    /**
     * Remplace les filtres d'une sauvegarde
     * @param string $id
     * @return \Cake\Http\Response
     * @throws Exception
     */
    public function ajaxOverwrite(string $id)
    {
        $this->request->allowMethod('ajax');
        $this->request->allowMethod('post');
        $this->loadModel('SavedFilters');
        $this->loadModel('Filters');

        $this->loadComponent('AsalaeCore.Modal');
        $this->loadComponent('AsalaeCore.Filter');
        $data = $this->request->getData();
        $filters = $this->Filter->dataToFilters(array_filter($data, 'is_array'));

        if (empty($filters)) {
            $this->Modal->fail();
            $report = "Empty filters";
        } else {
            $entity = $this->SavedFilters->find()
                ->where(['id' => $id, 'user_id' => $this->Auth->user('id')])
                ->contain(['Filters'])
                ->firstOrFail()
            ;
            $this->Filters->deleteAll([
                'id IN' => Hash::extract($entity->get('filters'), '{n}.id')
            ]);
            $this->SavedFilters->patchEntity($entity, compact('filters'));
            if ($this->SavedFilters->save($entity, ['associated' => 'Filters'])) {
                $this->Modal->success();
                $report = [
                    'id' => $entity->get('id'),
                    'name' => h($entity->get('name')),
                    'filters' => $entity->get('filters')
                ];
            } else {
                $this->Modal->fail();
                $report = $entity->getErrors();
            }
        }
        return $this->renderData($report);
    }
}
