<?php
/**
 * AsalaeCore\Controller\RenderDataTrait
 */

namespace AsalaeCore\Controller;

use AsalaeCore\Http\AppServerRequest;
use Cake\Http\Response;
use AsalaeCore\Utility\XmlUtility;
use Cake\Http\Exception\NotAcceptableException;
use Cake\ORM\Entity;
use Cake\ORM\ResultSet;
use Cake\View\ViewBuilder;

/**
 * Permet le rendu d'une variable en divers formats (ex: json, xml)
 *
 * @category Controller
 *
 * @author    Ludovic Ganée <ludovic.ganee@libriciel.coop>
 * @copyright (c) 2017, Libriciel
 * @license   http://www.cecill.info/licences/Licence_CeCILL_V2-fr.html
 * @property AppServerRequest $request
 * @property Response $response
 */
trait RenderDataTrait
{
    /**
     * Effectue un rendu de fichier json
     * @param string $json
     * @param Response|null $response
     * @return Response|null
     */
    protected function renderJson(string $json, Response $response = null)
    {
        if ($response === null) {
            $response = $this->response;
        }
        $body = $response->getBody();
        $body->write($json);
        $response = $response
            ->withBody($body)
            ->withType('json')
        ;
        return $response;
    }

    /**
     * Effectue un rendu de fichier xml
     * @param string $xml
     * @param Response|null $response
     * @return Response|null
     */
    protected function renderXml(string $xml, Response $response = null)
    {
        if ($response === null) {
            $response = $this->response;
        }
        $body = $response->getBody();
        $body->write($xml);
        $response = $response
            ->withBody($body)
            ->withType('xml')
        ;
        return $response;
    }

    /**
     * Effectue un rendu de fichier html
     * @param mixed $data
     * @param Response|null $response
     * @return Response|null
     */
    protected function renderHtml($data, Response $response = null)
    {
        if ($response === null) {
            $response = $this->response;
        }
        $this->response = $response;
        /** @var ViewBuilder $View */
        $View = $this->viewBuilder();
        $View->setLayout('empty')->setTemplate('/Element/export');
        $this->set(compact('data'));
        return $this->render();
    }

    /**
     * Effectue le rendu des données selon le(s) accept(s) de la requête
     * @param array $accepts
     * @param mixed $output
     * @param Response|null $response
     * @return Response|null
     */
    protected function renderByAccepts(array $accepts, $output, Response $response = null)
    {
        if ($output instanceof ResultSet) {
            $output = $output->toArray();
        }
        if (is_array($output)) {
            $output = array_map(function ($v) {
                if ($v instanceof Entity) {
                    $v = $v->toArray();
                }
                return $v;
            }, $output);
        }
        switch (current($accepts)) {
            case 'json':
                return $this->renderJson(json_encode($output), $response);
            case 'xml':
                /** @var AppServerRequest $request */
                $request = $this->getRequest();
                $controller = $request->getParam('controller');
                $xml = XmlUtility::arrayToXmlString(
                    ['output' => [$controller => $output]]
                );
                return $this->renderXml($xml, $response);
            default:
                return $this->renderHtml($output, $response);
        }
    }

    /**
     * Effectue un rendu selon le accept
     * @param mixed $output
     * @param Response|null $response
     * @return Response
     * @throws \Aura\Intl\Exception
     */
    protected function renderData($output, Response $response = null): Response
    {
        if ($response === null) {
            $response = $this->response;
        }
        $acceptable = ['json', 'xml', 'html', 'xhtml'];
        /** @var AppServerRequest $request */
        $request = $this->getRequest();
        $accepts = current($response->mapType($request->parseAccept()));
        if (count($accepts) === 1 && current($accepts) === 'html') {
            $accepts = ['json'];
        }
        $intersect = array_intersect($accepts, $acceptable);
        if (!$intersect) {
            throw new NotAcceptableException(__("406 Not Acceptable"));
        }
        return $this->renderByAccepts($intersect, $output, $response);
    }
}
