<?php
/**
 * AsalaeCore\Error\AppErrorHandler
 */

namespace AsalaeCore\Error;

use Cake\Error\ErrorHandler;
use Cake\Console\ConsoleErrorHandler;
use Cake\Error\Debugger;

/**
 * Ajoute des gestions d'erreur à l'application
 *
 * @category    Error
 *
 * @author      Ludovic Ganée <ludovic.ganee@libriciel.coop>
 * @copyright   (c) 2017, Libriciel
 * @license     http://www.cecill.info/licences/Licence_CeCILL_V2-fr.html
 */
class AppErrorHandler extends ErrorHandler
{
    /**
     * @var ConsoleErrorHandler Error handler for Cake console.
     */
    public $ConsoleErrorHandler;

    /**
     * @var bool Affiche l'erreur si le debug > 2
     */
    private static $errorOutput = true;

    /**
     * @var array stock les erreurs si l'affichage est désactivé
     */
    private static $errors = [];

    /**
     * Constructor
     *
     * @param array $options The options for error handling.
     */
    public function __construct($options = [])
    {
        if (PHP_SAPI === 'cli') {
            $this->ConsoleErrorHandler = new ConsoleErrorHandler($options);
        }
        parent::__construct($options);
    }

    /**
     * Appelé dans config/bootstrap.php
     * Enregistre les fonctions de callback à effectuer lors de l'arret d'un script php
     */
    public function register()
    {
        register_shutdown_function("\Libriciel\Filesystem\Utility\Filesystem::rollbackAll");
        if (PHP_SAPI === 'cli') {
            $this->ConsoleErrorHandler->register();
        } else {
            parent::register();
        }
    }

    /**
     * Display an error.
     *
     * Template method of BaseErrorHandler.
     *
     * Only when debug > 2 will a formatted error be displayed.
     *
     * @param array $error An array of error data.
     * @param bool $debug Whether or not the app is in debug mode.
     * @return void
     */
    protected function _displayError($error, $debug)
    {
        if (!static::$errorOutput) {
            static::$errors[] = $error;
            return;
        }
        if (!$debug) {
            return;
        }
        /** @noinspection PhpParamsInspection */
        Debugger::getInstance()->outputError($error);
    }

    /**
     * Désactive l'affichage des erreurs (elles sont stockés dans self::$errors)
     */
    public static function disableErrorOutput()
    {
        static::$errorOutput = false;
    }

    /**
     * Active l'affichage des erreurs
     */
    public static function enableErrorOutput()
    {
        static::$errorOutput = true;
    }

    /**
     * Permet de capturer les erreurs d'une fonction en particulier
     *
     * @param callable $callback la fonction a executer sans erreurs
     * @param array $args arguments à envoyer au callback
     * @return array ['result' => 'résultat du callback', 'errors' => [errors]]
     */
    public static function captureErrors(callable $callback, ...$args): array
    {
        static::disableErrorOutput();
        $oldErrors = static::$errors;
        static::$errors = [];
        $result = call_user_func_array($callback, $args);
        $errors = static::$errors;
        static::$errors = $oldErrors;
        static::enableErrorOutput();
        return compact('result', 'errors');
    }

    /**
     * Log an error.
     *
     * @param string $level The level name of the log.
     * @param array $data Array of error data.
     * @return bool
     */
    protected function _logError($level, $data)
    {
        if (static::$errorOutput) {
            return parent::_logError($level, $data);
        }
    }
}
