<?php
/**
 * AsalaeCore\Error\ContextDebugger
 */

namespace AsalaeCore\Error;

use Cake\Error\Debugger;
use SplFileObject;

/**
 * Amélioration du debugger de CakePHP afin d'afficher en plus, le contexte.
 *
 * @category Debugger
 *
 * @author    Ludovic Ganée <ludovic.ganee@libriciel.coop>
 * @copyright (c) 2017, Libriciel
 * @license   http://www.cecill.info/licences/Licence_CeCILL_V2-fr.html
 */
class ContextDebugger
{
    /**
     * @var array Structure d'affichage du debug :
     *            container - lineInfoContainer, context, degug
     *            lineInfo - fileShort, traceLine
     *            lineInfoContainer - lineInfo, uid
     *            context - uid, lineContent
     *            debug - vars
     *            var - key, value
     */
    public static $templates = [
        'container' => '<div class="cake-debug-output">%s%s%s</div>',
        'lineInfo' => '<span><strong>%s</strong> (line <strong>%s</strong>)</span>',
        'lineInfoContainer' => '%s<br><a href="javascript:(function(){var a=document.getElementById(\'%s\').style;a.display = a.display==\'none\'?\'block\':\'none\'})()" role="button">Context</a>',
        'context' => '<pre class="cake-debug" style="display:none" id="%s">%s</pre>',
        'debug' => '<pre class="cake-debug">%s</pre>',
        'var' => '<b style="background:lightgray;display:block;margin:15px 0 5px;border-bottom:2px solid gray">%s</b>%s'
    ];

    /**
     * @var array Liste des methodes à ignorer dans le debug_backtrace
     */
    private static $debugFunctions = ['AsalaeCore\Error\ContextDebugger::debug'];

    /**
     * Ajoute une methode à la liste des méthodes
     * de debug pour être ignoré par le debug_backtrace
     * @param string $method obtenu par la constante magique __METHOD__
     *                       ex: AsalaeCore\Error\MyDebugger::debug
     */
    public static function registerDebugFunction(string $method)
    {
        if (!in_array($method, self::$debugFunctions)) {
            self::$debugFunctions[] = $method;
        }
    }

    /**
     * Fonction de debug, permet de chainer les arguments
     * ex: d('foo', 'bar') affichera 'foo' et 'bar'
     * Ajoute quelques lignes du contexte (code)
     *
     * @see debug()
     * @param array $args
     * @return mixed
     */
    public static function debug(...$args)
    {
        list($file, $traceLine, $caller) = self::getFileTrace();

        $spl = new SplFileObject($file);
        $linesToParse = 10; // incrementation dans self::readFile
        do {
            $extracted = self::readFile($spl, $linesToParse, $traceLine);
            $lineContent = $extracted['lineContent'];
            $fullContent = $extracted['fullContent'];
            $lines = $extracted['lines'];

            $allMatches = self::findMatchingArgs($fullContent, $lines, $traceLine, $caller);
            $parsedArgs = self::parseArgs($allMatches, count($args));
        } while (count($args) > count($parsedArgs) && $linesToParse < 100);
        foreach ($args as $k => $value) {
            $args[$parsedArgs[$k]] = $value;
            unset($args[$k]);
        }

        self::renderDebug($args, $file, $traceLine, $lineContent);

        return current($args);
    }

    /**
     * Permet d'obtenir le contexte à partir d'un nom de fichier et du numéro de
     * ligne
     * @param string $filename
     * @param int $lineNumber
     * @param int $linesAround
     * @return string
     */
    public static function getContext(string $filename, int $lineNumber, int $linesAround = 5): string
    {
        $spl = new SplFileObject($filename);

        $lineContent = '';
        for ($line = $lineNumber - $linesAround; $line <= $lineNumber + $linesAround; $line++) {
            if (!$spl->eof() && $line > 0) {
                $spl->seek($line -1);
                $current = $spl->current();
                $content = sprintf("%0".strlen($line)."d : %s", ($line), h($current));
                $lineContent .= $line === $lineNumber ? '<b>'.$content.'</b>' : $content;
            }
        }
        return $lineContent;
    }

    /**
     * Transforme le contenu d'une variable en HTML
     * @param $value
     * @return null|string
     */
    public static function exportVar($value)
    {
        if (method_exists('Symfony\Component\VarDumper\VarDumper', 'dump')) {
            $cloner = new \Symfony\Component\VarDumper\Cloner\VarCloner();
            $dumper = new \Symfony\Component\VarDumper\Dumper\HtmlDumper();
            return $dumper->dump($cloner->cloneVar($value), true);
        } else {
            return h(Debugger::exportVar($value, 25));
        }
    }

    /**
     * Utilise les templates afin de générer le code html affiché
     * @param array $args
     * @param string $file
     * @param string $traceLine
     * @param string $lineContent
     */
    private static function renderDebug(array $args, string $file, string $traceLine, string $lineContent)
    {
        $uid = uniqid();
        $fileShort = str_replace([ROOT, CAKE_CORE_INCLUDE_PATH], '', $file);
        $vars = '';
        foreach ($args as $key => $value) {
            $vars .= sprintf(self::$templates['var'], h($key), self::exportVar($value));
        }

        $lineInfo = sprintf(self::$templates['lineInfo'], $fileShort, $traceLine);
        $lineInfoContainer = sprintf(self::$templates['lineInfoContainer'], $lineInfo, $uid);
        $context = sprintf(self::$templates['context'], $uid, trim($lineContent));
        $debug = sprintf(self::$templates['debug'], $vars);
        printf(self::$templates['container'], $lineInfoContainer, $context, $debug);
    }

    /**
     * Permet d'obtenir l'emplacement de l'appel du debug
     * @return array ['/path/to/file', '123']
     */
    private static function getFileTrace(): array
    {
        $traces = Debugger::trace(['start' => 2, 'depth' => 5, 'format' => 'array']);
        krsort($traces);
        foreach ($traces as $trace) {
            $function = $trace['function'];
            $class = isset($trace['class']) ? $trace['class'] : '';
            $type = isset($trace['type']) ? $trace['type'] : '';
            if (in_array($class . $type . $function, self::$debugFunctions)) {
                return [
                    $trace['file'],
                    $trace['line'],
                    $function
                ];
            }
        }
        return ['', 0];
    }

    /**
     * Permet de lire un fichier autour des lignes $traceLine
     * @param SplFileObject $spl
     * @param int $linesToParse
     * @param int $traceLine
     * @return array
     */
    private static function readFile(SplFileObject $spl, int &$linesToParse, int $traceLine): array
    {
        $linesToParse += 5;
        $lineContent = '';
        $fullContent = '';
        $lines = [];
        for ($i = 0; $i < $linesToParse; $i++) {
            if (!$spl->eof() && ($traceLine + $i -8) >= 0) {
                $spl->seek($traceLine + $i -8);
                $current = $spl->current();
                $lines[$traceLine + $i -7] = $current;
                $content = sprintf("%0".strlen($traceLine + 7)."d : %s", ($traceLine + $i -7), h($current));
                $lineContent .= $i === 7 ? '<b>'.$content.'</b>' : $content;
                $fullContent .= $current;
            } else {
                $lineContent .= PHP_EOL;
            }
        }
        return compact('lineContent', 'fullContent', 'lines');
    }

    /**
     * Permet de récupérer les arguments possible de l'appel du debug
     * @param string $fullContent
     * @param array $lines
     * @param int $traceLine
     * @param string $caller fonction appelante (ex: debug)
     * @return array
     */
    private static function findMatchingArgs(string $fullContent, array $lines, int $traceLine, string $caller): array
    {
        $allMatches = [];
        $offset = -1;
        while (preg_match('/(?:^|\s|;|\{|\}|\*\/)(?:[\\\\\w\$]+(?:::|->))?'.$caller.'\((.*)\)\s*;/s', $fullContent, $matches, PREG_OFFSET_CAPTURE, $offset +1)) {
            // Detection du numéro de la ligne du paramètre
            $possibleNumbers = [];
            foreach (explode(PHP_EOL, $matches[1][0]) as $line) {
                if (empty($line)) {
                    continue;
                }
                $search = array_filter($lines, function ($value) use ($line, $traceLine) {
                    return strpos($value, $line) !== false;
                });
                $possibleNumbers[] = array_keys($search);
            }

            $count = count($possibleNumbers);
            for ($i = 0; $i < $count; $i++) {
                if (isset($possibleNumbers[$i +1])) {
                    $possibleNumbers[$i] = array_filter($possibleNumbers[$i], function ($value) use ($possibleNumbers, $i) {
                        return array_search($value +1, $possibleNumbers[$i +1]) !== false;
                    });
                }
                if (isset($possibleNumbers[$i -1])) {
                    $possibleNumbers[$i] = array_filter($possibleNumbers[$i], function ($value) use ($possibleNumbers, $i, $traceLine) {
                        return array_search($value -1, $possibleNumbers[$i -1]) !== false;
                    });
                }
                ksort($possibleNumbers[$i]);
            }
            ksort($possibleNumbers);
            $matchingLineNumber = current($possibleNumbers[$count -1]) - $count +1;
            $offset = $matches[1][1];
            if ($matchingLineNumber && $matchingLineNumber <= $traceLine) {
                $allMatches[$matchingLineNumber] = $matches[1][0];
            }
        }
        krsort($allMatches);
        return $allMatches;
    }

    /**
     * Extraction des informations contenu dans le code
     * @param array $allMatches
     * @param int $targetArgCount
     * @return array
     */
    private static function parseArgs(array $allMatches, int $targetArgCount): array
    {
        $parsedArgs = [];
        foreach ($allMatches as $match) {
            $cut = self::extractArgs($match);
            $parsedArgs = self::argsStringToArray($cut);
            if (count($parsedArgs) === $targetArgCount) {
                break;
            }
        }
        return $parsedArgs;
    }

    /**
     * Extraction du code de la fonction de debug
     * ex: debug('foo', 'bar') return "'foo', 'bar'"
     * @param string $match
     * @return string
     */
    private static function extractArgs(string $match): string
    {
        $cut = '';
        foreach (token_get_all('<?php '.$match) as $token) {
            if (is_array($token) && $token[0] === T_OPEN_TAG) {
                continue;
            }
            if (is_array($token)) {
                $cut .= $token[1];
            } elseif ($token === ';') {
                if ($cut[strlen($cut)-1] === ')') {
                    $cut = substr($cut, 0, strlen($cut)-1); // <?php 'test_1'); d('test_2'); => $cut = "'test_1'"
                }
                break;
            } else {
                $cut .= $token;
            }
        }
        return $cut;
    }

    /**
     * Transforme les arguments 'foo', 'bar' en array['foo', 'bar']
     * @param string $args
     * @return array
     */
    private static function argsStringToArray(string $args): array
    {
        $parsedArgs = [];
        $parsedArg = '';
        $pCount = 1;
        foreach (token_get_all('<?php '.$args) as $token) {
            if (is_array($token) && $token[0] === T_OPEN_TAG) {
                continue;
            }
            if ($token === '(') {
                $pCount++;
                if ($pCount > 1) {
                    $parsedArg .= $token;
                }
            } elseif ($token === ')') {
                $pCount--;
                $parsedArg .= $token;
            } elseif ($token === '[') {
                $parsedArg .= $token;
                $pCount++;
            } elseif ($token === ']') {
                $parsedArg .= $token;
                $pCount--;
            } elseif ($token === ',') {
                if ($pCount === 1) {
                    $parsedArgs[] = trim($parsedArg);
                    $parsedArg = '';
                } else {
                    $parsedArg .= $token;
                }
            } elseif (is_array($token) && !empty($token[1])) {
                $parsedArg .= $token[1];
            } else {
                $parsedArg .= $token;
            }
        }
        $parsedArgs[] = trim($parsedArg);
        return $parsedArgs;
    }
}
