<?php
/**
 * AsalaeCore\Form\ConfigurationForm
 */

namespace AsalaeCore\Form;

use Cake\Database\Connection;
use Cake\Form\Form;
use Cake\Form\Schema;
use Cake\Validation\Validator;
use Libriciel\Filesystem\Utility\Filesystem;
use Symfony\Component\Filesystem\Exception\IOException;
use Exception;
use FileValidator\Utility\FileValidator;
use Cake\Datasource\ConnectionManager;
use PDOException;
use Cake\Utility\Hash;
use AsalaeCore\Validation\ValidationRule;

/**
 * Formulaire de modification de la configuration
 *
 * @category Form
 *
 * @author    Ludovic Ganée <ludovic.ganee@libriciel.coop>
 * @copyright (c) 2017, Libriciel
 * @license   http://www.cecill.info/licences/Licence_CeCILL_V2-fr.html
 */
class ConfigurationForm extends Form
{
    /**
     * A hook method intended to be implemented by subclasses.
     *
     * You can use this method to define the schema using
     * the methods on Cake\Form\Schema, or loads a pre-defined
     * schema from a concrete class.
     *
     * @param Schema $schema The schema to customize.
     * @return Schema The schema to use.
     */
    protected function _buildSchema(Schema $schema): Schema
    {
        $schema->addField('config', ['type' => 'text'])
            ->addField('hash_algo', ['type' => 'string'])
            ->addField('domain', ['type' => 'string'])
            ->addField('datacompressor_encodefilename', ['type' => 'string'])
            ->addField('password_complexity', ['type' => 'integer'])
            ->addField('filesystem_useshred', ['type' => 'boolean'])
            ->addField('assets_login_logo', ['type' => 'string'])
            ->addField('assets_login_background', ['type' => 'string'])
            ->addField('assets_global_background', ['type' => 'string'])
            ->addField('assets_global_background_position', ['type' => 'string'])
            ->addField('emails_from', ['type' => 'string'])
            ->addField('emails_host', ['type' => 'string'])
            ->addField('emails_port', ['type' => 'integer'])
            ->addField('emails_username', ['type' => 'string'])
            ->addField('emails_password', ['type' => 'string'])
            ->addField('datasources_default_driver', ['type' => 'string'])
            ->addField('datasources_default_host', ['type' => 'string'])
            ->addField('datasources_default_username', ['type' => 'string'])
            ->addField('datasources_default_password', ['type' => 'string'])
            ->addField('datasources_default_database', ['type' => 'string'])
            ->addField('datasources_default_log', ['type' => 'boolean'])
            ->addField('datasources_debug_kit_driver', ['type' => 'string'])
            ->addField('datasources_debug_kit_host', ['type' => 'string'])
            ->addField('datasources_debug_kit_username', ['type' => 'string'])
            ->addField('datasources_debug_kit_password', ['type' => 'string'])
        ;
        // Fix Undefined method (bug phpstorm) (trop de fonctions chainés ?)
        $schema->addField('datasources_debug_kit_database', ['type' => 'string'])
            ->addField('datasources_test_driver', ['type' => 'string'])
            ->addField('datasources_test_host', ['type' => 'string'])
            ->addField('datasources_test_username', ['type' => 'string'])
            ->addField('datasources_test_password', ['type' => 'string'])
            ->addField('datasources_test_database', ['type' => 'string'])
        ;
        return $schema;
    }

    /**
     * Valide l'onglet Utilitaires
     *
     * @param Validator $validator
     * @return \self
     * @throws \Aura\Intl\Exception
     */
    private function validateUtilities(Validator $validator): self
    {
        $validator
            ->add('config', 'custom', [
                'rule' => function ($value) {
                    json_decode($value);
                    $error = json_last_error();
                    return $error === JSON_ERROR_NONE;
                },
                'message' => __("N'est pas un json valide")
            ])
            ->add('hash_algo', 'inList', [
                'rule' => ['inList', array_keys($this->options('hash_algo'))]
            ])
            ->allowEmpty('hash_algo')
            ->add('password_complexity', 'isInteger', [
                'rule' => 'isInteger'
            ])
            ->allowEmpty('password_complexity')
            ->add('ratchet_connect', 'custom', [
                'rule' => ['custom', '/^wss?:\/\/[\w\.\/]+$/']
            ])
            ->add('domain', 'custom', [
                'rule' => function ($value) {
                    try {
                        $dns = @dns_get_record($value);
                    } catch (Exception $ex) {
                        $dns = null;
                    }
                    if (empty($dns)) {
                        $context = stream_context_create([
                            "http" => ["method" => "GET"],
                            "ssl" => [
                                "verify_peer" => false,
                                "allow_self_signed" => true,
                                "verify_peer_name" => false,
                            ],
                        ]);
                        $dns = @file_get_contents('http://'.$value, false, $context);
                        if ($dns === false) {
                            $dns = @file_get_contents('https://'.$value, false, $context);
                        }
                    }
                    return !empty($dns);
                },
                'message' => __("N'est pas un dns valide")
            ])
            ->add('datacompressor_encodefilename', 'inList', [
                'rule' => ['inList', array_keys($this->options('datacompressor_encodefilename'))]
            ])
            ->allowEmpty('datacompressor_encodefilename')
            ->add('filesystem_useshred', 'inList', [
                'rule' => ['inList', array_keys($this->options('filesystem_useshred'))]
            ])
            ->allowEmpty('filesystem_useshred');
        return $this;
    }

    /**
     * Valide l'onglet Apparence
     *
     * @param Validator $validator
     * @return \self
     * @throws \Aura\Intl\Exception
     */
    private function validateAppearance(Validator $validator): self
    {
        $validator
            ->add('assets_login_logo', 'is_file', [
                'rule' => function ($value) {
                    return is_file($value);
                },
                'message' => __("N'est pas un fichier")
            ])
            ->add('assets_login_logo', 'readable', [
                'rule' => function ($value) {
                    return !is_file($value) || is_readable($value);
                },
                'message' => __("Ne peut pas être lu, vérifiez les autorisations")
            ])
            ->add('assets_login_logo', 'is_image', [
                'rule' => function ($value) {
                    return !is_readable($value)
                        || ((bool)preg_match('/^image\//', mime_content_type($value))
                        && FileValidator::check($value));
                },
                'message' => __("N'est pas une image valide")
            ])
            ->allowEmpty('assets_login_logo')
            ->allowEmpty('assets_login_background')
            ->add('assets_global_background', 'is_file', [
                'rule' => function ($value) {
                    return is_file($value);
                },
                'message' => __("N'est pas un fichier")
            ])
            ->add('assets_global_background', 'readable', [
                'rule' => function ($value) {
                    return !is_file($value) || is_readable($value);
                },
                'message' => __("Ne peut pas être lu, vérifiez les autorisations")
            ])
            ->add('assets_global_background', 'is_image', [
                'rule' => function ($value) {
                    return !is_readable($value)
                        || ((bool)preg_match('/^image\//', mime_content_type($value))
                        && FileValidator::check($value));
                },
                'message' => __("N'est pas une image valide")
            ])
            ->allowEmpty('assets_global_background')
            ->allowEmpty('assets_global_background_position');
        return $this;
    }

    /**
     * Valide l'onglet E-mails
     *
     * @param Validator $validator
     * @return \self
     * @throws \Aura\Intl\Exception
     */
    private function validateEmails(Validator $validator): self
    {
        $validator
            ->add('emails_from', 'email', ['rule' => 'email'])
            ->allowEmpty('emails_from')
            ->allowEmpty('emails_host')
            ->add('emails_port', 'isInteger', [
                'rule' => 'isInteger'
            ])
            ->add('emails_port', 'is_port', [
                'rule' => function ($value) {
                    return (int)$value > 0 && (int)$value <= 65535;
                },
                'message' => __("N'est pas une image valide")
            ])
            ->allowEmpty('emails_port')
            ->allowEmpty('emails_username')
            ->allowEmpty('emails_password');
        return $this;
    }

    /**
     * Valide l'onglet Datasources
     *
     * @param Validator $validator
     * @return \self
     * @throws \Aura\Intl\Exception
     */
    private function validateDatasources(Validator $validator): self
    {
        $default = include CONFIG . 'app_default.php';
        foreach (['default', 'debug_kit', 'test'] as $datasource) {
            /** @noinspection PhpUnusedParameterInspection */
            $rule = new ValidationRule([
                'rule' => function (/** @noinspection PhpUnusedParameterInspection */
                    $value,
                    $context
                ) use (
                    $datasource,
                    $default
) {
                    try {
                        $json = json_decode($context['data']['config'], true);
                    } catch (Exception $e) {
                        $json = [];
                    }
                    if (isset($json['Datasources'][$datasource])) {
                        $default['Datasources'][$datasource] = array_merge(
                            $default['Datasources'][$datasource],
                            $json['Datasources'][$datasource]
                        );
                    }
                    foreach (['driver', 'host', 'username', 'password', 'database'] as $k) {
                        $v = $context['data']['datasources_'.$datasource.'_'.$k];
                        if (!empty($v)) {
                            $default['Datasources'][$datasource][$k] = $v;
                        }
                    }

                    try {
                        ConnectionManager::drop('validate_datasource');
                        ConnectionManager::setConfig(
                            'validate_datasource',
                            Hash::get($default, 'Datasources.'.$datasource)
                        );
                        $conn = ConnectionManager::get('validate_datasource');
                        if ($conn instanceof Connection) {
                            $conn->query('select 1');
                        }
                        return true;
                    } catch (PDOException $e) {
                        $rule = $this->getValidator()
                            ->field('datasources_'.$datasource.'_database')
                            ->rule('connectable');
                        if ($rule instanceof ValidationRule) {
                            $rule->setMessage($e->getMessage());
                        }
                        return false;
                    }
                },
                'message' => __("La connexion à la base de données a échoué")
            ]);
            $validator
                ->add('datasources_'.$datasource.'_driver', 'inList', [
                    'rule' => ['inList', array_keys($this->options('datasources_'.$datasource.'_driver'))]
                ])
                ->allowEmpty('datasources_'.$datasource.'_driver')
                ->allowEmpty('datasources_'.$datasource.'_host')
                ->allowEmpty('datasources_'.$datasource.'_username')
                ->allowEmpty('datasources_'.$datasource.'_password')
                ->allowEmpty('datasources_'.$datasource.'_database')
                ->add('datasources_'.$datasource.'_database', 'connectable', $rule)
            ;
        }
        return $this;
    }

    /**
     * A hook method intended to be implemented by subclasses.
     *
     * You can use this method to define the validator using
     * the methods on Cake\Validation\Validator or loads a pre-defined
     * validator from a concrete class.
     *
     * @param Validator $validator The validator to customize.
     * @return Validator The validator to use.
     * @throws \Aura\Intl\Exception
     */
    protected function _buildValidator(Validator $validator): Validator
    {
        $this->validateUtilities($validator)
            ->validateAppearance($validator)
            ->validateEmails($validator)
            ->validateDatasources($validator)
        ;
        return $validator;
    }

    /**
     * Sauvegarde l'onglet Utilitaires
     *
     * @param array $data
     * @param array $config
     * @return \self
     */
    private function saveUtilities(array $data, array &$config): self
    {
        if (!empty($data['domain'])) {
            $config['Domain'] = $data['domain'];
        } else {
            unset($config['domain']);
        }
        if (!empty($data['hash_algo'])) {
            $config['hash_algo'] = $data['hash_algo'];
        } else {
            unset($config['hash_algo']);
        }
        if (!empty($data['password_complexity']) || $data['password_complexity'] === '0') {
            $config['Password']['complexity'] = (int)$data['password_complexity'];
        } else {
            unset($config['Password']['complexity']);
            if (empty($config['Password'])) {
                unset($config['Password']);
            }
        }
        if (!empty($data['ratchet_connect'])) {
            $config['Ratchet']['connect'] = $data['ratchet_connect'];
        } else {
            unset($config['Ratchet']['connect']);
            if (empty($config['Ratchet'])) {
                unset($config['Ratchet']);
            }
        }
        if (!empty($data['datacompressor_encodefilename'])) {
            $config['DataCompressorUtility']['encodeFilename'] = $data['datacompressor_encodefilename'];
        } elseif ($data['datacompressor_encodefilename'] === '0') {
            $config['DataCompressorUtility']['encodeFilename'] = false;
        } else {
            unset($config['DataCompressorUtility']['encodeFilename']);
            if (empty($config['DataCompressorUtility'])) {
                unset($config['DataCompressorUtility']);
            }
        }
        if ($data['filesystem_useshred'] !== '') {
            $config['FilesystemUtility']['useShred'] = $data['filesystem_useshred'] === '1';
        } else {
            unset($config['FilesystemUtility']['useShred']);
            if (empty($config['FilesystemUtility'])) {
                unset($config['FilesystemUtility']);
            }
        }
        return $this;
    }

    /**
     * Sauvegarde l'onglet Apparence
     *
     * @param array $data
     * @param array $config
     * @return \self
     */
    private function saveAppearance(array $data, array &$config): self
    {
        if (!empty($data['assets_login_logo'])) {
            $config['Assets']['login']['logo-client'] = $data['assets_login_logo'];
        } else {
            unset($config['Assets']['login']['logo-client']);
        }
        if (!empty($data['assets_login_background'])) {
            $config['Assets']['login']['background'] = $data['assets_login_background'];
        } else {
            unset($config['Assets']['login']['background']);
        }
        if (empty($config['Assets']['login'])) {
            unset($config['Assets']['login']);
        }
        if (!empty($data['assets_global_background'])) {
            $config['Assets']['global-background'] = $data['assets_global_background'];
        } else {
            unset($config['Assets']['global-background']);
        }
        if (!empty($data['assets_global_background_position'])) {
            $config['Assets']['global-background-position'] = $data['assets_global_background_position'];
        } else {
            unset($config['Assets']['global-background-position']);
        }
        if (empty($config['Assets'])) {
            unset($config['Assets']);
        }
        return $this;
    }

    /**
     * Sauvegarde l'onglet E-mails
     *
     * @param array $data
     * @param array $config
     * @return \self
     */
    private function saveEmails(array $data, array &$config): self
    {
        if (!empty($data['emails_from'])) {
            $config['Email']['default']['from'] = $data['emails_from'];
        } else {
            unset($config['Email']['default']['from']);
            if (empty($config['Email']['default'])) {
                unset($config['Email']['default']);
            }
            if (empty($config['Email'])) {
                unset($config['Email']);
            }
        }
        if (!empty($data['emails_host'])) {
            $config['EmailTransport']['default']['host'] = $data['emails_host'];
        } else {
            unset($config['EmailTransport']['default']['host']);
        }
        if (!empty($data['emails_port'])) {
            $config['EmailTransport']['default']['port'] = $data['emails_port'];
        } else {
            unset($config['EmailTransport']['default']['port']);
        }
        if (!empty($data['emails_username'])) {
            $config['EmailTransport']['default']['username'] = $data['emails_username'];
        } else {
            unset($config['EmailTransport']['default']['username']);
        }
        if (!empty($data['emails_password'])) {
            $config['EmailTransport']['default']['password'] = $data['emails_password'];
        } else {
            unset($config['EmailTransport']['default']['password']);
        }
        if (empty($config['EmailTransport']['default'])) {
            unset($config['EmailTransport']['default']);
        }
        if (empty($config['EmailTransport'])) {
            unset($config['EmailTransport']);
        }
        return $this;
    }

    /**
     * Sauvegarde l'onglet Datasources
     *
     * @param array $data
     * @param array $config
     * @return \self
     */
    private function saveDatasources(array $data, array &$config): self
    {
        foreach (['default', 'debug_kit', 'test'] as $datasource) {
            foreach (['driver', 'host', 'username', 'password', 'database', 'log'] as $k) {
                if (!empty($data['datasources_'.$datasource.'_'.$k])) {
                    $config['Datasources'][$datasource][$k]
                        = $data['datasources_'.$datasource.'_'.$k];
                } else {
                    unset($config['Datasources'][$datasource][$k]);
                }
            }
            if (empty($config['Datasources'][$datasource])) {
                unset($config['Datasources'][$datasource]);
            }
        }
        if (empty($config['Datasources'])) {
            unset($config['Datasources']);
        }
        return $this;
    }

    /**
     * Hook method to be implemented in subclasses.
     *
     * Used by `execute()` to execute the form's action.
     *
     * @param array $data Form data.
     * @return bool
     * @throws Exception
     * @throws \Aura\Intl\Exception
     */
    protected function _execute(array $data): bool
    {
        $config = json_decode($data['config'], true);

        $this->saveUtilities($data, $config)
            ->saveAppearance($data, $config)
            ->saveEmails($data, $config)
            ->saveDatasources($data, $config)
        ;

        ksort($config);
        $json = json_encode($config, JSON_PRETTY_PRINT);

        try {
            Filesystem::begin('edit_config');
            Filesystem::dumpFile(
                include CONFIG . 'path_to_local.php',
                $json
            );
            Filesystem::commit('edit_config');
            return true;
        } catch (IOException $e) {
            Filesystem::rollback('edit_config');
            trigger_error($e->getMessage());
            return false;
        }
    }

    /**
     * Permet d'obtenir les options possible pour un $field
     *
     * @param string $field
     * @return array
     * @throws \Aura\Intl\Exception
     */
    public static function options(string $field): array
    {
        switch ($field) {
            case 'hash_algo':
                return [
                    'crc32' => __x('hash_algo', 'crc32 - Très rapide'),
                    'md5' => __x('hash_algo', 'md5 - Très connu'),
                    'sha1' => __x('hash_algo', 'sha1 - Très connu'),
                    'sha256' => __x('hash_algo', 'sha256 - Lourd'),
                    'sha512' => __x('hash_algo', 'sha512 - Très lourd'),
                    'whirlpool' => __x('hash_algo', 'whirlpool - Ultra lourd (+ secure)'),
                ];
            case 'password_complexity':
                return [
                    -1 =>  __x('password_complexity', '--- Valeur custom ---'),
                    PASSWORD_ULTRA_WEAK => __x('password_complexity', 'Très faible (49) - (~8 caractères avec chiffres lettres majuscules et spéciaux)'),
                    PASSWORD_WEAK => __x('password_complexity', 'Faible (78) - Taille minimale recommandée par l\'ANSSI (~12)'),
                    PASSWORD_MEDIUM => __x('password_complexity', 'Moyen (82) - Taille recommandée par l\'ANSSI (~16)'),
                    PASSWORD_STRONG => __x('password_complexity', 'Fort (130) - Equivalent clef 128bits (~20)'),
                ];
            case 'datacompressor_encodefilename':
                return [
                    '0' => __x('datacompressor_encodefilename', 'Ne pas ré-encoder les noms des fichiers'),
                    'IBM850' => __x('datacompressor_encodefilename', 'IBM850 - Encodage Windows'),
                    'ASCII//TRANSLIT' => __x('datacompressor_encodefilename', 'ASCII//TRANSLIT - Retire les caractères spéciaux (compatibilité max)'),
                ];
            case 'filesystem_useshred':
                return [
                    '0' => __('Non'),
                    '1' => __('Oui'),
                ];
            case 'datasources_default_driver':
            case 'datasources_debug_kit_driver':
            case 'datasources_test_driver':
                return [
                    'Cake\Database\Driver\Mysql' => __x('datasources_driver', 'Mysql'),
                    'Cake\Database\Driver\Postgres' => __x('datasources_driver', 'Postgres'),
                    'Cake\Database\Driver\Sqlite' => __x('datasources_driver', 'Sqlite'),
                    'Cake\Database\Driver\Sqlserver' => __x('datasources_driver', 'Sqlserver'),
                ];
            default:
                return [];
        }
    }
}
