<?php
/**
 * AsalaeCore\Http\AppServerRequest
 */
namespace AsalaeCore\Http;

use ArrayAccess;
use Cake\Http\ServerRequest;
use Psr\Http\Message\ServerRequestInterface;

/**
 * A class that helps wrap Request information and particulars about a single request.
 * Provides methods commonly used to introspect on the request headers and request body.
 */
class AppServerRequest extends ServerRequest implements ArrayAccess, ServerRequestInterface
{
    /**
     * @var AppServerRequest instance singleton
     */
    private static $instance;

    /**
     * Permet de récupérer l'instance unique de la classe
     * @param array $config
     * @return AppServerRequest
     */
    public static function getInstance(array $config = [])
    {
        return self::$instance = self::$instance ?? new self($config);
    }

    /**
     * Nettoyage des fichiers temporaires
     */
    public function __destruct()
    {
        foreach (array_keys($this->files) as $path) {
            if (is_file($path)) {
                unlink($path);
            }
        }
    }

    /**
     * @var array fichiers à supprimer
     */
    private $files = [];

    /**
     * PHP ne traite que les méthodes POST pour remplir $_POST et $_FILES
     * La surcharge de cette méthode permet d'utiliser par exemple PUT pour
     * envoyer un formulaire.
     * Le formulaire est parsé manuellement à partir d'un multipart/form-data
     * @param array $data Array of post data.
     * @return array
     */
    protected function _processPost($data)
    {
        $method = $this->getEnv('REQUEST_METHOD');
        if (in_array($method, ['PUT', 'DELETE', 'PATCH']) &&
            strpos($this->contentType(), 'multipart/form-data') === 0
        ) {
            return $this->parseMultipartFormData();
        }

        return parent::_processPost($data);
    }

    /**
     * Parse un multipart/form-data
     * @return array
     */
    private function parseMultipartFormData()
    {
        if (!preg_match('/boundary=(.*)$/', $this->contentType(), $matches)) {
            return [];
        }
        $boundary = $matches[1];
        $input = $this->input();
        $rawDatas = preg_split('/-*'.$boundary.'/', $input);
        $data = [];
        foreach ($rawDatas as $raw) {
            $nl = strpos($raw, "\n\r");
            if ($nl) {
                $header = trim(substr($raw, 0, $nl));
                $value = substr($raw, $nl + strlen("\n\r") +1);
                $value = substr($value, 0, strlen($value) - strlen("\n\r"));
                $this->rawDataIntoCake($header, $value, $data);
            }
        }
        return $data;
    }

    /**
     * Transforme les données brutes en habituel request->data
     * @param string $header
     * @param string $value
     * @param array $data
     */
    private function rawDataIntoCake(string $header, string $value, array &$data)
    {
        $headers = [];
        foreach (explode(';', preg_replace('/\n\s*/', '; ', $header)) as $header) {
            list($k, $v) = preg_split('/:|=/', trim($header));
            $headers[$k] = trim($v, " \t\n\r\0\x0B\"");
        }

        if (isset($headers['filename'])) {
            $file = [
                'name' => $headers['filename'],
                'size' => strlen($value),
                'type' => $headers['Content-Type'] ?? 'application/octet-stream'
            ];
            $tmpDir = sys_get_temp_dir();
            if ($file['size'] > $this->uploadMaxFilesize()) {
                $file['error'] = UPLOAD_ERR_INI_SIZE;
            } elseif (empty($file['size'])) {
                $file['error'] = UPLOAD_ERR_NO_FILE;
            } elseif (empty($tmpDir)) {
                $file['error'] = UPLOAD_ERR_NO_TMP_DIR;
            } elseif (!is_writable($tmpDir)) {
                $file['error'] = UPLOAD_ERR_CANT_WRITE;
            } else {
                $file['error'] = UPLOAD_ERR_OK;
                $tmp = tmpfile();
                $file['tmp_name'] = stream_get_meta_data($tmp)['uri'];
                fwrite($tmp, $value);
                $this->files[$file['tmp_name']] = $tmp;
            }
            $value = $file;
        }
        $data[$headers['name']] = $value;
    }

    /**
     * @var int mémorise la valeur upload_max_filesize configurée en octet
     */
    private $uploadMaxFilesize = 0;

    /**
     * Converti le upload_max_filesize en octet (ex: 100K = 102400)
     * @link http://www.php.net/manual/en/function.ini-get.php
     * @return int|string
     */
    private function uploadMaxFilesize()
    {
        if (!$this->uploadMaxFilesize) {
            $val = trim(ini_get('upload_max_filesize'));
            $last = strtolower($val[strlen($val)-1]);
            switch ($last) {
                /** @noinspection PhpMissingBreakStatementInspection */
                case 'g':
                    $val *= 1024;
                /** @noinspection PhpMissingBreakStatementInspection */
                case 'm':
                    $val *= 1024;
                case 'k':
                    $val *= 1024;
            }
            $this->uploadMaxFilesize = $val;
        }
        return $this->uploadMaxFilesize;
    }
}
