<?php
/**
 * AsalaeCore\Utility\Check
 */

namespace AsalaeCore\Utility;

use Cake\Database\Connection;
use Cake\Datasource\ConnectionManager;
use PDOException;
use Cake\Core\Configure;
use Beanstalk\Utility\CakeEmitter;

/**
 * Utilitaire de vérification de l'application
 *
 * @category    Utility
 *
 * @author      Ludovic Ganée <ludovic.ganee@libriciel.coop>
 * @copyright   (c) 2017, Libriciel
 * @license     http://www.cecill.info/licences/Licence_CeCILL_V2-fr.html
 */
abstract class Check
{

    /**
     * Fonctions shell qui doivent être installé pour un fonctionnement optimal
     *
     * @var array   la valeur --help est défini par défaut,
     *              indiquer une autre commande si besoin (doit renvoyer un code 0)
     */
    public static $prerequis = [
        'bzip2',
        'gzip',
        'mediainfo' => '--Version',
        'rar' => '',
        'sf' => '-version',
        'zip',
    ];

    /**
     * Liste les service utilisés par l'application
     *
     * @var array
     * @deprecated
     */
    public static $services = [
        'apache2',
        'beanstalkd',
        'cron',
    ];

    /**
     * Liste des modules PHP utilisés par l'application
     *
     * @var array
     */
    public static $classes = [
        'ZipArchive',
        'ZMQContext',
        'IntlDateFormatter',
    ];

    /**
     * Vérifie la connexion à la base de données
     *
     * @return boolean Vrai si la base de donnée fonctionne
     */
    public static function database(): bool
    {
        try {
            $conn = ConnectionManager::get('default');
            if ($conn instanceof Connection) {
                $conn->query('select 1');
            }
            return true;
        } catch (PDOException $e) {
            return false;
        }
    }

    /**
     * Vérifie la connexion à Beanstalkd
     *
     * @return boolean Vrai si Beanstalkd fonctionne
     */
    public static function beanstalkd(): bool
    {
        return CakeEmitter::connected();
    }

    /**
     * Vérifie la connexion à Ratchet
     *
     * @return boolean Vrai si Ratchet fonctionne
     */
    public static function ratchet(): bool
    {
        $header = @get_headers(Configure::read('Ratchet.ping'));
        return !empty($header[0]) && $header !== 'HTTP/1.1 400 Bad Request';
    }

    /**
     * Liste des logiciels non installés
     *
     * @return array d'installations manquante
     */
    public static function missings(): array
    {
        $missing = [];
        foreach (static::$prerequis as $key => $value) {
            if (is_numeric($key)) {
                $key = $value;
                $value = '--help';
            }
            exec(trim($key.' '.$value).' 2> /dev/null', $output, $code);

            if ($code == 127) {
                $missing[] = $key;
            }
        }
        return $missing;
    }

    /**
     * Vérifi la présence des classes de module PHP
     *
     * @return array
     */
    public static function missingsClasses(): array
    {
        $missing = [];
        foreach (static::$classes as $value) {
            if (!class_exists($value)) {
                $missing[] = $value;
            }
        }
        return $missing;
    }

    /**
     * Permet de savoir si il faut installer un logiciel
     *
     * @return boolean Vrai si il manque une installation
     */
    public static function needInstall(): bool
    {
        return !empty(static::missings()) || !empty(static::missingsClasses());
    }

    /**
     * Vérifi l'existance d'un service
     *
     * @param string $serviceName
     * @return boolean
     */
    public static function serviceExists(string $serviceName): bool
    {
        $exists = false;
        foreach (static::getServicesList() as $service) {
            if (preg_match('/[\s\[\]\+\-]+'.$serviceName.'$/', $service)) {
                $exists = true;
                break;
            }
        }
        return $exists;
    }

    /**
     * Garde en mémoire le résultat de la commande "service --status-all"
     * @var array
     */
    private static $servicesList;

    /**
     * Permet d'obtenir la liste des services du systeme
     * @return array
     */
    private static function getServicesList(): array
    {
        if (empty(static::$servicesList)) {
            exec('service --status-all', static::$servicesList);
        }
        return static::$servicesList;
    }

    /**
     * Vérifi qu'un service est en cours de fonctionnement
     * @param string $serviceName
     * @return null|bool    null si le service n'a pas été trouvé,
     *                      vrai ou faux selon si il est lancé ou pas
     */
    public static function serviceIsRunning(string $serviceName)
    {
        $running = null;
        foreach (static::getServicesList() as $service) {
            if (preg_match('/[\s\[\]]+([\+\-])[\s\[\]]+'.$serviceName.'$/', $service, $match)) {
                $running = $match[1] === '+';
                break;
            }
        }
        return $running;
    }
}
