<?php
/**
 * AsalaeCore\View\Helper\Object\ModalForm
 */

namespace AsalaeCore\View\Helper\Object;

use AsalaeCore\Model\Entity\SavedFilter;
use AsalaeCore\View\Helper\FilterHelper;
use Cake\Core\Configure;
use Cake\Routing\Router;
use Cake\Utility\Inflector;
use Cake\View\Helper;

/**
 * Objet crée par AsalaeCore\View\Helper\FilterHelper
 *
 * @category Helper
 *
 * @author    Ludovic Ganée <ludovic.ganee@libriciel.coop>
 * @copyright (c) 2017-2018, Libriciel
 * @license   http://www.cecill.info/licences/Licence_CeCILL_V2-fr.html
 */
class Filter implements ObjectInterface
{
    /**
     * @var string Identifiant HTML
     */
    public $id;

    /**
     * @var array Liste des filtres
     */
    public $filters = [];

    /**
     * @var array Save
     */
    public $saves = [];

    /**
     * @var array Filtres chargés
     */
    public $loadedFilter = [];

    /**
     * @var int identifiant de la sauvegarde selectionnée
     */
    public $selectedSave = 0;

    /**
     * @var array liste des urls pour les requêtes ajax
     */
    public $urls;

    /**
     * @var FilterHelper parent
     */
    private $Helper;

    /**
     * @var string id unique, fait le lien entre les modales d'un même Filter
     */
    private $uid;

    /**
     * Initialisation d'un bloc de filtres
     * @param Helper $helper
     * @param string $id
     */
    public function __construct(Helper $helper, string $id)
    {
        $this->Helper = $helper;
        $this->id = Inflector::camelize(
            preg_replace('/[\W]/', '_', $id)
        );
        $this->urls = array_map(function ($v) {
            return Router::url('/'.trim($v, "/"));
        }, Configure::read('Filters.urls', [
            'delete' => 'filters/ajax-delete-save',
            'overwrite' => 'filters/ajax-overwrite',
            'new' => 'filters/ajax-new-save',
            'load' => 'filters/ajax-get-filters',
        ]));
        $this->uid = uniqid();
    }

    /**
     * Filtres sauvegardés
     * @param array $saves
     * @return \self
     */
    public function saves(array $saves): self
    {
        $this->saves = [];
        foreach ($saves as $savedFilter) {
            $this->saves[$savedFilter->get('id')] = $savedFilter->get('name');
        }
        return $this;
    }

    /**
     * Ajout d'un filtre de recherche
     * @param string $name
     * @param array $params
     * @return \self
     * @throws \Aura\Intl\Exception
     */
    public function filter(string $name, array $params = []): self
    {
        if (empty($params['label'])) {
            $params['label'] = $name;
        }
        if (!empty($params['type']) && $params['type'] === 'date') {
            $this->appendDatepicker($name, $params);
        }
        $this->filters[$name] = $params + ['required' => true];
        return $this;
    }

    /**
     * Ajoute à un champ de type date, un datepicker
     * et une selection d'operateur
     * @param string $name
     * @param array $params
     * @throws \Aura\Intl\Exception
     */
    private function appendDatepicker(string $name, array &$params)
    {
        $options = empty($params['options'])
            ? [
                '=' => __x('date', '='),
                '<=' => __x('date', '<='),
                '>=' => __x('date', '>='),
                '<' => __x('date', '<'),
                '>' => __x('date', '>')
            ]
            : $params['options'];

        $select = $this->Helper->Html->tag('label',
            __("Opérateur de comparaison pour la date"),
            ['class' => 'sr-only', 'for' => 'dateoperator_'.$name, 'escape' => false]
        );
        $select .= $this->Helper->Html->tag('select', null, [
            'name' => 'dateoperator_'.$name,
            'id' => 'dateoperator_'.$name,
        ]);
        foreach ($options as $k => $v) {
            $select .= $this->Helper->Html->tag('option', $v, ['value' => $k]);
        }
        $select .= $this->Helper->Html->tag('/select');

        $params['class'] = empty($params['class'])
            ? 'datepicker'
            : $params['class'] . ' datepicker';

        $params += [
            'prepend' => $select,
            'append' => $this->Helper->Date->picker(".datepicker")
        ];
        unset($params['type'], $params['options']);
    }

    /**
     * Génère le code html et javascript pour les filtres de recherches
     * @return string
     * @throws \Aura\Intl\Exception
     */
    public function generateFilters(): string
    {
        $formElements = [];
        $options = [];
        foreach ($this->filters as $name => $params) {
            $formElements[$name] = $this->Helper->Form->control($name, $params);
            $options[$name] = $params['label'];
        }
        $json = json_encode($formElements);

        $selectSave = $this->Helper->Html->tag(
                'div',
                $this->Helper->Form->control(null, [
                    'label' => __('Sélectionner la sauvegarde'),
                    'value' => $this->selectedSave ?: $this->Helper->request->getParam('?.SaveFilterSelect'),
                    'type' => 'select',
                    'class' => 'filterSelect',
                    'id' => $this->id.'-select-save',
                    'options' => ['-1' => __('-- Sélectionner une sauvegarde --')] + $this->saves
                ]),
                ['class' => 'select-filter-container']
            ).PHP_EOL;
        $deleteButton = $this->Helper->Html->tag('button',
            '<i class="fa fa-trash" aria-hidden="true"></i> '
            .__("Supprimer la sauvegarde"),
            [
                'type' => 'button',
                'class' => 'btn btn-danger delete',
                'disabled' => 'disabled',
            ]
        );
        $overwriteButton = $this->Helper->Html->tag('button',
            '<i class="fa fa-floppy-o" aria-hidden="true"></i> '
            .__("Enregistrer les modifications"),
            [
                'type' => 'button',
                'class' => 'btn btn-warning overwrite',
                'disabled' => 'disabled',
            ]
        );
        $loadButton = $this->Helper->Html->tag('button',
            '<i class="fa fa-caret-square-o-down" aria-hidden="true"></i> '
            .__("Charger"),
            [
                'type' => 'button',
                'class' => 'btn btn-primary load',
                'disabled' => 'disabled',
            ]
        );
        $script = $this->Helper->Html->tag('script',
            "AsalaeFilter.modalListButtons(".json_encode($this->urls).", '"
            .$this->Helper->Html->request->getParam('controller')."', '"
            .$this->Helper->Html->request->getParam('action')
            ."');"
        );
        $buttons = $this->Helper->Html->tag('div',
            $deleteButton . $overwriteButton . $loadButton . $script,
            ['class' => 'submit-btn-filter-container modal-footer modal-list-buttons'
        ]);
        $saveSection = $this->Helper->Html->tag('div',
            $selectSave . $buttons . '<hr>',
            ['class' => 'save-section']
        );

        $select = $this->Helper->Html->tag(
            'div',
            $this->Helper->Form->control(null, [
                'label' => __('Ajouter un filtre'),
                'value' => 'null',
                'empty' => __('--- Selectionnez un filtre ---'),
                'type' => 'select',
                'class' => 'filter',
                'id' => $this->id.'-select',
                'options' => $options
            ]),
            ['class' => 'select-filter-container']
        ).PHP_EOL;
        $save = $this->Helper->Form->control('SaveFilterSelect', [
            'value' => $this->selectedSave ?: $this->Helper->request->getParam('?.SaveFilterSelect'),
            'type' => 'hidden',
            'id' => $this->id.'-savefilterselect'
        ]);

        $serializedHelper = $this->Helper->Form->control('ConfigFilter', [
            'class' => 'configuration-filter',
            'type' => 'hidden',
            'id' => $this->id.'-config',
            'value' => json_encode([
                'id' => $this->id,
                'filters' => $this->filters,
                'saves' => $this->saves,
            ])
        ]).PHP_EOL;

        $content = '';
        $this->Helper->Form->setTemplates(
            ['dateWidget' => '{{day}}{{month}}{{year}}']
        );

        $savedFilters = array_merge(
            (array)$this->Helper->request->getParam('?'),
            $this->loadedFilter
        );
        if (!empty($savedFilters)) {
            foreach ($savedFilters as $key => $values) {
                if (!isset($this->filters[$key])) {
                    continue;
                }
                $params = $this->filters[$key];
                foreach ((array)$values as $index => $value) {
                    $params['value'] = $value;
                    $content .= $this->addInput($key, $index, $params).PHP_EOL;
                }
            }
        }

        $container = $this->Helper->Html->tag(
            'div',
            $content,
            [
                'id' => $this->id.'-container',
                'class' => 'filter-container'
            ]
        ).PHP_EOL;
        $deleteBtn = !empty($content)
            ? $this->Helper->Form->button(
                '<i class="fa fa-undo" aria-hidden="true"></i> '.__('Réinitialiser les filtres actifs'),
                [
                    'type' => 'button',
                    'onclick' => 'AsalaeGlobal.confirmRemoveFilters()',
                    'class' => 'btn btn-default'
                ]
            )
            : '';
        $newSaveButton = $this->Helper->Html->tag('button',
            '<i class="fa fa-plus-circle" aria-hidden="true"></i> '
            .__("Nouvelle sauvegarde"),
            [
                'type' => 'button',
                'onclick' => 'AsalaeGlobal.saveFilters.call(this)',
                'class' => 'btn btn-success save'
            ]
        );
        $submit = !empty($content)
            ? $this->Helper->Html->tag(
                'div',
                $deleteBtn . $newSaveButton . $this->Helper->Form->button('<i class="fa fa-filter" aria-hidden="true"></i> '.__('Filtrer'), ['class' => 'btn btn-primary']),
                ['class' => 'submit-btn-filter-container modal-footer']
            )
            : '';
        $script = $this->Helper->Html->tag('script',
            "AsalaeFilter.selectHandler('#{$this->id}-select', '#{$this->id}-container', $json);
            $('#{$this->id}-form').on('change.filter', AsalaeFilter.formChangedHandler);
            "
        );

        return $save . $serializedHelper . $saveSection . $select . $container . $submit . $script;
    }

    /**
     * Génère le code html et javascript pour les filtres de recherches
     * @param array $params
     * @param array $args
     * @return string
     * @throws \Aura\Intl\Exception
     */
    public function generate(array $params = [], ...$args): string
    {
        if (isset($params['id'])) {
            $id = $params['id'];
            unset($params['id']);
        } else {
            $id = 'modal-filters';
        }
        $form = $this->Helper->Html->tag(
            'form',
            $this->generateFilters(),
            $params + [
                'novalidate',
                'class' => 'ModalFilters minifiable-target no-padding',
                'method' => 'get',
                'id' => $this->id.'-form',
                'action' => $this->Helper->Url->build([
                        'controller' => $this->Helper->request->getParam('controller'),
                        'action' => $this->Helper->request->getParam('action'),
                    ] + $this->Helper->request->getParam('pass'))
            ]
        );

        return $this->getModals($id, $form);
    }

    /**
     * Génère les modales
     * @param string $id
     * @param string $form
     * @return string
     * @throws \Aura\Intl\Exception
     */
    private function getModals(string $id, string $form)
    {
        $main = $this->Helper->Modal->create(['id' => $id, 'data-uid' => $this->uid])
            . $this->Helper->Modal->header(__('Filtres de recherche'))
            . $this->Helper->Modal->body($form)
            . $this->Helper->Modal->end()
        ;
        $save = $this->Helper->Modal->create(['id' => $this->id."NewSave", 'class' => 'filter-save-modal', 'data-uid' => $this->uid])
            . $this->Helper->Modal->header(__('Nouvelle sauvegarde de recherche'))
            . $this->Helper->Modal->body($this->newSaveModalContent())
            . $this->Helper->Modal->end()
        ;
        return $main . $save;
    }

    /**
     * Formulaire de sauvegarde des filtres
     * @return string
     * @throws \Aura\Intl\Exception
     */
    private function newSaveModalContent()
    {
        $saveName = $this->Helper->Html->tag(
                'div',
                $this->Helper->Form->control('FilterName', [
                    'label' => __('Nom de la nouvelle sauvegarde'),
                    'class' => 'filterNewSave',
                    'id' => $this->id.'-new-save',
                ]),
                ['class' => 'select-filter-save-container']
            ).PHP_EOL;
        $cancelButton = $this->Helper->Html->tag('button',
            '<i class="fa fa-times-circle" aria-hidden="true"></i> '
            .__("Annuler"),
            [
                'type' => 'button',
                'class' => 'btn btn-default cancel',
                'data-toggle' => 'modal',
                'data-target' => '#'.$this->id.'NewSave',
            ]
        );
        $newSaveButton = $this->Helper->Html->tag('button',
            '<i class="fa fa-floppy-o" aria-hidden="true"></i> '
            .__("Nouvelle sauvegarde"),
            [
                'type' => 'button',
                'data-toggle' => 'modal',
                'data-target' => '#'.$this->id.'Save',
                'class' => 'btn btn-primary save'
            ]
        );
        $script = $this->Helper->Html->tag('script',
            "AsalaeFilter.modalAddSaveButtons('".$this->urls['new']."', '"
                    .$this->Helper->Html->request->getParam('controller')."', '"
                    .$this->Helper->Html->request->getParam('action')
                ."');"
        );
        $saveDiv = $this->Helper->Html->tag('div',
            $cancelButton . $newSaveButton . $script,
            ['class' => 'submit-btn-filter-container modal-footer modal-save-buttons'
            ]);
        return $saveName . $saveDiv;
    }

    /**
     * Genère un filtre à partir des valeurs dans le request->params['?']
     * @param string $name
     * @param int $index
     * @param array $params
     * @return string
     * @throws \Aura\Intl\Exception
     */
    private function addInput(string $name, int $index, array $params): string
    {
        if ($s = strpos($name, '[')) {
            $name = substr($name, 0, $s) . '[' . $index . ']' . substr($name, $s);
        } else {
            $name = $name . '[' . $index . ']';
        }
        $close = $this->Helper->Html->tag(
            'button',
            $this->Helper->Html->tag('span', '×'),
            [
                'class' => 'close',
                'type' => 'button',
                'aria-label' => __('Retirer le filtre de recherche'),
                'onclick' => 'var p = $(this).parent(), f = p.closest("form"); p.slideUp(400, function() { p.remove(); f.change(); })'
            ]
        );

        foreach (['prepend', 'append'] as $key) {
            if (!empty($params[$key])
                && preg_match('/name=[\'"]([^\'"]+)[\'"]/', $params[$key], $match)
            ) {
                if ($s = strpos($match[1], '[')) {
                    $subname = substr($match[1], 0, $s) . '[' . $index . ']' . substr($match[1], $s);
                } else {
                    $subname = $match[1] . '[' . $index . ']';
                }
                $insert = 'name="' . $subname . '"';
                $value = isset($this->loadedFilter[$match[1]][$index])
                    ? $this->loadedFilter[$match[1]][$index]
                    : $this->Helper->request->getParam('?.'.$match[1].'.'.$index);

                // Insert la valeur
                $postScript = $value
                    ? "<script>$('[name=\"$subname\"]').last().val('".addcslashes((string)$value, "'")."')</script>"
                    : '';

                $params[$key] = str_replace($match[0], $insert, $params[$key]) . $postScript;
            }
        }

        return $this->Helper->Html->tag(
            'div',
            $close . $this->Helper->Form->control($name, $params)
        );
    }

    /**
     * Permet de charger les filtres sauvegardés en base
     * @param SavedFilter $save
     * @return self
     */
    public function loadFilters(SavedFilter $save): self
    {
        $this->selectedSave = $save->get('id');
        foreach ($save->get('filters') as $filter) {
            /** @var \AsalaeCore\Model\Entity\Filter $filter */
            if (preg_match('/^([\w]+)\[([\d]+)\](?:\[([\d]+)\])*$/', $filter->get('key'), $match)) {
                if (count($match) >= 4) {
                    $this->loadedFilter[$match[1]][$match[2]][$match[3]] = $filter->get('value');
                } else {
                    $this->loadedFilter[$match[1]][$match[2]] = $filter->get('value');
                }
            }
        }
        return $this;
    }

    /**
     * Appel self::generate() sans spécifier de paramètres
     * @return string
     * @throws \Aura\Intl\Exception
     */
    public function __toString(): string
    {
        return $this->generate();
    }
}
