<?php
/**
 * AsalaeCore\View\Helper\Object\ModalForm
 */

namespace AsalaeCore\View\Helper\Object;

use AsalaeCore\View\Helper\ModalFormHelper;
use Cake\View\Helper;
use Exception;
use Cake\Core\Configure;

/**
 * Objet crée par AsalaeCore\View\Helper\ModalFormHelper
 *
 * @category Helper
 *
 * @author    Ludovic Ganée <ludovic.ganee@libriciel.coop>
 * @copyright (c) 2017, Libriciel
 * @license   http://www.cecill.info/licences/Licence_CeCILL_V2-fr.html
 */
class ModalForm implements ObjectInterface
{
    /**
     * @var string Identifiant de la modal
     */
    public $id;

    /**
     * @var array id, class etc...
     */
    public $params;

    /**
     * @var string Ce que contient la modal
     */
    public $content = '';

    /**
     * @var bool Recharge la page lors de l'envoi du formulaire sans erreurs
     */
    public $refresh = false;

    /**
     * @var string Callback javascript, appelé après l'envoi du formulaire
     */
    public $callback = '';

    /**
     * @var string Titre de la modal
     */
    public $title = '';

    /**
     * @var array Options pour le rendu du bouton d'activation de la modale
     */
    public $output = [];

    /**
     * @var ModalFormHelper parent
     */
    private $Helper;

    /**
     * @var string Bouton annuler
     */
    public $cancelButton;

    /**
     * @var string Bouton Enregistrer
     */
    public $acceptButton;

    /**
     * ModalForm constructor.
     * @param Helper $helper
     * @param string $id
     * @param array $params
     * @throws \Aura\Intl\Exception
     */
    public function __construct(Helper $helper, string $id, array $params)
    {
        $this->Helper = $helper;
        $this->id = $id;
        $this->cancelButton = $params['cancelButton']
            ?? $this->Helper->Form->button('<i class="fa fa-times-circle-o" aria-hidden="true"></i> ' . __('Annuler'), ['data-dismiss' => 'modal', 'class' => 'cancel'])
        ;
        $this->acceptButton = $params['acceptButton']
            ?? $this->Helper->Form->button('<i class="fa fa-floppy-o" aria-hidden="true"></i> ' . __('Enregistrer'), ['bootstrap-type' => 'primary', 'class' => 'accept'])
        ;
        unset($params['cancelButton'], $params['acceptButton']);
        $this->params = $params + [
            'id' => 'modal-'.$id, // FIXME retirer le prefix
            'data-backdrop' => 'static',
            'data-keyboard' => "false",
            'class' => 'ui-draggable-handle'
        ];
    }

    /**
     * Défini le titre et le contenu de la modal
     * @param string $title
     * @param string $content
     * @return \self
     */
    public function modal(string $title, string $content = ''): self
    {
        $this->content = $content ?: $this->Helper->Html->tag(
            'div',
            $this->Helper->Html->image('ajax-loader.gif'),
            ['class' => 'text-center']
        );
        $this->title = $title;

        return $this;
    }

    /**
     * Recharge la page en cas de succès
     * @param bool $refresh
     * @return \self
     */
    public function refreshOnSuccess(bool $refresh = true): self
    {
        $this->refresh = $refresh;
        return $this;
    }

    /**
     * Défini le callback javascript lié à la modal
     * Envoi les attributs javascript suivant :
     * (content, textStatus, jqXHR)
     * @param string $funcName
     * @return \self
     * @throws Exception
     */
    public function javascriptCallback(string $funcName): self
    {
        if (strpos($funcName, '(')) {
            throw new Exception(__("Le callback sera appelé avec les arguments suivants : (content, textStatus, jqXHR), ne pas indiquer de parentèses"));
        }
        $this->callback = $funcName . '(content, textStatus, jqXHR);';
        return $this;
    }

    /**
     * Génère un fonction javacript qui pourra être appelé pour ouvrir la modal
     * @param string $function
     * @param string|array $url
     * @return string
     * @throws \Aura\Intl\Exception
     */
    public function generateFunction(string $function, $url): string
    {
        $modalId = $this->params['id'];
        $refresh = $this->refresh ? 'location.reload();' : '';
        $callback = $this->callback ?: '';

        $script = <<<EOT
<script>
    function $function() {
        var id = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : '';
        AsalaeModal.append($('#$modalId').modal('show'));
        {$this->javascriptFunctionContent($this->Helper->Url->build($url), $modalId, $refresh, $callback, '+id')}
    }
</script>
EOT;
        return $this->getModal().$script;
    }

    /**
     * Javascript de la modal
     * @param string $url
     * @param string $modalId
     * @param string $refresh
     * @param string $callback
     * @param string $id
     * @return string
     * @throws \Aura\Intl\Exception
     */
    private function javascriptFunctionContent(string $url, string $modalId, string $refresh, string $callback, string $id = ''): string
    {
        $debug = '';
        if (Configure::read('debug')) {
            $debug = "default: $('#$modalId div.modal-body').find('#debug-output').remove();".PHP_EOL
                . "$('#$modalId div.modal-body').append('<div id=\"debug-output\" class=\"danger\"><h2>Debug output</h2>'+content+'</div>');";
        }
        $content = addcslashes($this->content, "'");
        $message = __("Vous êtes sur le point de quitter la page pendant la saisie d'un formulaire, si vous continuer les données saisie seront perdu. Voulez vous continuer ?");
        $msgError = __("Une erreur inattendue s'est produite, quitter le formulaire ? (entrainera la perte des données saisies)");
        $confirmCancel = __("Souhaitez-vous fermer le formulaire ? (vous perdrez les informations saisies)");
        $uFunction = 'modalFormSubmit'.uniqid();
        return <<<EOT
window.onbeforeunload = function(e) {
            return e.returnValue = "$message";
        };
        var $uFunction = function(event) {
            var form = $(this).get(0);
            event.preventDefault();
            if (form.checkValidity()) {
                $(form).removeClass('display-validation');
                AsalaeGlobal.addWaitingAjaxResponse('#$modalId');
                $(this).find('button').each(function() {
                    var isDisabled = $(this).prop('disabled');
                    if ($(this).prop('disabled')) {
                        $(this).data('do-not-enable-after-submit', true);
                    } else {
                        $(this).disable();            
                    }
                });
                $.ajax({
                    url: '$url/'$id,
                    type: 'POST',
                    data: $('#$modalId').find('form').serialize(),
                    success: function(content, textStatus, jqXHR) {
                        $('#$modalId')
                            .trigger('ajax.success', jqXHR)
                            .find('div.modal-body')
                            .removeClass('alert')
                            .removeClass('alert-danger')
                        switch (jqXHR.getResponseHeader('X-Asalae-Success')) {
                            case 'true':
                                $('#$modalId').find('button:disabled').each(function() {
                                    $(this).data('do-not-enable-after-submit', true);
                                });
                                $('#$modalId').modal("hide");
                                window.onbeforeunload = null;
                                $refresh
                                break;
                            case 'false':
                                $('#$modalId').find('div.modal-body')
                                    .find('*').off().remove().end()
                                    .html(content);
                                $('#$modalId').find('form').off().submit($uFunction);
                                break;
                            $debug
                        }
                        $callback
                    },
                    error: function() {
                        $('#$modalId').trigger('ajax.error');
                        if (confirm("$msgError")) {
                            $('#$modalId').modal("hide");
                        }
                    },
                    complete: function() {
                        var hiddenSubmit = $('<button type="submit" class="hide">Hidden submit</button>')
                            .on('click', function() {
                                $(this).closest('form').submit();                        
                            })
                        ; 
                        AsalaeGlobal.removeWaitingAjaxResponse('#$modalId');
                        $('#$modalId')
                            .trigger('ajax.complete')
                            .find('form')
                            .find('button[type="submit"], input[type="submit"]')
                            .remove().end()
                            .append(hiddenSubmit)
                            .find('button:disabled').each(function() {
                                if ($(this).data('do-not-enable-after-submit')) {
                                    $(this).data('do-not-enable-after-submit', false);
                                } else {
                                    $(this).enable();
                                }
                            })
                        ;
                    }
                });
            } else {
                var invalid = $(form).find('input:invalid, select:invalid, textarea:invalid').first(); 
                if (invalid.is('visible') === false) {
                    $('[href="#'+invalid.closest('.ui-tabs-panel').attr('id')+'"]').click();
                }
                form.reportValidity();
                $(form).addClass('display-validation');
            }
        };
        $('#$modalId').find('div.modal-body')
            .find('*').off().remove().end()
            .html('{$content}');
        $.ajax({
            url: '$url/'$id,
            success: function(content, textStatus, jqXHR) {
                $('#$modalId')
                    .find('div.modal-body')
                    .removeClass('alert')
                    .removeClass('alert-danger')
                    .html(content);
                $('#$modalId').trigger('ajax.success', jqXHR)
            },
            error: function(error) {
                AsalaeModal.error(error, $('#$modalId')); 
                $('#$modalId').trigger('ajax.error');
            },
            complete: function() {
                var hiddenSubmit = $('<button type="submit" class="hide">Hidden submit</button>');
                $('#$modalId')
                    .find('form')
                    .off()
                    .submit($uFunction); 
                AsalaeGlobal.removeWaitingAjaxResponse('#$modalId');
                $('#$modalId').find('form')
                    .find('button[type="submit"], input[type="submit"]')
                    .remove().end()
                    .append(hiddenSubmit)
                ;
                $('#$modalId').trigger('ajax.complete');
            }
        });

        $('#$modalId').find('div.modal-footer button.cancel').off().click(function(eve) {
            window.onbeforeunload = null;
        }).on('keydown', function(event) {
            if (event.keyCode === 13) {
                return confirm('$confirmCancel');
            }
        });
        
        $('#$modalId').find('div.modal-footer button.accept').off().click(function(event) {
            event.preventDefault();
            $('#$modalId').find('form').submit();
        });
EOT;
    }

    /**
     * Permet de définir de quel type sera la sortie HTML
     * ex: un lien, un bouton, une fonction...
     * @param string $type
     * @param string $text
     * @param string|array $url
     * @return \self
     */
    public function output(string $type, string $text, $url): self
    {
        $this->output = compact('type', 'text', 'url');
        return $this;
    }

    /**
     * Génère la modal à partir des options défini précedement
     * @param array $params
     * @param array $args
     * @return string
     * @throws Exception
     */
    public function generate(array $params = [], ...$args): string
    {
        if (empty($this->output)) {
            throw new Exception(__("Vous devez faire appel à la méthode ModalForm::output() avant d'utiliser ModalForm::generate()."));
        }
        if ($this->output['type'] === 'function') {
            return $this->generateFunction($this->output['text'], $this->output['url']);
        }

        return $this->getModal().$this->getButton($params);
    }

    /**
     * Génère le boutton, à utiliser avec $this->getModal()
     * @param array $params
     * @return string
     * @throws Exception
     */
    public function getButton(array $params = []): string
    {
        if (empty($this->output)) {
            throw new Exception(__("Vous devez faire appel à la méthode ModalForm::output() avant d'utiliser ModalForm::getButton()."));
        }
        $modalId = $this->params['id'];
        $btnId = $this->output['type'].'-'.$this->id;
        $btn = $this->Helper->Html->tag($this->output['type'], $this->output['text'], $params + [
                'id' => $btnId,
                'class' => 'btn btn-primary',
        ]);
        $refresh = $this->refresh ? 'location.reload();' : '';
        $callback = $this->callback ?: '';

        // TODO remplacer le click() par un on('click.namespace')
        $script = <<<EOT
<script>
    $('#$btnId').click(function() {
        AsalaeModal.append($('#$modalId').modal('show'));
        {$this->javascriptFunctionContent($this->Helper->Url->build($this->output['url']), $modalId, $refresh, $callback)}
    });
</script>
EOT;
        return $btn.$script;
    }

    /**
     * Construit la modal
     * @return string
     */
    public function getModal(): string
    {
        return $this->Helper->Modal->create($this->params)
            . $this->Helper->Modal->header($this->title, ['close' => false])
            . $this->Helper->Modal->body($this->content)
            . $this->Helper->Modal->end([
                $this->cancelButton,
                $this->acceptButton
            ])
            . $this->Helper->modalScript($this->params['id'])
        ;
    }

    /**
     * Appel self::generate() sans spécifier de paramètres
     * @return string
     * @throws Exception
     */
    public function __toString(): string
    {
        return $this->generate();
    }
}
