<?php
/**
 * AsalaeCore\View\Helper\Object\ModalView
 */

namespace AsalaeCore\View\Helper\Object;

use AsalaeCore\View\Helper\ModalViewHelper;
use Cake\View\Helper;
use Exception;

/**
 * Objet crée par AsalaeCore\View\Helper\ModalViewHelper
 *
 * @category Helper
 *
 * @author    Ludovic Ganée <ludovic.ganee@libriciel.coop>
 * @copyright (c) 2017, Libriciel
 * @license   http://www.cecill.info/licences/Licence_CeCILL_V2-fr.html
 */
class ModalView implements ObjectInterface
{
    /**
     * @var string Identifiant de la modal
     */
    public $id;

    /**
     * @var array id, class etc...
     */
    public $params;

    /**
     * @var string Ce que contient la modal
     */
    public $content = '';

    /**
     * @var string Callback javascript, appelé après l'envoi du formulaire
     */
    public $callback = '';

    /**
     * @var string Titre de la modal
     */
    public $title = '';

    /**
     * @var array Options pour le rendu du bouton d'activation de la modale
     */
    public $output = [];

    /**
     * @var ModalViewHelper parent
     */
    private $Helper;

    /**
     * ModalView constructor.
     * @param Helper $helper
     * @param string $id
     * @param array $params
     */
    public function __construct(Helper $helper, string $id, array $params)
    {
        $this->Helper = $helper;
        $this->id = $id;
        $this->params = $params + [
            'id' => $id,
            'class' => 'ui-draggable-handle'
        ];
    }

    /**
     * Défini le titre et le contenu de la modal
     * @param string $title
     * @param string $content
     * @return \self
     */
    public function modal(string $title, string $content = ''): self
    {
        $this->content = $content ?: $this->Helper->Html->tag(
            'div',
            $this->Helper->Html->image('ajax-loader.gif'),
            ['class' => 'text-center']
        );
        $this->title = $title;

        return $this;
    }

    /**
     * Défini le callback javascript lié à la modal
     * Envoi les attributs javascript suivant :
     * (content, textStatus, jqXHR)
     * @param string $funcName
     * @return \self
     * @throws Exception
     */
    public function javascriptCallback(string $funcName): self
    {
        if (strpos($funcName, '(')) {
            throw new Exception(__("Le callback sera appelé avec les arguments suivants : (content, textStatus, jqXHR), ne pas indiquer de parentèses"));
        }
        $this->callback = $funcName . '(content, textStatus, jqXHR);';
        return $this;
    }

    /**
     * Génère un fonction javacript qui pourra être appelé pour ouvrir la modal
     * @param string $function
     * @param string|array $url
     * @return string
     */
    public function generateFunction(string $function, $url): string
    {
        $modalId = $this->params['id'];
        $callback = $this->callback ?: '';

        $script = <<<EOT
<script>
    $function = function() {
        var id = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : '';
        AsalaeModal.append($('#$modalId').modal('show'));
        {$this->javascriptFunctionContent($this->Helper->Url->build($url), $modalId, $callback, '+id')}
    }
</script>
EOT;
        return $this->getModal().$script;
    }

    /**
     * Javascript de la modal
     * @param string $url
     * @param string $modalId
     * @param string $callback
     * @param string $id
     * @return string
     */
    private function javascriptFunctionContent(string $url, string $modalId, string $callback, string $id = ''): string
    {
        $content = addcslashes($this->content, "'");
        return <<<EOT
        $('#$modalId').find('div.modal-body')
            .find('*').off().remove().end()
            .html('{$content}');
        $.ajax({
            url: '$url/'$id,
            success: function(content) {
                $('#$modalId').find('div.modal-body')
                    .html(content)
                ;
                $callback
            },
            error: function(error) {
                AsalaeModal.error(error, $('#$modalId')); 
            },
            complete: function() {
                AsalaeGlobal.removeWaitingAjaxResponse('#$modalId');
            }
        });
EOT;
    }

    /**
     * Permet de définir de quel type sera la sortie HTML
     * ex: un lien, un bouton, une fonction...
     * @param string $type
     * @param string $text
     * @param string|array $url
     * @return \self
     */
    public function output(string $type, string $text, $url): self
    {
        $this->output = compact('type', 'text', 'url');
        return $this;
    }

    /**
     * Génère la modal à partir des options défini précedement
     * @param array $params
     * @param array $args
     * @return string
     * @throws Exception
     */
    public function generate(array $params = [], ...$args): string
    {
        if (empty($this->output)) {
            throw new Exception(__("Vous devez faire appel à la méthode ModalView::output() avant d'utiliser ModalView::generate()."));
        }
        if ($this->output['type'] === 'function') {
            return $this->generateFunction($this->output['text'], $this->output['url']);
        }

        return $this->getModal().$this->getButton($params);
    }

    /**
     * Génère le boutton, à utiliser avec $this->getModal()
     * @param array $params
     * @return string
     * @throws Exception
     */
    public function getButton(array $params = []): string
    {
        if (empty($this->output)) {
            throw new Exception(__("Vous devez faire appel à la méthode ModalView::output() avant d'utiliser ModalView::getButton()."));
        }
        $modalId = $this->params['id'];
        $btnId = $this->output['type'].'-'.$this->id;
        $btn = $this->Helper->Html->tag($this->output['type'], $this->output['text'], $params + [
                'id' => $btnId,
                'class' => 'btn btn-primary',
            ]);
        $callback = $this->callback ?: '';

        // TODO remplacer le click() par un on('click.namespace')
        $script = <<<EOT
<script>
    $('#$btnId').click(function() {
        AsalaeModal.append($('#$modalId').modal('show'));
        {$this->javascriptFunctionContent($this->Helper->Url->build($this->output['url']), $modalId, $callback)}
    });
</script>
EOT;
        return $btn.$script;
    }

    /**
     * Construit la modal
     * @return string
     */
    public function getModal(): string
    {
        return $this->Helper->Modal->create($this->params)
            . $this->Helper->Modal->header($this->title)
            . $this->Helper->Modal->body($this->content)
            . $this->Helper->Modal->end()
            . $this->Helper->modalScript($this->params['id'])
        ;
    }

    /**
     * Appel self::generate() sans spécifier de paramètres
     * @return string
     * @throws Exception
     */
    public function __toString(): string
    {
        return $this->generate();
    }
}
