<?php
/**
 * AsalaeCore\View\Helper\Object\Table
 */

namespace AsalaeCore\View\Helper\Object;

use AsalaeCore\View\Helper\TableHelper;
use Cake\I18n\I18n;
use Cake\Log\Log;
use Cake\ORM\ResultSet;
use Cake\Utility\Inflector;
use Cake\View\Helper;
use Cake\ORM\Entity;
use Cake\Utility\Hash;
use Exception;

/**
 * Objet crée par AsalaeCore\View\Helper\TableHelper
 *
 * @category Helper
 *
 * @author    Ludovic Ganée <ludovic.ganee@libriciel.coop>
 * @copyright (c) 2017, Libriciel
 * @license   http://www.cecill.info/licences/Licence_CeCILL_V2-fr.html
 */
class Table implements ObjectInterface
{
    /**
     * @var string Contient le HTML de la table en construction
     */
    private $html;

    /**
     * @var string Contient l'id de la table
     */
    private $id;

    /**
     * @var array Contient les champs
     */
    private $fields = [];

    /**
     * @var array Contient les données
     */
    private $data = [];

    /**
     * @var array Contient les actions
     */
    private $actions = [];

    /**
     * @var string nom de l'objet javascript
     */
    public $tableObject = '';

    /**
     * @var array Paramètres du tableau
     */
    public $params = [];

    /**
     * @var TableHelper parent
     */
    private $Helper;

    /**
     * Initialise la création d'un tableau HTML
     * @param  Helper $helper
     * @param  string $id
     * @param  array $params
     * @throws Exception
     */
    public function __construct(Helper $helper, string $id, array $params = [])
    {
        if (empty($id)) {
            throw new Exception(__('Vous devez saisir un identifiant pour le tableau'));
        }

        $this->Helper = $helper;
        $tableVar = Inflector::camelize(
            'generated_'.preg_replace('/[^\w]/', '_', $id)
        );
        $this->tableObject = $tableVar;
        $params['id'] = $id;
        $params += [
            'class' => $this->Helper->defaultCssClass,
            'data-for' => $tableVar
        ];

        $this->id = $id;
        $this->html = $this->Helper->Html->tag('table', '', $params);
    }

    /**
     * Récupère le HTML de l'objet
     * @param array $params
     * @param array $args
     * @return string
     * @throws Exception
     */
    public function generate(array $params = [], ...$args): string
    {
        $this->isCompleteOrDie();
        $paramsJson = $this->getParamsJson();
        $tableVar = $this->tableObject;

        $actions = $this->actions
            ? json_encode($this->formatActions(), $this->Helper->jsonStyle)
            : 'null';

        $messages = json_encode(
            [
                'favorite' => __('Ajouter/supprimer des favoris'),
                'confirmSwitchView' => __('Voulez vous enregistrer vos modifications ?'),
                'confirmChangeMaxPerPage' => __('Vous devez recharger la page pour prendre en compte toutes les modifications. Voulez-vous le faire maintenant ?'),
                'selectCheckbox' => __("Select"),
                'selectAllCheckboxes' => __("Select all")
            ],
            $this->Helper->jsonStyle
        );


        $local = I18n::getLocale();
        if (strpos($local, '_')) {
            /** @noinspection PhpUnusedLocalVariableInspection */
            list($lang, $territory) = explode('_', I18n::getLocale());
        } else {
            $lang = $local;
        }

        $javascript = <<<EOD
            <script>
                TableHelper.locale('$lang');
                var $tableVar = new TableGenerator($('#{$this->id}'));
                $tableVar.messages = $messages;

                (function() {
                    var data = {$this->data};
                    $tableVar.actions = $actions;

                    if ($tableVar.actions) {
                        data = TableGenerator.appendActions(data, $tableVar.actions);
                    }
                    $tableVar.data = data;
                })();

                $tableVar.params($paramsJson);
                $tableVar.applyCookies();
                $tableVar.generateAll();
                $tableVar.insertConfiguration('#modal-{$this->id}');
            </script>
EOD;

        return $this->html . $javascript;
    }


    /**
     * Renseigne les champs utilisable
     * @param  array $fields
     * @return \self
     * @throws \Aura\Intl\Exception
     */
    public function fields(array $fields): self
    {
        $this->fields = Hash::normalize($fields);

        foreach ($this->fields as $fieldname => $params) {
            $field = $this->fieldParams($fieldname, (array)$params);
            if (!empty($field['order'])) {
                if ($field['order'] === true) {
                    $field['order'] = $fieldname;
                }
                $label = $field['label'];
                $sr = $this->Helper->Html->tag('span', __("Trier par").' ', ['class' => 'sr-only']);
                $field['label'] =
                    $this->Helper->Paginator->sort(
                        $field['order'],
                        $sr . $label,
                        [
                            'asc' => ['title' => __("Trier par ''{0}'' par ordre croissant", $label)],
                            'desc' => ['title' => __("Trier par ''{0}'' par ordre décroissant", $label)],
                            'escape' => false
                        ]
                    )
                ;
            }
            $this->fields[$fieldname] = $field;
        }

        return $this;
    }

    /**
     * Permet d'obtenir les paramètres d'un field
     * @param  string $fieldname
     * @param  array $params
     * @return array
     * @throws \Aura\Intl\Exception
     */
    private function fieldParams(string $fieldname, array $params): array
    {
        if (!isset($params['label'])) {
            $params['label'] = $this->translate($fieldname);
        }
        if (isset($params['type'])) {
            switch ($params['type']) {
                case 'date':
                    unset($params['type']);
                    $params['callback'] = 'TableHelper.date("LL")';
                    break;
                case 'datetime':
                    unset($params['type']);
                    $params['callback'] = 'TableHelper.date("LLL")';
                    break;
                case 'bool':
                case 'boolean':
                    unset($params['type']);
                    $params['callback'] = 'TableHelper.boolean("'.__("Oui").'", "'.__('Non').'")';
                    break;
            }
        }
        if (isset($params['thead'])) {
            $params['thead'] = Hash::normalize($params['thead']);
            foreach ($params['thead'] as $subFieldname => $subParams) {
                $params['thead'][$subFieldname] =
                    $this->fieldValue($subFieldname, (array)$subParams);
            }
        }
        if (!isset($params['title'])) {
            $params['title'] = $params['label'];
        }

        return $params;
    }

    /**
     * Assure une utilisation de la méthode toArray() sur toutes les profondeurs
     * sans modification de la structure de données
     * @param mixed $data
     */
    private function dataToArray(&$data)
    {
        if ($data instanceof ResultSet) {
            $newData = [];
            foreach ($data as $value) {
                if ($value instanceof Entity) {
                    $newData[][$value->getSource()] = $value->toArray();
                }
            }
            $data = $newData;
        }
        foreach ($data as $key => $value) {
            if ($value instanceof Entity) {
                $source = $value->getSource();
                $value = $value->toArray();
                if (is_numeric($key)) {
                    $data[$key][$source] = $value;
                } else {
                    $data[$key] = $value;
                }
                $this->dataToArray($data[$key]);
            } elseif (is_array($value) || $value instanceof \Traversable) {
                $this->dataToArray($data[$key]);
            }
        }
    }

    /**
     * Renseigne les champs utilisable
     * @param  mixed $data
     * @return \self
     */
    public function data($data): self
    {
        $this->dataToArray($data);
        $this->data = json_encode($data, $this->Helper->jsonStyle);

        return $this;
    }

    /**
     * Permet de formatter les données lorsque elles sont représenté par
     * une entité
     * @param Entity $entity
     * @param array  $output
     */
    private function recursiveEntityFormater(Entity $entity, array &$output)
    {
        $output[$entity->getSource()] = [];

        foreach ($entity->toArray() as $fieldname => $value) {
            if ($entity->$fieldname instanceof Entity) {
                $this->recursiveEntityFormater($entity->$fieldname, $output);
            } elseif (!preg_match('/^\[.*\]$/', $fieldname)) {
                $output[$entity->getSource()][$fieldname] = $value;
            }
        }
    }

    /**
     * Paramètres du tableau
     * @param  array $params
     * @return self
     * @throws Exception
     */
    public function params(array $params): self
    {
        if (!Hash::get($params, 'identifier')) {
            throw new Exception(
                __("Impossible de trouver la paramètre 'identifier'.")
            );
        }
        $this->params = $params;

        return $this;
    }

    /**
     * Permet d'insérer des actions sur le tableau de résultat
     * @param  array $actions
     * @return \self
     */
    public function actions(array $actions)
    {
        $this->actions = $actions;
        return $this;
    }

    /**
     * Permet d'obtenir le lien pour la configuration du tableau
     * @param array $params Pour configurer le tag "A"
     * @return string
     * @throws \Aura\Intl\Exception
     */
    public function getConfigureLink(array $params = array()): string
    {
        $params += [
            'class' => 'float-right btn-link',
            'data-toggle' => 'modal',
            'data-target' => '#modal-'.$this->id,
            'content' => $this->Helper->Html->tag('i', '', ['class' => 'fa fa-sliders', 'title' => __("Paramétrage du tableau de résultat")])
                .$this->Helper->Html->tag('span', __("Paramétrage du tableau de résultat"), ['class' => 'sr-only'])
        ];
        $content = $params['content'];
        unset($params['content']);
        return $this->Helper->Html->tag('button', $content, $params);
    }

    /**
     * Permet d'obtenir le json de paramètre à passer au javascript
     * @return string json
     * @throws \Aura\Intl\Exception
     */
    public function getParamsJson(): string
    {
        $paramsJs = [];

        foreach ($this->fields as $fieldname => $params) {
            $params = (array)$params;
            $paramsJs['thead'][0][] = $this->fieldValue($fieldname, $params);
        }

        if ($this->actions) {
            $paramsJs['thead'][0][] = [
                '_id' => 'actions',
                'label' => __('Actions'),
                'title' => __('Actions'),
                'link' => false,
                'type' => 'th',
                'data-view' => 'column',
                'display' => true,
                'colspan' => 1,
                'target' => 'actions',
                'thead' => $this->actionsToThead(),
                'callback' => null,
                'class' => 'action'
            ];
        }

        $paramsJs['tbody'] = $this->params;
        return json_encode($paramsJs, $this->Helper->jsonStyle);
    }

    /**
     * Donne un champ avec tout ses paramètres
     *
     * '_id' => identifiant unique de l'objet
     * 'label' => Nom qui sera affiché à l'utilisateur
     * 'title' => attribut title de l'objet
     * 'link' => url qui sera inséré dans le 'th' pour le tri sur colonne
     * 'type' => 'th' ou 'td' pour la colonne dans le thead
     * 'data-view' => default, div ou column pour le basculement d'une vue à
     *                l'autre
     * 'display' => colonne affiché ou non par défaut
     * 'colspan' => attribut colspan de la colonne
     * 'target' => Pointeur vers la donnée dans "$results[index]"
     *              ex: Monmodel.field
     * 'thead' => Si la colonne en contient d'autres, les informations seront
     *            contenues ici
     * 'callback' => function javascript qui reçois la valeur de la colonne en
     *               paramètre et renvoi la valeur modifié
     * @param  string $fieldname
     * @param  array $params
     * @return array
     * @throws \Aura\Intl\Exception
     */
    private function fieldValue(string $fieldname, array $params): array
    {
        $display = Hash::get($params, 'display') === false
        || Hash::get($params, 'display') === 'none'
            ? false
            : true;
        $label = !isset($params['label']) ? $fieldname : $params['label'];

        $value = [
            '_id' => preg_replace('/[\W]/', '_', $fieldname),
            'label' => $label,
            'title' => Hash::get($params, 'title') ?: $label,
            'link' => Hash::get($params, 'link') ?: false,
            'type' => Hash::get($params, 'type') ?: 'th',
            'data-view' => Hash::get($params, 'data-view') ?: 'default',
            'display' => $display,
            'colspan' => Hash::get($params, 'colspan') ?: 1,
            'target' => isset($params['target']) ? $params['target'] : $fieldname,
            'thead' => Hash::get($params, 'thead'),
            'callback' => Hash::get($params, 'callback'),
            'style' => Hash::get($params, 'style'),
            'class' => Hash::get($params, 'class'),
        ];

        if ($value['thead']) {
            $value['thead'] = Hash::normalize((array)$value['thead']);
            foreach ($value['thead'] as $subfield => $subparams) {
                $value['thead'][$subfield] =
                    $this->fieldParams($subfield, (array)$subparams);
            }
        }

        return $value;
    }

    /**
     * Vérifi que toutes les méthodes obligatoire ont été appelés
     * @return bool true
     * @throws Exception
     */
    private function isCompleteOrDie(): bool
    {
        if (empty($this->fields)) {
            throw new Exception(__('Il faut appeler la méthode {0}() du helper.', 'field'));
        } elseif (empty($this->data)) {
            throw new Exception(__('Il faut appeler la méthode {0}() du helper.', 'data'));
        } elseif (empty($this->params)) {
            throw new Exception(__('Il faut appeler la méthode {0}() du helper.', 'params'));
        } elseif (empty($this->html)) {
            throw new Exception(__('Il faut appeler la méthode {0}() du helper.', 'create'));
        }

        return true;
    }

    /**
     * Transforme les actions en json exploitable par le javascript
     * @return array
     */
    private function actionsToThead(): array
    {
        $thead = [];
        foreach ($this->formatActions() as $actionName => $action) {
            $display = Hash::get($action, 'display') === false
            || Hash::get($action, 'display') === 'none'
                ? false
                : true;

            $thead[$actionName] = [
                '_id' => $actionName,
                'label' => false,
                'title' => Hash::get($action, 'title') ?: $actionName,
                'link' => false,
                'type' => 'th',
                'data-view' => 'default',
                'display' => $display,
                'colspan' => 1,
                'target' => $actionName,
                'thead' => null,
                'callback' => Hash::get($action, 'callback'),
            ];
        }

        return $thead;
    }

    /**
     * Assure que les actions ont un identifiant
     * @return array
     */
    private function formatActions(): array
    {
        $actions = [];
        foreach ($this->actions as $key => $action) {
            $actionName = is_numeric($key) ? 'action_'.$key : $key;
            $actions[$actionName] = $action;
        }
        return $actions;
    }

    /**
     * Fourni la traduction
     * @param string $name
     * @return string
     * @throws \Aura\Intl\Exception
     */
    private function translate(string $name): string
    {
        switch ($name) {
            case '':
                return __('');
        }
        Log::write(LOG_INFO, 'default::'.$name);
        return I18n::getTranslator()->translate($name);
    }

    /**
     * Appel self::generate() sans spécifier de paramètres
     * @return string
     * @throws Exception
     */
    public function __toString(): string
    {
        return $this->generate();
    }
}
