<?php
/**
 * Table beanstalk_jobs
 *
 * Défini la validation et assure la synchronisation avec le serveur beanstalk.
 *
 * @category    Beanstalk
 *
 * @author      Ludovic Ganée <ludovic.ganee@libriciel.coop>
 * @copyright   (c) 2017, Libriciel
 * @license     http://www.cecill.info/licences/Licence_CeCILL_V2-fr.html
 */

namespace Beanstalk\Model\Table;

use Beanstalk\Model\Entity\BeanstalkJob;
use Beanstalk\Utility\CakeEmitter;
use Cake\Event\Event;
use Cake\ORM\Table;
use Cake\Utility\Hash;
use Cake\Validation\Validator;
use Exception;

/**
 * Classe de la table
 *
 * Si un job n'existe plus sur le serveur Beantalk, il est automatiquement
 * effacé de la base de donnée grâce à la méthode <b>sync()</b>.
 *
 * <b>saveAndEmit()</b> est à utiliser à la place de <b>save()</b> car en plus
 * de créer une ligne dans la table, elle envoie l'information au serveur
 * Beanstalk.
 */
class BeanstalkJobsTable extends Table
{
    /**
     * Configuration initale de la table
     *
     * @param array $config
     */
    public function initialize(array $config)
    {
        $this->addBehavior('Timestamp');
    }

    /**
     * Default validation rules.
     *
     * @param \Cake\Validation\Validator $validator Validator instance.
     * @return \Cake\Validation\Validator
     */
    public function validationDefault(Validator $validator)
    {
        $validator
            ->integer('id')
            ->allowEmpty('id', 'create');

        $validator
            ->integer('jobid')
            ->allowEmpty('jobid', 'create');

        $validator
            ->add('tube', 'validFormat', [
                'rule' => [
                    'custom',
                    "/^[a-z0-9\+\/;\.\$\(\)][a-z0-9\+\/;\.\$\(\)\-]*$/i"
                ],
                'message' => "Format de text invalid"
            ])
            ->allowEmpty('tube'); // default: "default"

        $validator
            ->notEmpty('data');

        $validator
            ->integer('priority')
            // @see vendor/pda/pheanstalk/doc/protocol-1.3.txt
            ->range('priority', [0, 4294967295])
            ->allowEmpty('priority');

        $validator
            ->integer('delay')
            ->nonNegativeInteger('delay')
            ->allowEmpty('delay');

        $validator
            ->integer('ttr')
            ->nonNegativeInteger('ttr')
            ->allowEmpty('ttr');

        return $validator;
    }

    /**
     * Surcharge la sauvegarde afin de communiquer les informations à Beanstalkd
     *
     * @param array $data
     * @param array $options
     * @return boolean|BeanstalkJob
     * @throws Exception
     */
    public function saveAndEmit($data, $options = array())
    {
        $data += [
            'tube' => 'default',
            'priority' => 1024,
            'delay' => 0,
            'ttr' => 60,
            'user_id' => Hash::get($data, 'data.user_id')
        ];
        $entity = $this->newEntity($data);

        if (CakeEmitter::connected() && !$entity->getErrors()) {
            $entity->set('jobid', CakeEmitter::emitNoSave(
                $data['tube'],
                $data['data'],
                $data['priority'],
                $data['delay'],
                $data['ttr']));

            return $this->save($entity, $options);
        }

        return false;
    }

    /**
     * La table est t-elle synchronisée ?
     *
     * @var boolean
     */
    public $sync = false;

    /**
     * A lancer avant un find
     *
     * @param \Cake\Event\Event $Event
     */
    public function beforeFind(/** @noinspection PhpUnusedParameterInspection */Event $Event)
    {
        if (!$this->sync) {
            $this->sync = true;
            $this->sync();
        }
    }

    /**
     * Synchronise la base de donnée avec Beanstalk
     *
     * @return boolean
     */
    public function sync()
    {
        if (!CakeEmitter::connected()) {
            return false;
        }

        $query = $this->find()->select(['id', 'jobid']);
        foreach ($query as $entity) {
            try {
                CakeEmitter::getEmitter()
                    ->Pheanstalk->statsJob($entity->jobid);
            } catch (Exception $ex) {
                $this->delete($entity);
            }
        }

        return true;
    }

    /**
     * Options disponnibles pour le champ selectionné
     *
     * @param string $field
     * @return array
     */
    public function options($field)
    {
        if ($field === 'state') {
            return [
                'ready' => 'Prêt',
                'buried' => 'En pause',
                'delayed' => 'En attente',
                'reserved' => 'En cours'
            ];
        }
        return [];
    }
}
