<?php
/**
 * Beanstalk\Shell\AbstractWorker
 */

namespace Beanstalk\Shell;

use Beanstalk\Exception\CantWorkException;
use Beanstalk\lib\Worker;
use Exception;
use ReflectionClass;

/**
 * Squelette d'un worker
 *
 * Pour créer un worker standard, il est utile d'étendre cette classe
 *
 * @category    Beanstalk
 *
 * @author      Ludovic Ganée <ludovic.ganee@libriciel.coop>
 * @copyright   (c) 2017, Libriciel
 * @license     http://www.cecill.info/licences/Licence_CeCILL_V2-fr.html
 */
abstract class AbstractWorker implements WorkerInterface
{
    /**
     * Permet de maintenir le worker en vie après son travail
     *
     * @var boolean
     */
    public $keepAlive;

    /**
     * @var string Titre/nom public du worker
     */
    public $title;

    /**
     * Paramètres additionnels
     *
     * @var array
     */
    public $params;

    /**
     * Instance du worker
     *
     * @var Worker
     */
    protected $Worker;

    /**
     * Constructeur de classe
     *
     * @param Worker $Worker
     * @param boolean $keepAlive
     * @param array $params
     */
    public function __construct(Worker $Worker,
        $keepAlive = true,
        array $params = array())
    {
        $this->Worker = $Worker;
        $this->keepAlive = $keepAlive;
        $this->params = $params;
        $this->title = $this->title ?: (new ReflectionClass($this))->getShortName();
    }

    /**
     * {@inheritdoc}
     */
    public function start()
    {
        $data = [];
        $this->log("Starting ".get_called_class());
        while ($this->Worker->getNext()) {
            $this->log("New job reserved");
            try {
                $data = $this->Worker->getData();
                if ($data === 'shutdown') {
                    $this->Worker->done();
                    $this->shutdown();
                } elseif ($data === 'getprerequisites') {
                    $this->submitPrerequisites();
                    $this->Worker->done();
                    $this->log("Prerequisites submited");
                    continue;
                }
                $this->beforeWork($data);
                $this->work($data);
                $this->afterWork($data);
                $this->Worker->done();
                $this->log("Job work done");
            } catch (CantWorkException $ex) {
                $this->afterWork($data);
                $this->Worker->done();
                $this->log("Job skipped");
            } catch (Exception $ex) {
                $this->fail($ex, $data);
                $this->Worker->fail();
            }

            if (!$this->keepWorking($this->keepAlive)) {
                break;
            }
        }
        $this->log("Work finished");
        $this->Worker->Pheanstalk->getConnection()->disconnect();
        exit(0);
    }

    /**
     * Lancé avant work()
     * @param mixed $data
     */
    protected function beforeWork($data)
    {
    }

    /**
     * Lancé après work()
     * @param mixed $data
     */
    protected function afterWork($data)
    {
    }

    /**
     * {@inheritdoc}
     *
     * @param mixed $params
     * @return boolean
     */
    public function keepWorking($params)
    {
        if (!empty($params['sleep']) && is_numeric($params['sleep'])) {
            sleep($params['sleep']);
        }
        return $params;
    }

    /**
     * {@inheritdoc}
     *
     * @param array $params
     */
    public function touch($params = array())
    {
        $this->Worker->touch();
        return $params;
    }

    /**
     * {@inheritdoc}
     *
     * @param string $string to log
     */
    public function log($string)
    {
        echo date('Y-m-d h:i:s').' - '.$string."\n";
    }

    /**
     * {@inheritdoc}
     *
     * @param Exception $ex Exception envoyé depui le bloc start()
     * @param mixed $data Data obtenu par Beanstalkd
     */
    public function fail(Exception $ex, $data)
    {
        $this->log("Job work fail");
        debug([
            'worker' => __CLASS__,
            'message' => $ex->getMessage(),
            'pid' => getmypid()
        ]);
    }

    /**
     * {@inheritdoc}
     */
    public function shutdown()
    {
        exit(0);
    }

    /**
     * {@inheritdoc}
     */
    public function submitPrerequisites()
    {
    }
}
