<?php
/**
 * Interface d'un worker
 *
 * Tout worker devrait implémenter cet interface afin de respecter une cohérence
 * entre les différents workers
 *
 * @category    Beanstalk
 *
 * @author      Ludovic Ganée <ludovic.ganee@libriciel.coop>
 * @copyright   (c) 2017, Libriciel
 * @license     http://www.cecill.info/licences/Licence_CeCILL_V2-fr.html
 */

namespace Beanstalk\Shell;

use Beanstalk\Exception\CantWorkException;
use Exception;

/**
 * Interface d'un worker
 *
 * <b>start()</b> se charge de lancer la procédure.
 * <b>work()</b> effectue le travail
 * <b>keepWorking()</b> permet de relancer <b>work()</b>
 */
interface WorkerInterface
{
    /**
     * Effectue le travail voulu
     *
     * @param mixed $data
     * @throws CantWorkException
     */
    public function work($data);

    /**
     * Permet de maintenir le worker en vie
     *
     * @param mixed $params
     * @return boolean
     */
    public function keepWorking($params);

    /**
     * Lance la boucle du worker
     * Appellera de façon consécutive, la fonction <b>work()</b> puis
     * la fonction <b>keepWorking()</b>.
     */
    public function start();

    /**
     * Permet de mettre à jour le timer
     *
     * @param array $params
     */
    public function touch($params = array());

    /**
     * Permet d'envoyer un message formaté dans le but de garnir le fichier log
     *
     * @param string $string to log
     */
    public function log($string);

    /**
     * Callback utilisé en cas d'Exception dans le bloc d'execution du worker
     *
     * @param Exception $ex Exception envoyé depui le bloc start()
     * @param mixed $data Data obtenu par Beanstalkd
     */
    public function fail(Exception $ex, $data);

    /**
     * Met fin à l'execution du worker
     */
    public function shutdown();

    /**
     * Permet d'obtenir les prérequis du worker et de savoir s'ils sont résolus
     * Enverra dans le tube prerequisites:
     *  [
     *      'tube' => 'montube',
     *      'date' => '2018-04-25T14:05:11',
     *      'prerequisites' => ['mon prérequis' => true, ...]
     *  ]
     */
    public function submitPrerequisites();
}
