<?php
/**
 * Shell de lancement des workers
 *
 * Permet de lancer un worker en fonction de la configuration et en assurant
 * le suivi sur base de donnée.
 *
 * @category    Beanstalk
 *
 * @author      Ludovic Ganée <ludovic.ganee@libriciel.coop>
 * @copyright   (c) 2017, Libriciel
 * @license     http://www.cecill.info/licences/Licence_CeCILL_V2-fr.html
 */

namespace Beanstalk\Shell;

use Beanstalk\Utility\CakeWorker;
use Cake\Console\Shell;
use Cake\Core\Configure;
use Cake\ORM\TableRegistry;
use Exception;

/**
 * Classe du shell
 *
 * Contient une méthode unique qui utilise le premier paramètre pour choisir
 * quel worker lancer.
 *
 * Le nom du worker doit correspondre au nom dans la configuration sous la clef
 * "Beanstalk.workers".
 *
 * Le worker doit implémenter une interface définie dans la constante
 * WORKER_INTERFACE
 *
 * Le suivi en base de donnée est assurée, le pid sera enregistré en base puis
 * retiré lorsque le worker aura fini de fonctionner.
 */
class WorkerShell extends Shell
{

    /**
     * Interface que doit avoir le worker
     */
    const WORKER_INTERFACE = 'Beanstalk\Shell\WorkerInterface';

    /** @noinspection PhpHierarchyChecksInspection */
    /**
     * Worker Shell
     *
     * @return int
     * @throws Exception
     */
    public function main()
    {
        if (!isset($this->args[0])) {
            throw new Exception("Le worker n'a pas été indiqué");
        }

        $workers = Configure::read('Beanstalk.workers');
        $config = null;
        foreach ($workers as $name => $conf) {
            if ($name === $this->args[0]) {
                $config = $conf;
                break;
            }
        }

        if (!$config) {
            throw new Exception("Le worker n'a pas été configuré");
        } elseif (!is_file($config['path'])) {
            throw new Exception("Le fichier du worker n'a pas été trouvé");
        } elseif (!is_readable($config['path'])) {
            throw new Exception("Le fichier du worker n'est pas lisible");
        }

        $className = $config['namespace'].'\\'.basename($config['path'], '.php');
        if (!in_array(
                static::WORKER_INTERFACE,
                class_implements($className))
        ) {
            throw new Exception(
                "Le worker ne possède pas l'interface "
                . static::WORKER_INTERFACE);
        }

        $BeanstalkWorkers = TableRegistry::getTableLocator()->get('Beanstalk.BeanstalkWorkers');
        $hostname = gethostname();
        $entity = $BeanstalkWorkers->find()
            ->where(['name' => $this->args[0].'-'.$hostname])
            ->first();

        if (!$entity) {
            $entity = $BeanstalkWorkers->newEntity();
            $entity->set('name', $this->args[0].'-'.$hostname);
            $entity->set('tube', $config['tube']);
            $entity->set('path', $config['path']);
        }

        $entity->set('last_launch', time());
        $entity->set('pid', getmypid());
        $entity->set('hostname', $hostname);
        $BeanstalkWorkers->save($entity);

        $Worker = new $className(
            CakeWorker::getWorker($config['tube']),
            $config['keep_alive'],
            (array)$config['params']);

        if ($Worker instanceof WorkerInterface) {
            $Worker->start();
        }
        return 0;
    }

    /**
     * Détruit proprement le worker
     */
    public function __destruct()
    {
        try {
            $BeanstalkWorkers = TableRegistry::getTableLocator()->get('Beanstalk.BeanstalkWorkers');
            $entity = $BeanstalkWorkers->find()
                ->where(['name' => $this->args[0].'-'.gethostname()])
                ->firstOrFail();
            $BeanstalkWorkers->delete($entity);
        } catch (\Exception $e) {
            $this->abort($e->getMessage());
        }
    }
}
