<?php
/**
 * Permet d'établir une connexion entre Cake et Le client Beanstalk
 *
 * Regroupe les méthodes communes des clients CakePhp3
 *
 * @category    Beanstalk
 *
 * @author      Ludovic Ganée <ludovic.ganee@libriciel.coop>
 * @copyright   (c) 2017, Libriciel
 * @license     http://www.cecill.info/licences/Licence_CeCILL_V2-fr.html
 */

namespace Beanstalk\Utility;

use Cake\ORM\TableRegistry;
use Exception;
use Pheanstalk\Pheanstalk;

/**
 * Classe abstraite du client pour CakePhp3
 *
 * Utilisation de type Multi-ton en fonction du tube choisi par le client.
 *
 * <b>connected()</b> permet de savoir si le client est connecté au serveur
 * Beanstalk sans générer d'exception.
 *
 * <b>getCount()</b> permet de connaître la quantité de travail à fournir.
 * Utile pour éviter d'attendre un travail en vérifiant à l'avance qu'il y
 * en ai.
 */
abstract class CakeClient implements ClientInterface
{
    /**
     * Une instance par tube et par classe
     *
     * @var array self
     */
    protected static $instance = [];

    /**
     * Model BeanstalkJobs
     *
     * @var \Beanstalk\Model\Table\BeanstalkJobsTable
     */
    protected $BeanstalkJobs = null;

    /**
     * @var Pheanstalk instance de Pheanstalk
     */
    protected $Pheanstalk;

    /**
     * Constructeur multi-ton
     *
     * @param string $tube
     */
    protected function __construct($tube = 'default')
    {
        $this->BeanstalkJobs = TableRegistry::getTableLocator()->get('Beanstalk.BeanstalkJobs');
    }

    /**
     * Permet d'obtenir l'instance selon le tube
     *
     * @param string $tube
     * @return CakeClient
     */
    public static function getInstance($tube = 'default')
    {
        if (!isset(static::$instance[get_called_class()][$tube])) {
            static::$instance[get_called_class()][$tube] = new static($tube);
        }

        return static::$instance[get_called_class()][$tube];
    }

    /**
     * {@inheritdoc}
     *
     * @return bool
     */
    public static function connected()
    {
        try {
            static::getInstance();
            return true;
        } catch (Exception $ex) {
            return false;
        }
    }

    /**
     * {@inheritdoc}
     *
     * @param string $tube
     */
    public static function getPheanstalk($tube = 'default')
    {
        return static::getClient($tube)->Pheanstalk;
    }

    /**
     * {@inheritdoc}
     *
     * @param string $tube
     * @return int
     */
    public static function getCount($tube = 'default')
    {
        try {
            $stats = static::getPheanstalk($tube)->statsTube($tube);
            return (int)$stats['current-jobs-ready'];
        } catch (Exception $ex) {
            return 0;
        }
    }
}
