<?php
/**
 * Émetteur client cake
 *
 * Permet d'envoyer des message au serveur Beanstalk tout en gardant la trace
 * de la transaction sous base de donnée.
 *
 * @category    Beanstalk
 *
 * @author      Ludovic Ganée <ludovic.ganee@libriciel.coop>
 * @copyright   (c) 2017, Libriciel
 * @license     http://www.cecill.info/licences/Licence_CeCILL_V2-fr.html
 */

namespace Beanstalk\Utility;

use Beanstalk\Model\Table\BeanstalkJobsTable;
use Cake\Core\Configure;
use Cake\ORM\TableRegistry;
use Beanstalk\lib\Emitter;
use Exception;
use Pheanstalk\Connection;

/**
 * Classe de l'emetteur cake
 *
 * Permet notamment d'envoyer des message de n'importe où.
 * Un simple <b>CakeEmitter::emit('montube', $data)</b> enverra au serveur
 * Beanstalk un message contenant $data dans le bon tube et la transaction sera
 * sauvegardée en base de donnée.
 */
class CakeEmitter extends CakeClient
{
    /**
     * Instance Emitter
     *
     * @var \Beanstalk\lib\Emitter
     */
    protected $emitter = null;

    /**
     * Constructeur multi-ton
     *
     * @param string $tube
     * @throws Exception
     */
    protected function __construct($tube = 'default')
    {
        $conf = Configure::read('Beanstalk');
        $this->emitter = new Emitter(
            $tube,
            $conf['host'],
            $conf['port'],
            $conf['timeout'],
            $conf['persistant']
        );

        $conn = $this->emitter->Pheanstalk->getConnection();
        if ($conn instanceof Connection && !$conn->isServiceListening()) {
            throw new Exception("Pheanstalk not listening Beanstalkd");
        }

        parent::__construct($tube);
    }

    /**
     * Permet de récupérer l'instance de l'émetteur
     *
     * @param string $tube
     * @return Emitter
     * @throws Exception
     */
    public static function getEmitter($tube = 'default')
    {
        $instance = static::getInstance($tube);
        if (!$instance instanceof CakeEmitter) {
            throw new Exception("Instance != CakeEmitter");
        }
        return $instance->emitter;
    }

    /**
     * Envoi un message
     *
     * @param string $tube
     * @param mixed $data
     * @param int $priority From 0 (most urgent) to 0xFFFFFFFF (least urgent)
     * @param int $delay Seconds to wait before job becomes ready
     * @param int $ttr Time To Run: seconds a job can be reserved for
     * @return bool
     * @throws Exception
     */
    public static function emit(
        $tube, $data, $priority = 1024, $delay = 0, $ttr = 60)
    {
        $BeanstalkJobs = TableRegistry::getTableLocator()->get('Beanstalk.BeanstalkJobs');
        if ($BeanstalkJobs instanceof BeanstalkJobsTable) {
            return $BeanstalkJobs
                ->saveAndEmit(compact('tube', 'data', 'priority', 'delay', 'ttr'));
        }
        return false;
    }

    /**
     * Envoi un message
     *
     * @param string $tube
     * @param mixed $data
     * @param int $priority From 0 (most urgent) to 0xFFFFFFFF (least urgent)
     * @param int $delay Seconds to wait before job becomes ready
     * @param int $ttr Time To Run: seconds a job can be reserved for
     * @return int
     * @throws Exception
     */
    public static function emitNoSave(
        $tube, $data, $priority = 1024, $delay = 0, $ttr = 60)
    {
        $instance = static::getInstance($tube);
        if (!$instance instanceof CakeEmitter) {
            throw new Exception("Instance != CakeEmitter");
        }

        $jobid = $instance->emitter->emit($data, $priority, $delay, $ttr);

        return $jobid;
    }

    /**
     * Méthode générique pour obtenir le client
     *
     * @param string $tube
     * @return \Beanstalk\lib\Emitter
     * @throws Exception
     */
    public static function getClient($tube = 'default')
    {
        return static::getEmitter($tube);
    }
}
