<?php
/**
 * Gestionnaire des workers
 *
 * Assure un lancement correct des workers en fonction de la configuration
 * définie. Les workers seront lancés avec les bons paramètres en arrière plan
 * et le suivi sera fait par base de donnée.
 *
 * @category    Beanstalk
 *
 * @author      Ludovic Ganée <ludovic.ganee@libriciel.coop>
 * @copyright   (c) 2017, Libriciel
 * @license     http://www.cecill.info/licences/Licence_CeCILL_V2-fr.html
 */
namespace Beanstalk\Utility;

use Cake\Core\Configure;
use Cake\ORM\TableRegistry;
use Beanstalk\lib\Worker;
use Exception;

/**
 * La classe du gestionnaire
 *
 * Lors de l'initialisation, la base de donnée est synchronisée avec la
 * configuration à l'aide des méthodes <b>createConfiguredInDb()</b> et
 * <b>deleteNotConfigured()</b>.
 *
 * <b>launch()</b> lance un worker
 *
 * <b>autoLaunch()</b> lance tout les workers configuré avec la clef
 * 'auto_launch' à true.
 */
class WorkerManager
{

    /**
     * Instance unique de cette classe
     *
     * @var WorkerManager
     */
    protected static $instance;

    /**
     * Table liée
     *
     * @var \Beanstalk\Model\Table\BeanstalkWorkersTable
     */
    public $BeanstalkWorkers;

    /**
     * Worker pour communiquer les pids
     *
     * @var \Beanstalk\lib\Worker
     */
    protected $worker;

    /**
     * Constructeur de classe
     */
    protected function __construct()
    {
        $this->BeanstalkWorkers = TableRegistry::getTableLocator()->get('Beanstalk.BeanstalkWorkers');
        $conf = Configure::read('Beanstalk');
        try {
            $this->worker = new Worker(
                'workers', $conf['host'], $conf['port'], $conf['timeout'], $conf['persistant']
            );
        } catch (Exception $e) {
            trigger_error($e->getMessage(), E_USER_WARNING);
        }
    }

    /**
     * Permet d'obtenir l'instance unique de cette classe
     *
     * @return WorkerManager
     */
    public static function getInstance()
    {
        if (!static::$instance) {
            static::$instance = new static();
        }

        return static::$instance;
    }

    /**
     * Synchronise la classe avec la base de donnée
     */
    public static function initialize()
    {
        static::getInstance();
    }

    /**
     * Vérifie la connexion au serveur sans provoquer d'exceptions
     *
     * @return boolean
     */
    public static function connected()
    {
        return CakeEmitter::connected();
    }

    /**
     * Lance un worker
     *
     * @param string $worker nom du worker
     * @return boolean lancement effectué
     * @throws Exception
     */
    public static function launch($worker)
    {
        $instance = static::getInstance();
        $config = Configure::read('Beanstalk.workers.' . $worker);

        if (empty($config)) {
            throw new Exception(
            "Le worker n'a pas été trouvé dans la configuration");
        }

        $data = $instance->BeanstalkWorkers->find()
            ->where(['name' => $worker.'-'.gethostname()])
            ->first();
        if ($data) {
            if (file_exists(sprintf('/proc/%d', $data->get('pid')))) {
                return false;
            }

            if (!file_exists(LOGS . "worker_$worker.log")) {
                touch(LOGS . "worker_$worker.log");
            }
        }
        $cmd = '%s/bin/cake Beanstalk.Worker %s > %sworker_%s.log &';
        exec(sprintf($cmd, ROOT, $worker, LOGS, $worker));

        return true;
    }

    /**
     * Lance tout les workers
     * qui sont configurés pour être lancé automatiquement
     */
    public static function autoLaunch()
    {
        if (Configure::read('Beanstalk.autoLaunch')) {
            foreach (Configure::read('Beanstalk.workers') as $worker => $conf) {
                if ($conf['auto_launch']) {
                    static::launch($worker);
                }
            }
        }
    }
}
