<?php
/**
 * Worker Beanstalk
 *
 * Permet de récupérer des messages au serveur Beanstalk.
 *
 * @category    Beanstalk
 *
 * @author      Ludovic Ganée <ludovic.ganee@libriciel.coop>
 * @copyright   (c) 2017, Libriciel
 * @license     http://www.cecill.info/licences/Licence_CeCILL_V2-fr.html
 */

namespace Beanstalk\lib;

use Pheanstalk\Job;

/**
 * Classe du worker
 *
 * <b>getNext()</b> Récupère le prochain message, met en attente le script php
 * jusqu'a obtention du job.
 *
 * <b>touch()</b> indique au serveur Beanstalk que le worker est toujours en
 * train de travailler sur le job afin de conserver sa réservation. Évite que le
 * job soit récupérer par un autre worker du même tube.
 *
 * <b>getData()</b> permet de récupérer un message formaté par <b>Emitter</b>.
 *
 * <b>done()</b> indique que le travail est accompli.
 *
 * <b>fail()</b> indique que le travail a échoué.
 *
 * <b>cancel()</b> retire la réservation
 */
class Worker extends Client
{
    /**
     * @var Job Dernier job reservé
     */
    public $job;

    /**
     * Supprime le job
     * @return self
     */
    public function done()
    {
        $this->Pheanstalk->delete($this->job);
        return $this;
    }

    /**
     * Enterre le job
     * @return self
     */
    public function fail()
    {
        $this->Pheanstalk->bury($this->job);
        return $this;
    }

    /**
     * Libère le job
     * @return self
     */
    public function cancel()
    {
        $this->Pheanstalk->release($this->job);
        return $this;
    }

    /**
     * Remet à zero le chrono du ttr (time to run)
     * Conserve ainsi la réservation
     * @return self
     */
    public function touch()
    {
        $this->Pheanstalk->touch($this->job);
        return $this;
    }

    /**
     * Récupère le job suivant
     * @return bool
     */
    public function getNext()
    {
        $this->job = $this->Pheanstalk->watch($this->tube)->reserve();
        return true;
    }

    /**
     * Récupère les données
     *
     * @return string|array|bool
     */
    public function getData()
    {
        $rawData = json_decode($this->job->getData(), true);

        if ($rawData) {
            $data = $rawData['data'];
            $key = $rawData['key'];

            $json = json_encode($data);

            if (sha1($json.strlen($json).'salted') === $key) {
                return $data;
            }
        }

        return false;
    }
}
