<?php
/**
 * Datacompressor\Utility\DataCompressor
 */

namespace Datacompressor\Utility;

use Cake\Core\Configure;
use Libriciel\Filesystem\Utility\Filesystem;
use ZipArchive;
use PharData;
use Exception;
use RecursiveIteratorIterator;
use RecursiveDirectoryIterator;
use FilesystemIterator;

/**
 * Utilitaire de manipulation des archives (zip, rar, tar, gz)
 *
 * Utilise la même structure d'objet que \ZipArchive (en moins fourni)
 * @see http://php.net/manual/fr/class.ziparchive.php
 *
 * @category    Utility
 *
 * @author      Ludovic Ganée <ludovic.ganee@libriciel.coop>
 * @copyright   (c) 2017, Libriciel
 * @license     http://www.cecill.info/licences/Licence_CeCILL_V2-fr.html
 */
class DataCompressor
{
    /**
     * @var array [UTF-8, ASCII, ISO-8859-1, ...]
     */
    const FILENAME_ENCODAGES = [
        'UTF-8', 'ASCII',
        'ISO-8859-1', 'ISO-8859-2', 'ISO-8859-3', 'ISO-8859-4', 'ISO-8859-5',
        'ISO-8859-6', 'ISO-8859-7', 'ISO-8859-8', 'ISO-8859-9', 'ISO-8859-10',
        'ISO-8859-13', 'ISO-8859-14', 'ISO-8859-15', 'ISO-8859-16',
        'Windows-1251', 'Windows-1252', 'Windows-1254',
    ];

    /**
     * @var string namespace pour Filesystem
     */
    private static $transaction = 'no-transaction';

    /**
     * Défini l'espace de nom à utiliser pour les transactions Filesystem
     *
     * @param string $namespace
     */
    public static function useTransaction(string $namespace = 'default')
    {
        static::$transaction = $namespace;
    }

    /**
     * Fonction générique de compression (compresse en zip)
     *
     * @param string|array|\Traversable $files A filename, an array of files, or a \Traversable instance
     * @param string $archiveName
     * @param string $baseDir ex: /tmp/dir1/file1 et /tmp/dir1/file2 avec
     *                          /tmp en $baseDir donnera un zip avec
     *                          dir1/file1 et dir1/file2.
     *                          Par défaut, essayera de deviner à l'aide de $files
     * @param bool $overwrite
     * @return bool
     * @throws Exception
     */
    public static function compress($files, string $archiveName, string $baseDir = null, bool $overwrite = true): bool
    {
        if (!is_dir(dirname($archiveName))) {
            if (!class_exists('Libriciel\Filesystem\Utility\Filesystem')) {
                throw new Exception("Le dossier de destination du zip n'existe pas");
            }
            Filesystem::setNamespace(static::$transaction);
            Filesystem::mkdir(dirname($archiveName));
        }
        if (file_exists($archiveName)) {
            if (!$overwrite) {
                throw new \Exception("Un fichier du même nom existe déjà !");
            } elseif (class_exists('Libriciel\Filesystem\Utility\Filesystem')) {
                Filesystem::setNamespace(static::$transaction);
                Filesystem::revertFileIfRollback($archiveName);
            }
        } elseif (class_exists('Libriciel\Filesystem\Utility\Filesystem')) {
            Filesystem::setNamespace(static::$transaction);
            Filesystem::addToDeleteIfRollback($archiveName);
        }
        $zip = new ZipArchive;
        $zip->open($archiveName, ZipArchive::CREATE | ZipArchive::OVERWRITE);

        if (is_string($files)) {
            if (is_file($files)) {
                $files = [$files];
            } elseif (is_dir($files)) {
                $baseDir = $baseDir ?: $files;
                $files = glob($files . DS . '*');
            }
        } elseif ($files instanceof \Traversable) {
            $files = iterator_to_array($files, false);
        }

        if (!$baseDir) {
            $baseDir = !class_exists('Libriciel\Filesystem\Utility\Filesystem')
                ? '/'
                : Filesystem::baseDir($files);
        }

        $addedFile = false;
        foreach ($files as $file) {
            if (is_link($file) || is_dir($file)) {
                continue;
            } elseif (is_readable($file)) {
                $addedFile = static::addFormatedFilename($zip, $file, $baseDir);
            } elseif (file_exists($file)) {
                trigger_error("File is not readable : ".$file);
            } elseif (!file_exists($file)) {
                trigger_error("File not found : ".$file);
            }
            if (!$addedFile) {
                throw new Exception("Erreur lors de l'ajout d'un fichier au zip");
            }
        }
        $success = $zip->close();
        return $success;
    }

    /**
     * Décompresse une archive
     *
     * @param string $path
     * @param string $destination
     * @return bool|array false si échec | array de nom de fichiers
     * @throws Exception
     */
    public static function uncompress(string $path, string $destination)
    {
        $type = static::detectType($path);
        if (!$type) {
            throw new Exception("Le type de fichier n'est pas supporté");
        }
        if (!is_dir($destination)) {
            if (!class_exists('Libriciel\Filesystem\Utility\Filesystem')) {
                throw new Exception("Le dossier de destination d'existe pas");
            }
            $existing = [];
            Filesystem::setNamespace(static::$transaction);
            Filesystem::mkdir($destination);
        } else {
            $existing = static::fileIterator($destination);
        }
        if (!is_writable($destination)) {
            throw new Exception("Le dossier de destination n'est pas inscriptible");
        }
        $uncompressDestination = !empty($existing)
            ? sys_get_temp_dir() . DS . uniqid('uncompress-destination-')
            : $destination;
        if (!is_dir($uncompressDestination)) {
            mkdir($uncompressDestination);
        }
        switch ($type) {
            case 'zip':
                $success = static::uncompressZip($path, $uncompressDestination);
                break;
            case 'gzip':
                $success = static::uncompressGzip($path, $uncompressDestination);
                break;
            default:
                throw new Exception("Aucune méthode pour décompresser le type : ".$type);
        }
        if ($success) {
            if ($uncompressDestination !== $destination) {
                return static::safeMoveto($uncompressDestination, $destination);
            } else {
                return iterator_to_array(static::fileIterator($destination));
            }
        }
        return false;
    }

    /**
     * Déplace les fichiers en utilisant Filesystem si disponnible (pour les transactions)
     *
     * @param string $origin Dossier d'origine
     * @param string $target Dossier final
     * @return array
     */
    private static function safeMoveto(string $origin, string $target): array
    {
        $targetDir = trim($target, DS);
        $files = iterator_to_array(
            static::fileIterator($origin));
        $length = strlen(rtrim($origin, DS)) +1;
        $newFiles = [];

        if (class_exists('Libriciel\Filesystem\Utility\Filesystem')) {
            Filesystem::setNamespace(static::$transaction);
            foreach ($files as $file) {
                $newDir = DS . $targetDir . DS . substr($file, $length);
                $newFiles[] = $newDir;
                Filesystem::mkdir(dirname($newDir));
                Filesystem::rename($file, $newDir, true);
            }
        } else {
            foreach ($files as $file) {
                $newDir = DS . $targetDir . DS . substr($file, $length);
                $newFiles[] = $newDir;
                if (file_exists($newDir)) {
                    unlink($newDir);
                }
                static::mkdir(dirname($newDir));
                rename($file, $newDir);
            }
        }
        return $newFiles;
    }

    /**
     * Créer les dossiers de façon récursive
     *
     * @param string $dir
     */
    private static function mkdir(string $dir)
    {
        $path = '';
        foreach (explode(DS, trim($dir, DS)) as $dirName) {
            $path .= DS . $dirName;
            if (!is_dir($path)) {
                mkdir($path);
            }
        }
    }

    /**
     * Donne la liste complète des fichiers d'un répertoire
     *
     * @param string $dirname
     * @return RecursiveIteratorIterator
     */
    private static function fileIterator(string $dirname): RecursiveIteratorIterator
    {
        $flags = FilesystemIterator::CURRENT_AS_PATHNAME
            | FilesystemIterator::SKIP_DOTS;
        return new RecursiveIteratorIterator(
            new RecursiveDirectoryIterator($dirname, $flags));
    }

    /**
     * Décompresse une archive zip
     *
     * @param string $path
     * @param string $destination
     * @return bool success
     * @throws Exception
     */
    private static function uncompressZip(string $path, string $destination): bool
    {
        $zip = new ZipArchive;
        if (!$zip->open($path)) {
            throw new Exception("Impossible d'ouvrir le fichier zip");
        }

        for ($i = 0; $i < $zip->numFiles; $i++) {
            $indexName = $zip->getNameIndex($i);
            $filename = static::formatFilename($indexName, 'UTF-8');
            // suppression des fichiers d'indexation de MacOS
            if (strpos($filename, '__MACOSX') !== false) {
                $zip->deleteIndex($i);
                continue;
            }
            if ($indexName !== $filename) {
                $zip->renameIndex($i, $filename);
            }
        }
        $success = $zip->extractTo($destination);
        $zip->unchangeAll();
        $zip->close();
        return $success;
    }

    /**
     * Décompresse une archive tar.gz
     *
     * NOTE : Peut altérer son contenu (supprime __MAXOSX et renomme en UTF-8)
     * La sauvegarde du fichier serait possible mais couteuse en ressources
     *
     * @param string $path
     * @param string $destination
     * @return bool success
     * @throws Exception
     */
    private static function uncompressGzip(string $path, string $destination): bool
    {
        $tarName = substr($path, 0, strlen($path) - strlen('.gz'));
        $existingTar = is_file($tarName);
        if ($existingTar) {
            // Note : Impossible de décompresser un tar.gz là ou il a été compressé
            // -> on déplace le fichier dans un autre dossier pour résoudre le problème
            Filesystem::begin('DataCompressor_uncompressGzip');
            Filesystem::setSafemode(true);
            Filesystem::rename($tarName, $tarName.'.bak');
            $newPath = dirname($path).DS.'tmp'.DS.basename($path);
            if (!is_dir(dirname($path).DS.'tmp')) {
                Filesystem::mkdir(dirname($path).DS.'tmp');
            }
            Filesystem::rename($path, $newPath);
            $path = $newPath;
        }
        $gzip = new PharData($path);
        $compressed = $gzip->isCompressed();

        if ($compressed) {
            $gzip = $gzip->decompress();
        }

        try {
            $gzip->extractTo($destination, null, true);
            $success = true;
        } catch (Exception $e) {
            $success = false;
        }

        if ($compressed) {
            if (is_file($tarName)) {
                unlink($tarName);
            }
            if ($existingTar) {
                Filesystem::rollback('DataCompressor_uncompressGzip');
            }
        }

        return $success;
    }

    /**
     * Ajoute au zip le fichier/dossier $filename formaté selon
     * l'encodage choisi et $baseDir retiré
     *
     * @param ZipArchive $zip
     * @param string $filename
     * @param string $baseDir
     * @return bool
     */
    public static function addFormatedFilename(ZipArchive $zip, string $filename, string $baseDir = '/'): bool
    {
        $normalized = static::formatFilename($filename);
        $inzipName = trim(substr($normalized, strlen($baseDir)), DS);

        if (is_dir($filename)) {
            return $zip->addEmptyDir($inzipName);
        }
        return $zip->addFile($filename, $inzipName);
    }

    /**
     * Renvoi un nom de fichier encodé en UTF-8 et normalisé
     *
     * L'encodage UTF-8 : Peut être modifié dans la configuration sous la
     * clef DataCompressor.encodeFilename
     *
     * @param string $filename
     * @param string $format
     * @return string
     */
    public static function formatFilename(string $filename, string $format = null): string
    {
        $useFormat = $format ?:
            (Configure::read('DataCompressor.encodeFilename') ?: 'UTF-8');
        $filenameFormat = mb_detect_encoding($filename, static::FILENAME_ENCODAGES);

        $formated = $useFormat !== $filenameFormat
            ? iconv($filenameFormat, $useFormat, $filename)
            : $filename;

        $normalized = normalizer_is_normalized($formated)
            ? $formated
            : normalizer_normalize($formated, \Normalizer::FORM_C);

        return $normalized;
    }

    /**
     * Permet de détecter le type d'archive
     *
     * @param string $filename
     * @return boolean|string false|zip|gzip
     */
    public static function detectType(string $filename)
    {
        if (is_file($filename)) {
            switch (mime_content_type($filename)) {
                case 'application/x-zip':
                case 'application/x-zip-compressed':
                case 'application/zip':
                case 'application/zip-compressed':
                case 'multipart/x-zip':
                    return 'zip';
                case 'application/x-rar':
                case 'application/x-rar-compressed':
                    return 'rar';
                case 'application/gzip':
                case 'application/tar+gzip':
                case 'application/x-gzip':
                case 'application/x-tgz':
                case 'application/x-gtar':
                case 'application/tar':
                case 'application/x-tar':
                case 'multipart/x-tar':
                    return 'gzip';
            }
        } else {
            switch ($ext = pathinfo($filename, PATHINFO_EXTENSION)) {
                case 'rar':
                case 'zip':
                    return $ext;
                case 'tar':
                case 'gz':
                    return 'gzip';
            }
        }

        return false;
    }
}
