<?php
/**
 * FileConverters\Utility\FileConverters
 */

namespace FileConverters\Utility;

/**
 * Conversion d'un fichier
 *
 * Permet de modifier un fichier vers un autre format au choix
 *
 * @category    File
 *
 * @author      Ludovic Ganée <ludovic.ganee@libriciel.coop>
 * @copyright   (c) 2017, Libriciel
 * @license     http://www.cecill.info/licences/Licence_CeCILL_V2-fr.html
 */
class FileConverters
{
    /**
     * Liste les callbacks pour convertir vers une extension en particulier
     *
     * match : type de fichiers @see self::$mimes
     * callback : callable
     *
     * @var array
     */
    public static $callbacks = [
        'jpg' => [
            'match' => [
                'images',
                'libre_office_writer',
                'ms_office_writer',
                'libre_office_calc',
                'ms_office_calc',
                'libre_office_pres',
                'ms_office_pres',
                'texts',
                'htmls'
            ],
            'callback' => '\FileConverters\Utility\FileConverters::convertToJpg'
        ],
        'png' => [
            'match' => [
                'images',
                'libre_office_writer',
                'ms_office_writer',
                'libre_office_calc',
                'ms_office_calc',
                'libre_office_pres',
                'ms_office_pres',
                'texts',
                'htmls'
            ],
            'callback' => '\FileConverters\Utility\FileConverters::convertToPng'
        ],
        'mp4' => [
            'match' => ['videos'],
            'callback' => '\FileConverters\Utility\FileConverters::convertToMp4'
        ],
        'mkv' => [
            'match' => ['videos'],
            'callback' => '\FileConverters\Utility\FileConverters::convertToMkv'
        ],
        'ods' => [
            'match' => ['libre_office_calc', 'ms_office_calc'],
            'callback' => '\FileConverters\Utility\FileConverters::convertToOds'
        ],
        'odp' => [
            'match' => ['libre_office_pres', 'ms_office_pres'],
            'callback' => '\FileConverters\Utility\FileConverters::convertToOdp'
        ],
        'odt' => [
            'match' => ['libre_office_writer', 'ms_office_writer', 'texts', 'htmls'],
            'callback' => '\FileConverters\Utility\FileConverters::convertToOdt'
        ],
        'html' => [
            'match' => [
                'libre_office_writer',
                'ms_office_writer',
                'libre_office_calc',
                'ms_office_calc',
                'libre_office_pres',
                'ms_office_pres',
                'texts',
                'htmls'
            ],
            'callback' => '\FileConverters\Utility\FileConverters::convertToHtml'
        ],
        'flac' => [
            'match' => ['audios'],
            'callback' => '\FileConverters\Utility\FileConverters::convertToFlac'
        ],
        'ogg' => [
            'match' => ['audios'],
            'callback' => '\FileConverters\Utility\FileConverters::convertToOgg'
        ],
        'wav' => [
            'match' => ['audios'],
            'callback' => '\FileConverters\Utility\FileConverters::convertToWav'
        ],
        'pdf' => [
            'match' => [
                'pdfs',
                'libre_office_writer',
                'ms_office_writer',
                'libre_office_calc',
                'ms_office_calc',
                'libre_office_pres',
                'ms_office_pres',
                'texts',
                'htmls'
            ],
            'callback' => '\FileConverters\Utility\FileConverters::convertToPdf'
        ]
    ];

    /**
     * Associe des codes mime avec des type de fichiers convertissable
     *
     * @var array
     */
    public static $mimes = [
        'images' => '/^image\/.*$/',
        'videos' => '/^video\/.*$/',
        'audios' => '/^audio\/.*$/',
        'libre_office_writer' => '/^application\/(vnd\.(oasis\.opendocument|sun\.xml|stardivision)\.|x-star)(text|writer)$/',
        'ms_office_writer' => '/^application\/(msword|vnd\.openxmlformats\-officedocument\.wordprocessingml\.document|vnd\.ms-word\.document.*)$/',
        'libre_office_calc' => '/^application\/(vnd\.(oasis\.opendocument|sun\.xml|stardivision)\.|x-star)(spreadsheet|calc)$/',
        'ms_office_calc' => '/^application\/vnd\.(ms\-excel(\.sheet.*)?|openxmlformats\-officedocument\.spreadsheetml\.sheet)$/',
        'libre_office_pres' => '/^application\/(vnd\.(oasis\.opendocument|sun\.xml|stardivision)\.|x-star)(presentation|impress)$/',
        'ms_office_pres' => '/^application\/vnd\.(ms\-powerpoint(\.presentation.*)?|openxmlformats\-officedocument\.presentationml\.presentation)$/',
        'texts' => '/^(text\/(plain|rtf)|application\/octet-stream)$/',
        'htmls' => '/^(text\/.*html|application\/xml)$/',
        'pdfs' => '/^application\/(x\-)?pdf$/',
    ];

    /**
     * A des fins de debug, permet de lister les commandes exécutés,
     * leurs valeurs de retour, les erreurs et les codes de retour
     *
     * @var array
     */
    public static $logs = [];

    /**
     * Converti un fichier vers un autre format
     *
     * @param string $filename chemin vers le fichier à convertir
     * @param string $convertTo extension du fichier une fois converti
     * @param string $mv destination où écrire le fichier converti (optionnel)
     * @param mixed $params paramètres à envoyer au callback
     * @return boolean|string fichier binaire
     */
    public static function convert($filename, $convertTo, $mv = null, $params = [])
    {
        if (!isset(static::$callbacks[$convertTo]) || !is_readable($filename)) {
            return false;
        }

        $bin = call_user_func(static::$callbacks[$convertTo]['callback'], $filename, $params);

        if (empty($bin)) {
            return false;
        }

        if ($mv) {
            $handle = fopen($mv, "w");

            if ($handle) {
                fwrite($handle, $bin);
            } else {
                trigger_error("Unable to write at '$mv' location");
            }
            fclose($handle);
        }

        return $bin;
    }

    /**
     * Permet à partir d'un nom de fichier de connaitre la liste des conversion
     * possible.
     *
     * @param string $filename
     * @return array|bool
     */
    public static function matchFormats($filename)
    {
        if (!is_readable($filename)) {
            return false;
        }
        $mime = mime_content_type($filename);
        $results = [];
        foreach (static::$mimes as $format => $regex) {
            if (!preg_match($regex, $mime)) {
                continue;
            }
            foreach (static::$callbacks as $key => $values) {
                if (in_array($format, $values['match'])) {
                    $results[] = $key;
                }
            }
        }

        return $results;
    }

    /**
     * Callback
     *
     * @param string $filename
     * @return bool|string fichier binaire
     * @internal param array $params
     */
    public static function convertToJpg($filename)
    {
        $mime = mime_content_type($filename);

        if (preg_match(static::$mimes['images'], $mime)) {
            return static::imageToImage($filename, 'jpg');
        } elseif (in_array('jpg', static::matchFormats($filename))) {
            $bin = static::officeTo($filename, 'png');
            if ($bin) {
                $tmpFile = static::tmpFile($bin);
                $image = static::imageToImage($tmpFile, 'jpg');
                unlink($tmpFile);
                return $image;
            }
        }

        return false;
    }

    /**
     * Callback
     *
     * @param string $filename
     * @return bool|string fichier binaire
     * @internal param array $params
     */
    public static function convertToPng($filename)
    {
        $mime = mime_content_type($filename);

        if (preg_match(static::$mimes['images'], $mime)) {
            return static::imageToImage($filename, 'png');
        } elseif (in_array('png', static::matchFormats($filename))) {
            return static::officeTo($filename, 'png');
        }

        return false;
    }

    /**
     * Callback
     *
     * @param string $filename
     * @param array $params
     * @return boolean|string fichier binaire
     */
    public static function convertToMp4($filename, $params = [])
    {
        $mime = mime_content_type($filename);

        if (preg_match(static::$mimes['videos'], $mime)) {
            return static::videoToVideo($filename, 'mp4', $params);
        }

        return false;
    }

    /**
     * Callback
     *
     * @param string $filename
     * @param array $params
     * @return boolean|string fichier binaire
     */
    public static function convertToMkv($filename, $params = [])
    {
        $mime = mime_content_type($filename);

        if (preg_match(static::$mimes['videos'], $mime)) {
            return static::videoToVideo($filename, 'mkv', $params);
        }

        return false;
    }

    /**
     * Callback
     *
     * @param string $filename
     * @return bool|string fichier binaire
     * @internal param array $params
     */
    public static function convertToPdf($filename)
    {
        if (in_array('pdf', static::matchFormats($filename))) {
            return static::officeToPdf($filename, 'pdf');
        }

        return false;
    }

    /**
     * Callback
     *
     * @param string $filename
     * @return bool|string fichier binaire
     * @internal param array $params
     */
    public static function convertToOdt($filename)
    {
        if (in_array('odt', static::matchFormats($filename))) {
            return static::officeToOffice($filename, 'odt');
        }

        return false;
    }

    /**
     * Callback
     *
     * @param string $filename
     * @return bool|string fichier binaire
     * @internal param array $params
     */
    public static function convertToOds($filename)
    {
        if (in_array('ods', static::matchFormats($filename))) {
            return static::officeToOffice($filename, 'odt');
        }

        return false;
    }

    /**
     * Callback
     *
     * @param string $filename
     * @return bool|string fichier binaire
     * @internal param array $params
     */
    public static function convertToOdp($filename)
    {
        if (in_array('odp', static::matchFormats($filename))) {
            return static::officeToOffice($filename, 'odp');
        }

        return false;
    }

    /**
     * Callback
     *
     * @param string $filename
     * @return bool|string fichier binaire
     * @internal param array $params
     */
    public static function convertToHtml($filename)
    {
        if (in_array('html', static::matchFormats($filename))) {
            return static::officeTo($filename, 'html');
        }

        return false;
    }

    /**
     * Callback
     *
     * @param string $filename
     * @return bool|string fichier binaire
     * @internal param array $params
     */
    public static function convertToFlac($filename)
    {
        $mime = mime_content_type($filename);

        if (preg_match(static::$mimes['audios'], $mime)) {
            return static::audioToAudio($filename, 'flac');
        }

        return false;
    }

    /**
     * Callback
     *
     * @param string $filename
     * @return bool|string fichier binaire
     * @internal param array $params
     */
    public static function convertToOgg($filename)
    {
        $mime = mime_content_type($filename);

        if (preg_match(static::$mimes['audios'], $mime)) {
            return static::audioToAudio($filename, 'ogg');
        }

        return false;
    }

    /**
     * Callback
     *
     * NOTE: le mp3 n'est pas un format libre, vous devez avoir la license pour
     * l'utiliser
     *
     * @param string $filename
     * @return bool|string fichier binaire
     * @internal param array $params
     */
    public static function convertToMp3($filename)
    {
        $mime = mime_content_type($filename);

        if (preg_match(static::$mimes['audios'], $mime)) {
            return static::audioToAudio($filename, 'mp3');
        }

        return false;
    }

    /**
     * Callback
     *
     * @param string $filename
     * @return bool|string fichier binaire
     * @internal param array $params
     */
    public static function convertToWav($filename)
    {
        $mime = mime_content_type($filename);

        if (preg_match(static::$mimes['audios'], $mime)) {
            return static::audioToAudio($filename, 'wav');
        }

        return false;
    }

    /**
     * Converti un pdf/office vers du pdf/a
     *
     * @param string $filename
     * @param string $type
     * @return boolean|string
     */
    protected static function officeToPdf($filename, $type = 'pdf')
    {
        $output = sys_get_temp_dir().DIRECTORY_SEPARATOR.uniqid('converted').'.'.$type;
        $cmd = sprintf(
            "unoconv -f %s -eSelectPdfVersion=1 -o %s %s",
            $type,
            escapeshellarg($output),
            escapeshellarg($filename));
        static::$logs[] = static::exec($cmd, $strout, $code);

        if ($code === 0 && is_file($output)) {
            $bin = file_get_contents($output);
        } else {
            $bin = false;
        }
        is_file($output) && unlink($output);

        return $bin;
    }

    /**
     * Converti une image vers une autre image
     *
     * @param string $filename
     * @param string $type ex: jpg, png, gif...
     * @return boolean|string
     */
    protected static function imageToImage($filename, $type = 'png')
    {
        $output = sys_get_temp_dir().DIRECTORY_SEPARATOR.uniqid('converted').'.'.$type;
        $cmd = sprintf(
            "convert %s[0] %s",
            escapeshellarg($filename),
            escapeshellarg($output));
        static::$logs[] = static::exec($cmd, $strout, $code);

        if ($code === 0 && is_file($output)) {
            $bin = file_get_contents($output);
        } else {
            $bin = false;
        }
        is_file($output) && unlink($output);

        return $bin;
    }

    /**
     * Converti une video vers un autre type de video
     *
     * @see https://handbrake.fr/docs/en/1.0.0/cli/cli-guide.html
     *
     * @param string $filename
     * @param string $type
     * @param array $params
     * @return boolean|string
     */
    protected static function videoToVideo($filename, $type = 'mp4', $params = [])
    {
        $params += [
            'video' => 'x264',
            'audio' => 'av_aac',
        ];

        $output = sys_get_temp_dir().DIRECTORY_SEPARATOR.uniqid('converted').'.'.$type;
        $cmd = sprintf(
            "HandBrakeCLI -i %s -o %s --encoder %s --aencoder %s",
            escapeshellarg($filename),
            escapeshellarg($output),
            $params['video'],
            $params['audio']);

        static::$logs[] = static::exec($cmd, $strout, $code);

        if ($code === 127) {
            trigger_error("Vous devez installer HandBrakeCLI pour effectuer la conversion video. Essayez sudo apt-get install handbrake-cli");
            return false;
        }

        if ($code === 0 && is_file($output)) {
            $bin = file_get_contents($output);
        } else {
            $bin = false;
        }
        is_file($output) && unlink($output);

        return $bin;
    }

    /**
     * Converti un fichier office vers un autre type
     *
     * @param string $filename
     * @param string $type
     * @return boolean
     */
    protected static function officeTo($filename, $type = 'pdf')
    {
        $output = sys_get_temp_dir()
            .DIRECTORY_SEPARATOR
            .basename($filename, pathinfo($filename, PATHINFO_EXTENSION))
            .'.'
            .$type;

        $cmd = sprintf(
            "libreoffice --headless --convert-to %s --outdir %s %s",
            $type,
            escapeshellarg(sys_get_temp_dir()),
            escapeshellarg($filename));

        static::$logs[] = static::exec($cmd, $strout, $code);

        if ($code === 0 && is_file($output)) {
            $bin = file_get_contents($output);
        } else {
            $bin = false;
        }
        is_file($output) && unlink($output);

        return $bin;
    }

    /**
     * Converti un fichier office vers un autre office
     *
     * @param string $filename
     * @param string $type
     * @return boolean
     */
    protected static function officeToOffice($filename, $type = 'odt')
    {
        $output = sys_get_temp_dir().DIRECTORY_SEPARATOR.uniqid('converted').'.'.$type;

        $cmd = sprintf(
            "unoconv -f %s -o %s %s",
            $type,
            escapeshellarg($output),
            escapeshellarg($filename));

        static::$logs[] = static::exec($cmd, $strout, $code);

        if ($code === 0 && is_file($output)) {
            $bin = file_get_contents($output);
        } else {
            $bin = false;
        }
        is_file($output) && unlink($output);

        return $bin;
    }

    /**
     * Créer un fichier temporaire et renvoi son path
     *
     * @param string $bin
     * @return string filename
     */
    protected static function tmpFile($bin)
    {
        $tmpfname = tempnam(sys_get_temp_dir(), 'tmpfileconverters');
        $handle = fopen($tmpfname, "w");
        fwrite($handle, $bin);
        fclose($handle);

        return $tmpfname;
    }

    /**
     * Converti un fichier audio vers un autre type
     *
     * @param string $filename
     * @param string $type
     * @return boolean|string
     */
    protected static function audioToAudio($filename, $type = 'ogg')
    {
        $output = sys_get_temp_dir().DIRECTORY_SEPARATOR.uniqid('converted').'.'.$type;
        is_file($output) && unlink($output);

        $cmd = sprintf(
            "ffmpeg -i %s %s",
            escapeshellarg($filename),
            escapeshellarg($output));

        static::$logs[] = static::exec($cmd, $strout, $code);

        if ($code === 127) {
            trigger_error("Vous devez installer ffmpeg pour effectuer la conversion audio. Essayez sudo apt install ffmpeg");
            return false;
        }

        if ($code === 0 && is_file($output)) {
            $bin = file_get_contents($output);
        } else {
            $bin = false;
        }
        is_file($output) && unlink($output);

        return $bin;
    }

    /**
     * Permet de lancer une commande en récupérant un array qui contient
     * - La commande envoyé
     * - La sorti standard
     * - La sorti d'erreur
     * - Le code de retour
     *
     * @param string $cmd
     * @param array $output
     * @param int $return_var
     * @return array [string 'cmd', string 'stdout', string 'stderr', int 'code']
     */
    protected static function exec($cmd, &$output = null, &$return_var = null)
    {
        $stdoutFile = tempnam(sys_get_temp_dir(), 'exec-out-');
        $stderrFile = tempnam(sys_get_temp_dir(), 'exec-err-');
        exec($cmd . ' > '. $stdoutFile .' 2> ' . $stderrFile, $output, $return_var);
        $stdout = file_get_contents($stdoutFile);
        $stderr = file_get_contents($stderrFile);
        unlink($stdoutFile);
        unlink($stderrFile);
        return [
            'cmd' => $cmd,
            'stdout' => $stdout,
            'stderr' => $stderr,
            'code' => $return_var
        ];
    }
}
