<?php
/**
 * Entité de la table beanstalk_jobs
 *
 * Défini des champs virtuels afin d'exploiter les informations détenues par le
 * serveur Beanstalk.
 *
 * @category    Beanstalk
 *
 * @author      Ludovic Ganée <ludovic.ganee@libriciel.coop>
 * @copyright   (c) 2017, Libriciel
 * @license     http://www.cecill.info/licences/Licence_CeCILL_V2-fr.html
 */

namespace Beanstalk\Model\Entity;

use Cake\ORM\Entity;
use Beanstalk\Utility\CakeEmitter;

/**
 * Classe de l'entité
 *
 * <b>beanstalkStats()</b> permet de récupérer et de stocker les informations
 * beanstalk sur le job de l'entité.
 * Cette information est récupérée par les méthodes des champs virtuels.
 *
 * <b>getServerBeanstalkStatus()</b> permet de récupérer les informations du
 * serveur beanstalk du job.
 */
class BeanstalkJob extends Entity
{
    /**
     * Champs virtuels
     *
     * @var array
     */
    protected $_virtual = [
        'state', 'age', 'data', 'statefr', 'related_id'
    ];

    /**
     * Stock le résultat de la function beanstalkStats
     *
     * @var array
     */
    protected $_stats = [];

    /**
     * Récupère les informations beanstalk sur un job en fonction de son id
     *
     * @return array
     * @throws \Exception
     */
    public function beanstalkStats()
    {
        if (empty($this->id) || !CakeEmitter::connected()) {
            return [];
        }

        if (!isset($this->_stats[$this->id])) {
            $this->_stats = [];
            $this->_stats[$this->id] = CakeEmitter::getEmitter()
                ->Pheanstalk->statsJob($this->_properties['jobid']);
        }

        return $this->_stats[$this->id];
    }


    /**
     * Demande le statut du serveur Beanstalkd
     *
     * @return array|object|\Pheanstalk\Response
     * @throws \Exception
     */
    public function getServerBeanstalkStatus()
    {
        if (CakeEmitter::connected()) {
            return CakeEmitter::getEmitter()->Pheanstalk->stats();
        } else {
            return [
                'current-jobs-urgent' => null,
                'current-jobs-ready' => null,
                'current-jobs-reserved' => null,
                'current-jobs-delayed' => null,
                'current-jobs-buried' => null,
                'current-workers' => null,
                'uptime' => null,
            ];
        }
    }

    /**
     * Champ virtuel
     *
     * @return string
     * @throws \Exception
     */
    protected function _getState()
    {
        $stats = $this->beanstalkStats();
        return $stats ? $stats['state'] : false;
    }

    /**
     * Champ virtuel
     *
     * @return string
     * @throws \Exception
     */
    protected function _getAge()
    {
        $stats = $this->beanstalkStats();
        return $stats ? $stats['age'] : false;
    }

    /**
     * Champ virtuel
     *
     * @return string
     * @throws \Exception
     */
    protected function _getData()
    {
        if (CakeEmitter::connected() && $this->id) {
            return CakeEmitter::getEmitter()
                ->Pheanstalk->peek($this->_properties['jobid'])->getData();
        } else {
            return false;
        }
    }

    /**
     * Champ virtuel
     *
     * @return string
     * @throws \Exception
     */
    protected function _getStatefr()
    {
        switch ($this->_getState()) {
            case 'ready': return 'Prêt';
            case 'buried': return 'En pause';
            case 'delayed': return 'En attente';
            case 'reserved': return 'En cours';
        }
        return null;
    }

    /**
     * Champ virtuel
     *
     * @return string
     * @throws \Exception
     */
    protected function _getRelatedId()
    {
        $data = json_decode($this->_getData());
        if ($data && !empty($data->data->id)) {
            return $data->data->id;
        }
        return null;
    }
}
